/*
 * Codetrail
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package codetrail;

import java.rmi.RemoteException;
import java.util.List;

/**
 * Represents a shared space of RDF triples.
 * 
 * @author maxg
 */
public interface GraphSpace extends GraphSpaceBase {
    
    /**
     * Close the graph space.
     * 
     * @throws RemoteException
     */
    public void close() throws RemoteException;
    
    /**
     * Add zero or more triples to the graph space.
     * 
     * @param tuples triples to add
     * @return this
     * @throws RemoteException
     */
    public GraphSpace add(Tuple... tuples) throws RemoteException;
    
    /**
     * Add zero or more triples to the graph space.
     * 
     * @param tuples list of triples to add
     * @return this
     * @throws RemoteException
     */
    public GraphSpace add(List<Tuple> tuples) throws RemoteException;
    
    /**
     * Remove zero or more triples from the graph space.
     * 
     * @param tuples triples to remove
     * @return this
     * @throws RemoteException
     */
    public GraphSpace remove(Tuple... tuples) throws RemoteException;
    
    /**
     * Remove zero or more triples from the graph space.
     * 
     * @param tuples list of triples to remove
     * @return this
     * @throws RemoteException
     */
    public GraphSpace remove(List<Tuple> tuples) throws RemoteException;
    
    /**
     * Remove all triples matching a template.
     * Non-null fields of the template must match exactly. Null fields act
     * as wildcards.
     * 
     * @param template triple template
     * @return this
     * @throws RemoteException
     */
    public GraphSpace removeAll(Tuple template) throws RemoteException;
    
    /**
     * Remove all triples.
     * 
     * @return this
     * @throws RemoteException
     */
    public GraphSpace removeAll() throws RemoteException;
    
    /**
     * Retrieve triples matching a template.
     * Non-null fields of the template must match exactly. Null fields act
     * as wildcards.
     * 
     * @param template triple template
     * @return iterable list of triples matching the template
     * @throws RemoteException
     */
    public Iterable<Tuple> get(Tuple template) throws RemoteException;
    
    /**
     * Run a SPARQL query over the graph space.
     * 
     * @param queryString query
     * @return iterable list of solutions to the query
     * @throws RemoteException
     */
    public Iterable<Solution> query(String queryString) throws RemoteException;
    
    /**
     * Run a SPARQL/Update statement on the graph space.
     * 
     * @param updateString update
     * @throws RemoteException
     */
    public void update(String updateString) throws RemoteException;
    
    /**
     * Register a listener for additions to the graph space. Listeners must
     * specify both a template triple and an "anchored" SPARQL query. For
     * each new triple matching the template, the given query will be run,
     * "anchored" at that triple; the listener is notified of resulting
     * solutions to the query.
     * 
     * @param listener listener to be notified
     * @param template triple template to match added triples
     * @param queryString query to run "anchored" at matching triples
     * @throws RemoteException
     */
    public void register(GraphSpaceListener listener, Tuple template, String queryString) throws RemoteException;
    
    /**
     * Unregister a listener from all of its current registrations.
     * 
     * @param listener listener to remove
     * @throws RemoteException
     */
    public void unregister(GraphSpaceListener listener) throws RemoteException;
    
    /**
     * Register a listener for base events on the graph space.
     * 
     * @param listener listener to be notified
     * @throws RemoteException
     */
    public void register(GraphSpaceBase listener) throws RemoteException;
    
    /**
     * Unregister a listener from base event registration.
     * 
     * @param listener listener to remove
     * @throws RemoteException
     */
    public void unregister(GraphSpaceBase listener) throws RemoteException;
    
    /**
     * Register a listener for answering questions directed to the given
     * source.
     * 
     * @param answerer listener to handle questions
     * @param source URI of source for which questions will be handled
     * @throws RemoteException
     */
    public void register(GraphSpaceListener answerer, URIString source) throws RemoteException;
    
    /**
     * Ask the user a question at the given source and receive notification of
     * the answer.
     * 
     * @param listener listener to be notified of the user's answer
     * @param source URI of source at which to ask
     * @param message question
     * @param options potential answers
     * @throws RemoteException
     */
    public void ask(GraphSpaceListener listener, URIString source, String message, String... options) throws RemoteException;
    
    /**
     * Report the user's answer to a question.
     * 
     * @param question URI of question
     * @param answer user's answer
     * @throws RemoteException
     */
    public void answer(URIString question, String answer) throws RemoteException;
    
    /**
     * Returns the number of triples in this graph space.
     * 
     * @return number of triples in the graph space
     */
    public long size();
    
    /**
     * Returns a XML representation of the contents of this graph space.
     * 
     * @return a XML representation of the graph space
     * @throws RemoteException
     */
    public String toXML() throws RemoteException;
    
}
