/*
 * Codetrail
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package codetrail;

import java.util.*;

/**
 * Trie data structure.
 * 
 * @author maxg
 *
 * @param <K> key component type
 * @param <V> value type
 */
public class Trie<K, V> implements Map<List<K>, V> {
    
    static Comparator<List<?>> LENGTH_COMPARATOR = new Comparator<List<?>>() {
        public int compare(List<?> l0, List<?> l1) {
            if (l0.size() == l1.size()) {
                return l0.toString().compareTo(l1.toString());
            }
            return l0.size() - l1.size();
        }
    };
    
    private static class Node<K> {
        
        final Set<List<K>> nexts;

        public Node(List<K> next) {
            nexts = new HashSet<List<K>>();
            nexts.add(new ArrayList<K>(next));
        }

        public void ensure(List<K> next) {
            nexts.add(new ArrayList<K>(next));
        }
    }
    
    private final Map<List<K>, Node<K>> edges;
    private final Map<List<K>, V> nodes;
    
    /**
     * Constructs an empty <code>Trie</code>.
     */
    public Trie() {
        edges = new HashMap<List<K>, Node<K>>();
        nodes = new HashMap<List<K>, V>();
    }
    
    /**
     * @see Trie#near(List, int)
     */
    public SortedMap<List<K>, V> near(K[] key, int minimum) {
        return near(Arrays.asList(key), minimum);
    }
    
    /**
     * Returns mappings <i>near</i> the given key.
     * 
     * @param key a key in the trie
     * @param minimum minimum number of matching key components
     * @return key-value pairs near the given key
     */
    public SortedMap<List<K>, V> near(List<K> key, int minimum) {
        if ( ! nodes.containsKey(key)) throw new IllegalArgumentException("Key must be in trie");
        
        SortedMap<List<K>, V> map = new TreeMap<List<K>, V>(LENGTH_COMPARATOR);
        
        List<K> parent = new ArrayList<K>();
        for (int i = 0; i < key.size() && edges.get(parent) != null; i++) {
            parent.add(key.get(i));
        }
        do {
            if (parent.size() < minimum) {
                return map;
            }
            parent.remove(parent.size()-1);
        } while(edges.get(parent).nexts.size() == 1);
        
        Set<List<K>> children = new HashSet<List<K>>();
        children.addAll(edges.get(parent).nexts);
        while ( ! children.isEmpty()) {
            Iterator<List<K>> it = children.iterator();
            List<K> child = it.next();
            it.remove();
            if (nodes.containsKey(child)) {
                map.put(child, nodes.get(child));
            }
            if (edges.get(child) != null) {
                children.addAll(edges.get(child).nexts);
            }
        }
        
        map.remove(key);
        return map;
    }
    
    public synchronized void clear() {
        edges.clear();
        nodes.clear();
    }

    public boolean containsKey(Object key) {
        return nodes.containsKey(key);
    }

    public boolean containsValue(Object value) {
        return nodes.containsValue(value);
    }

    public Set<Map.Entry<List<K>, V>> entrySet() {
        return nodes.entrySet();
    }

    public V get(Object key) {
        return nodes.get(key);
    }

    public boolean isEmpty() {
        return nodes.isEmpty();
    }

    public Set<List<K>> keySet() {
        return nodes.keySet();
    }

    /**
     * @see Trie#put(List, Object)
     */
    public V put(K[] key, V value) {
        return put(Arrays.asList(key), value);
    }

    public synchronized V put(List<K> key, V value) {
        List<K> curr = new ArrayList<K>();
        List<K> next = new ArrayList<K>();
        for (int i = 0; i < key.size(); i++) {
            if (i > 0) curr.add(key.get(i-1));
            next.add(key.get(i));
            if (edges.containsKey(curr)) {
                edges.get(curr).ensure(next);
            } else {
                edges.put(new ArrayList<K>(curr), new Node<K>(next));
            }
        }
        return nodes.put(new ArrayList<K>(key), value);
    }

    public void putAll(Map<? extends List<K>, ? extends V> t) {
        for (Map.Entry<? extends List<K>, ? extends V> entry : t.entrySet()) {
            put(entry.getKey(), entry.getValue());
        }
    }

    public synchronized V remove(Object key) {
        return nodes.remove(key); // XXX should clean up edges!
    }

    public int size() {
        return nodes.size();
    }

    public Collection<V> values() {
        return nodes.values();
    }
    
    @Override public String toString() {
        return nodes.toString();
    }
}