/*
 * Codetrail
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package codetrail;

import static codetrail.URIString.rdf_type;
import static codetrail.URIString.uri;

import java.util.Date;
import java.util.UUID;

import jsjr.ConnectorImpl;
import jsjr.Marshalable;

/**
 * Represents a single RDF triple.
 * 
 * @author maxg
 */
public abstract class Tuple implements Marshalable {
    
    /**
     * Create a triple with literal object.
     * 
     * @param subject subject resource
     * @param predicate predicate resource
     * @param object object literal
     * @return the new triple
     */
    public static Tuple tuple(URIString subject, URIString predicate, String object) {
        return new TupleLeaf(subject, predicate, object);
    }
    
    /**
     * Create a triple with resource object.
     * 
     * @param subject subject resource
     * @param predicate predicate resource
     * @param object object resource
     * @return the new triple
     */
    public static Tuple tuple(URIString subject, URIString predicate, URIString object) {
        return new TupleEdge(subject, predicate, object);
    }
    
    /**
     * Create a triple with unique subject, predicate <tt>rdf:type</tt>, and
     * resource object with the given URI.
     * 
     * @param type URI of the type
     * @return the "unique" triple
     */
    public static Tuple unique(String type) {
        return tuple(uri(type + "-" + UUID.randomUUID().toString()),
                     rdf_type,
                     uri(type));
    }
    
    /**
     * Create a triple with the given subject, predicate
     * <tt>graphspace:timestamp</tt>, and literal object with the current
     * date and time.
     * 
     * @param subject subject resource
     * @return the "timestamp" triple
     */
    public static Tuple timestamp(URIString subject) {
        return tuple(subject,
                     uri("#timestamp"),
                     Util.iso8601().format(new Date()));
    }
    
    /**
     * Triple subject resource.
     */
    public final URIString subject;
    /**
     * Triple predicate resource.
     */
    public final URIString predicate;
    /**
     * Retrieve the triple object.
     * 
     * @return triple object
     */
    public abstract Object object();
    
    private final String hashString;
    
    Tuple(URIString subject, URIString predicate, Object object) {
        this.subject = subject;
        this.predicate = predicate;
        this.hashString = ("\n"+subject+"\n"+predicate+"\n"+object);
    }
    
    @Override public boolean equals(Object o) {
        if (o == null || ! getClass().equals(o.getClass())) return false;
        Tuple other = (Tuple)o;
        return hashString.equals(other.hashString);
    }
    
    @Override public int hashCode() {
        return hashString.hashCode();
    }

    public abstract <T> T accept(TupleVisitor<T> visitor);
}

class TupleLeaf extends Tuple {
    private static final long serialVersionUID = 1L;
    
    public final String object;
    
    public TupleLeaf(URIString subject, URIString predicate, String object) {
        super(subject, predicate, object);
        this.object = object;
    }
    
    public String object() {
        return object;
    }

    public <T> T accept(TupleVisitor<T> visitor) {
        return visitor.visit(this);
    }
    
    public String marshal() {
        return subject.marshal() + predicate.marshal() + ConnectorImpl.marshal(object)
                + "serial:class:codetrail.TupleLeaf&arguments:3\n";
    }
    
    @Override public String toString() {
        return "T<" + subject + "," + predicate + "," + object + ">";
    }
}

class TupleEdge extends Tuple {
    private static final long serialVersionUID = 1L;
    
    public final URIString object;
    
    public TupleEdge(URIString subject, URIString predicate, URIString object) {
        super(subject, predicate, object);
        this.object = object;
    }
    
    public URIString object() {
        return object;
    }
    
    public <T> T accept(TupleVisitor<T> visitor) {
        return visitor.visit(this);
    }
    
    public String marshal() {
        return subject.marshal() + predicate.marshal() + object.marshal()
                + "serial:class:codetrail.TupleEdge&arguments:3\n";
    }
    
    @Override public String toString() {
        return "T<" + subject + "," + predicate + "," + object + ">";
    }
}

interface TupleVisitor<T> {

    T visit(TupleLeaf tuple);
    
    T visit(TupleEdge tuple);

}