/*
 * Codetrail
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package codetrail;

import static codetrail.URIString.uri;

import com.hp.hpl.jena.rdf.model.*;

/**
 * RDF triple utilities.
 * 
 * Methods are static utility methods rather than instance methods of
 * <code>Tuple</code> so that <code>Tuple</code>s do not depend on RDF
 * library classes.
 * 
 * @author maxg
 */
public abstract class TupleUtil {
    
    public static Tuple tuple(Statement statement) {
        URIString subject = uri(statement.getSubject().getURI());
        URIString predicate = uri(statement.getPredicate().getURI());
        try {
            return Tuple.tuple(subject,
                               predicate,
                               statement.getLiteral().getString());
        } catch (LiteralRequiredException lre) {
            return Tuple.tuple(subject,
                               predicate,
                               uri(statement.getObject().toString()));
        }
    }
    
    public static Selector selectorIn(Tuple template, final Model model) {
        return template.accept(new TupleVisitor<Selector>() {
            public Selector visit(TupleLeaf t) {
                return new SimpleSelector(
                        t.subject == null ? null : model.createResource(t.subject.uri),
                        t.predicate == null ? null : model.createProperty(t.predicate.uri),
                        t.object == null ? null : model.createLiteral(t.object));
            }

            public Selector visit(TupleEdge t) {
                return new SimpleSelector(
                        t.subject == null ? null : model.createResource(t.subject.uri),
                        t.predicate == null ? null : model.createProperty(t.predicate.uri),
                        t.object == null ? null : model.createResource(t.object.uri));
            }
        });
    }
    
    public static Statement statementIn(Tuple tuple, final Model model) {
        return tuple.accept(new TupleVisitor<Statement>() {

            public Statement visit(TupleLeaf tuple) {
                return model.createStatement(
                        model.createResource(tuple.subject.uri),
                        model.createProperty(tuple.predicate.uri),
                        model.createLiteral(tuple.object));
            }

            public Statement visit(TupleEdge tuple) {
                return model.createStatement(
                        model.createResource(tuple.subject.uri),
                        model.createProperty(tuple.predicate.uri),
                        model.createResource(tuple.object.uri));
            }
        });
    }
    
    public static boolean matches(Tuple tuple, final Statement statement) {
        return tuple.accept(new TupleVisitor<Boolean>() {

            public Boolean visit(TupleLeaf tuple) {
                return subjectMatches(tuple, statement.getSubject())
                       && predicateMatches(tuple, statement.getPredicate())
                       && objectMatches(tuple, statement.getObject());
            }

            public Boolean visit(TupleEdge tuple) {
                return subjectMatches(tuple, statement.getSubject())
                       && predicateMatches(tuple, statement.getPredicate())
                       && objectMatches(tuple, statement.getObject());
            }
        });
    }
    
    public static boolean matches(Tuple tuple, final Tuple other) {
        if (subjectMatches(tuple, other) && predicateMatches(tuple, other)) {
            return tuple.accept(new TupleVisitor<Boolean>() {

                public Boolean visit(TupleLeaf tuple) {
                    return objectMatches(tuple, other);
                }

                public Boolean visit(TupleEdge tuple) {
                    return objectMatches(tuple, other);
                }
            });
        }
        return false;
    }
    
    private static boolean subjectMatches(Tuple tuple, Resource subject) {
        return tuple.subject == null || tuple.subject.uri.equals(subject.getURI());
    }
    
    private static boolean subjectMatches(Tuple tuple, Tuple other) {
        return tuple.subject == null || tuple.subject.equals(other.subject);
    }
    
    private static boolean predicateMatches(Tuple tuple, Property property) {
        return tuple.predicate == null || tuple.predicate.uri.equals(property.getURI());
    }
    
    private static boolean predicateMatches(Tuple tuple, Tuple other) {
        return tuple.predicate == null || tuple.predicate.equals(other.predicate);
    }
    
    private static boolean objectMatches(TupleLeaf tuple, RDFNode object) {
        return tuple.object == null || tuple.object.toString().equals(object.toString());
    }
    
    private static boolean objectMatches(TupleEdge tuple, RDFNode object) {
        return tuple.object == null || tuple.object.uri.equals(object.toString());
    }
    
    private static boolean objectMatches(final TupleLeaf tuple, Tuple other) {
        return tuple.object == null || other.accept(new TupleVisitor<Boolean>() {

            public Boolean visit(TupleLeaf other) {
                return tuple.object.equals(other.object);
            }

            public Boolean visit(TupleEdge other) {
                return false;
            }
        });
    }
    
    private static boolean objectMatches(final TupleEdge tuple, Tuple other) {
        return tuple.object == null || other.accept(new TupleVisitor<Boolean>() {

            public Boolean visit(TupleLeaf other) {
                return false;
            }

            public Boolean visit(TupleEdge other) {
                return tuple.object.equals(other.object);
            }
        });
    }
}
