/*
 * Codetrail
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package codetrail;

import java.io.IOException;
import java.io.OutputStream;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.*;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.*;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.xpath.*;

import org.w3c.dom.*;

/**
 * Utilities.
 * 
 * @author maxg
 */
public abstract class Util {
    
    private static final ThreadLocal<DateFormat> ISO8601 = new ThreadLocal<DateFormat>() {
        @Override protected DateFormat initialValue() {
            return new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
        }
    };
    /**
     * Date formatter conforming to the ISO 8601 specification.
     * 
     * @return ISO 8601 date formatter
     */
    public static DateFormat iso8601() {
        return ISO8601.get();
    }
    
    private static final ThreadLocal<DateFormat> SQLITE = new ThreadLocal<DateFormat>() {
        @Override protected DateFormat initialValue() {
            DateFormat sqlite = new SimpleDateFormat("yyyy-MM-dd' 'HH:mm:ss");
            sqlite.setTimeZone(TimeZone.getTimeZone("GMT"));
            return sqlite;
        }
    };
    /**
     * Date formatter matching the format of SQLite.
     * 
     * @return SQLite date formatter
     */
    public static DateFormat sqlite() {
        return SQLITE.get();
    }
    
    /**
     * XML document builder factory.
     */
    public static final DocumentBuilderFactory DOC_BUILDERER = DocumentBuilderFactory.newInstance();
    
    /**
     * XPath factory.
     */
    public static final XPathFactory XPATHER = XPathFactory.newInstance();
    
    public static final Node NODE = null;
    public static final Element ELEMENT = null;
    
    private Util() { }
    
    /**
     * Return a <code>Properties</code> initialized from the properties file for
     * the specified class. The properties file must be located in the same
     * directory as the class file. The properties file for class
     * <code>Foo</code> must be named <code>Foo.properties</code>.
     * 
     * @param clazz class to load properties for
     * @return a <code>Properties</code> initialized with properties for the
     *         specified class if they could be loaded
     */
    public static Properties propertiesFor(Class<?> clazz) {
        Properties properties = new Properties();
        try {
            properties.load(clazz.getResourceAsStream(clazz.getSimpleName() + ".properties"));
        } catch (IOException ioe) { }
        return properties;
    }
    
    /**
     * Return the value of an environment variable, or the given default if it
     * has no value.
     * 
     * @param name name of environment variable
     * @param defaultValue default value
     * @return environment variable value, or the default
     */
    public static String environment(String name, String defaultValue) {
        String value = System.getenv(name);
        return (value == null ? defaultValue : value);
    }
    
    /**
     * Return the value of an environment variable, or the given default if it
     * has no value.
     * 
     * @param name name of environment variable
     * @param defaultValue default value
     * @return environment variable value, or the default
     */
    public static int environment(String name, int defaultValue) {
        String value = System.getenv(name);
        return (value == null ? defaultValue : Integer.parseInt(value));
    }
    
    /**
     * Returns a human-readable comma-separated string listing of items.
     * Elements are separated by &quot;, &quot &mdash; a comma and a space.
     * 
     * @param items collection of items to list
     * @return human-readable comma-separated string list
     */
    public static String commaSeparated(Collection<?> items) {
        return Arrays.toString(items.toArray()).substring(1).replace("]", "");
    }
    
    /**
     * Splits a comma-separated string listing of items into a list. Elements
     * must be separated by &quot;, &quot &mdash; a comma and a space.
     * 
     * @param items comma-separated string listing of items
     * @return list of items
     */
    public static List<String> commaSeparate(String items) {
        return new ArrayList<String>(Arrays.asList(items.split(", ")));
    }
    
    /**
     * Returns a string representation of items.
     * 
     * @param items collection of items
     * @param separator separator between items
     * @return string of items with the given separator
     */
    public static String separated(Collection<?> items, String separator) {
        StringBuilder buff = new StringBuilder();
        for (Object item : items) {
            buff.append(separator).append(item);
        }
        return buff.substring(separator.length());
    }
    
    /**
     * Returns the first child of <code>doc</code> of the given type,
     * creating one if none exists.
     * 
     * @param doc XML document
     * @param childType child tag
     * @return child of <code>doc</code> of type <code>childType</code>
     */
    public static Element ensureChild(Document doc, String childType) {
        Element child = (Element)doc.getElementsByTagName(childType).item(0);
        if (child == null) {
            doc.appendChild(child = doc.createElement(childType));
        }
        return child;
    }
    
    /**
     * Returns the first child of <code>parent</code> of the given type,
     * creating one if none exists.
     * 
     * @param doc XML document
     * @param parent parent node in <code>doc</code>
     * @param childType child tag
     * @return child of <code>parent</code> of type <code>childType</code>
     */
    public static Element ensureChild(Document doc, Element parent, String childType) {
        Element child = (Element)parent.getElementsByTagName(childType).item(0);
        if (child == null) {
            parent.appendChild(child = doc.createElement(childType));
        }
        return child;
    }
    
    /**
     * Returns the value element of the given key in an Apple property list
     * structure, creating the key-value pair if it does not exist.
     * 
     * @param doc <tt>plist</tt> document
     * @param dict parent <tt>dict</tt> node
     * @param key property-value key name
     * @param valueType property-value value type
     * @return value element of the given key
     * @throws IOException if <code>key</code> has a value node with a type
     *                     other than <code>valueType</code>
     */
    public static Element ensureKeyValue(Document doc, Element dict, String key, String valueType) throws IOException {
        NodeList keys = dict.getElementsByTagName("key");
        Element value = null;
        for (int ii = 0; ii < keys.getLength(); ii++) {
            Node node = keys.item(ii);
            if (node.getTextContent().equals(key)) {
                do {
                    node = node.getNextSibling();
                } while (node.getNodeType() == Node.TEXT_NODE);
                if ( ! node.getNodeName().equals(valueType)) {
                    throw new IOException("Key-value type error");
                } else {
                    value = (Element)node;
                }
            }
        }
        if (value == null) {
            dict.appendChild(value = doc.createElement("key"));
            value.setTextContent(key);
            dict.appendChild(value = doc.createElement(valueType));
        }
        return value;
    }
    
    /**
     * Returns the value element of the given key in an Apple property list
     * structure.
     * 
     * @param dict parent <tt>dict</tt> node
     * @param key property-value key name
     * @return value element of the given key
     * @throws IOException if <code>key</code> has no value node
     */
    public static Element retrieveKeyValue(Element dict, String key) throws IOException {
        NodeList keys = dict.getElementsByTagName("key");
        for (int ii = 0; ii < keys.getLength(); ii++) {
            Node node = keys.item(ii);
            if (node.getTextContent().equals(key)) {
                do {
                    node = node.getNextSibling();
                } while (node.getNodeType() == Node.TEXT_NODE);
                return (Element)node;
            }
        }
        throw new IOException("No such key");
    }
    
    /**
     * Returns the first element of the given name in the given document.
     * 
     * @param document document
     * @param name tag name
     * @return first element with tag <code>name</code>
     */
    public static Element firstElementByTagName(Document document, String name) {
        return (Element)document.getElementsByTagName(name).item(0);
    }
    
    /**
     * Returns the first child element of the given name under the given
     * parent.
     * 
     * @param element parent node
     * @param name child tag name
     * @return first child with tag <code>name</code>
     */
    public static Element firstElementByTagName(Element element, String name) {
        return (Element)element.getElementsByTagName(name).item(0);
    }
    
    /**
     * Writes a XML document to a stream.
     * 
     * @param doc input XML document
     * @param out output stream
     * @throws TransformerException if the document cannot be transformed to text
     */
    public static void transform(Document doc, OutputStream out) throws TransformerException {
        Transformer trans = TransformerFactory.newInstance().newTransformer();
        trans.setOutputProperty(OutputKeys.INDENT, "yes");
        trans.transform(new DOMSource(doc), new StreamResult(out));
    }
    
    /**
     * Writes a XML document to a stream.
     * 
     * @param doc input XML document
     * @param out output stream
     * @param doctypePublic public part of DOCTYPE declaration
     * @param doctypeSystem system part of DOCTYPE declaration
     * @throws TransformerException if the document cannot be transformed to text
     */
    public static void transform(Document doc, OutputStream out, String doctypePublic, String doctypeSystem) throws TransformerException {
        Transformer trans = TransformerFactory.newInstance().newTransformer();
        trans.setOutputProperty(OutputKeys.INDENT, "yes");
        trans.setOutputProperty(OutputKeys.DOCTYPE_PUBLIC, doctypePublic);
        trans.setOutputProperty(OutputKeys.DOCTYPE_SYSTEM, doctypeSystem);
        trans.transform(new DOMSource(doc), new StreamResult(out));
    }
    
    /**
     * Evaluates a XPath expression.
     * 
     * @param node starting context for evaluation
     * @param xpath XPath expression
     * @return list of <code>Node</code>s that result
     * @throws XPathExpressionException if the expression cannot be compiled
     *                                  or evaluated
     */
    public static NodeList evaluate(Node node, String xpath) throws XPathExpressionException {
        XPathExpression expr = XPATHER.newXPath().compile(xpath);
        return (NodeList)expr.evaluate(node, XPathConstants.NODESET);
    }
    
    /**
     * Construct a <code>Node</code> iterable over a <code>NodeList</code>.
     * 
     * @param <T> type of nodes in the node list
     * @param t instance of <code>T</code>
     * @param nodes node list
     * @return equivalent iterable
     */
    public static <T> Iterable<T> iterable(T t, final NodeList nodes) {
        return new Iterable<T>() {
            public Iterator<T> iterator() {
                return new Iterator<T>() {
                    int i = 0;
                    public boolean hasNext() {
                        return i < nodes.getLength();
                    }

                    @SuppressWarnings("unchecked")
                    public T next() {
                        return (T)nodes.item(i++);
                    }

                    public void remove() {
                        throw new UnsupportedOperationException();
                    }
                };
            }
        };
    }
}
