/*
 * Codetrail
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package codetrail.eclipse;

import java.util.*;

import jsjr.Connector;
import jsjr.ConnectorImpl;

import org.eclipse.jface.util.IPropertyChangeListener;
import org.eclipse.jface.util.PropertyChangeEvent;
import org.eclipse.ui.plugin.AbstractUIPlugin;
import org.osgi.framework.BundleContext;

import codetrail.*;
import codetrail.eclipse.study.Data;

/**
 * Activator class that controls the plug-in life cycle.
 */
public class Activator extends AbstractUIPlugin implements IPropertyChangeListener {
    
    static final Properties PROPERTIES = Util.propertiesFor(Activator.class);
    static final boolean GRAPHSPACE = Boolean.parseBoolean(PROPERTIES.getProperty("graphspace"));
    static final String PREF_MODS_ENABLED = PROPERTIES.getProperty("pref-mods-enabled");

	// The plug-in ID
	public static final String PLUGIN_ID = "edu.mit.csail.uid.codetrail.core";

	// The shared instance
	private static Activator PLUGIN;

	/**
	 * Returns the shared instance.
	 *
	 * @return the shared instance
	 */
	public static Activator activator() {
		return PLUGIN;
	}
	
	private GraphSpace space = new NullGraphSpace();
	private Connector connector = new NullConnector();
	private List<Module> modulesAvailable;
	private final Map<Class<?>, Module> modulesEnabled = Collections.synchronizedMap(new HashMap<Class<?>, Module>());
	
	/**
     * The constructor.
     */
    public Activator() {
    }

    public void start(BundleContext context) throws Exception {
        super.start(context);
        PLUGIN = this;
        
        Log.ok("=== Starting " + PLUGIN_ID + " ===");
        
        if (GRAPHSPACE) {
            space = new EclipsedGraphSpace(getStateLocation());
        }
        
        try {
            connector = new ConnectorImpl(Util.environment("GRAPHSPACE_CONNECTOR_PORT", 3943));
            connector.export(Util.environment("GRAPHSPACE_NAME", "theGraphSpace"), space);
        } catch (Exception e) {
            Log.warn("Failed to bind for remote calls", e);
        }
        
        List<String> classes = Arrays.asList(PROPERTIES.getProperty("all-mods").split("\\s+"));
        //Collections.sort(classes);
        List<Module> mods = new ArrayList<Module>();
        for (String clazz : classes) {
            try {
                if ( ! clazz.equals("")) {
                    mods.add((Module)Class.forName(clazz).newInstance());
                }
            } catch (Exception e) {
                Log.error("Failed to instantiate " + clazz, e);
            }
        }
        modulesAvailable = Collections.unmodifiableList(mods);
        
        getPreferenceStore().addPropertyChangeListener(this);
    }
    
    public void stop(BundleContext context) throws Exception {
        try {
            for (Module module : modulesAvailable()) {
                modulesEnabled.remove(module.getClass());
                module.disconnect();
            }
            space.close();
            connector.close();
        } finally {
            PLUGIN = null;
            super.stop(context);
        }
    }
    
    public GraphSpace space() {
        return space;
    }
    
    public Connector connector() {
        return connector;
    }
    
    List<Module> modulesAvailable() {
        return modulesAvailable;
    }
    
    @SuppressWarnings("unchecked")
    public <T extends Module> T module(Class<T> clazz) {
        return (T)modulesEnabled.get(clazz);
    }
    
    public void propertyChange(PropertyChangeEvent event) {
        List<String> classes = Util.commaSeparate(getPreferenceStore().getString(PREF_MODS_ENABLED));
        Log.ok("Module state: " + classes);
        for (Module module : modulesAvailable()) {
            try {
                if (classes.contains(module.getClass().getName())) {
                    module.connect(space());
                    modulesEnabled.put(module.getClass(), module);
                } else {
                    modulesEnabled.remove(module.getClass());
                    module.disconnect();
                }
            } catch (Exception e) {
                Log.error("Error updating module state", e);
            }
        }
        Data.ideEvent("modules", null, classes.toString());
        Log.ok("Module state updated");
    }
}
