/*
 * Codetrail
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package codetrail.eclipse;

import static codetrail.Tuple.*;
import static codetrail.URIString.*;
import static java.sql.ResultSet.CONCUR_UPDATABLE;
import static java.sql.ResultSet.TYPE_FORWARD_ONLY;

import java.io.ByteArrayOutputStream;
import java.rmi.ConnectException;
import java.rmi.RemoteException;
import java.sql.*;
import java.util.*;
import java.util.Date;
import java.util.concurrent.*;

import org.eclipse.core.runtime.IPath;

import codetrail.*;

import com.hp.hpl.jena.db.DBConnection;
import com.hp.hpl.jena.db.IDBConnection;
import com.hp.hpl.jena.db.impl.DriverRDB;
import com.hp.hpl.jena.db.impl.IRDBDriver;
import com.hp.hpl.jena.graph.*;
import com.hp.hpl.jena.graph.compose.Difference;
import com.hp.hpl.jena.query.*;
import com.hp.hpl.jena.query.ResultSet;
import com.hp.hpl.jena.rdf.listeners.StatementListener;
import com.hp.hpl.jena.rdf.model.*;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.shared.JenaException;
import com.hp.hpl.jena.update.*;
import com.hp.hpl.jena.util.iterator.ExtendedIterator;

/**
 * <code>GraphSpace</code> implementation for Eclipse.
 * 
 * @author maxg
 */
public class EclipsedGraphSpace implements GraphSpace {
    
    static final Properties PROPERTIES = Util.propertiesFor(EclipsedGraphSpace.class);
    
    static IDBConnection dbConnectionFor(String logPath, String dbPath) throws ClassNotFoundException {
        System.getProperties().put(PROPERTIES.getProperty("dblogprop"), logPath);
        Class.forName(PROPERTIES.getProperty("dbdriver"));
        
        String dburl = PROPERTIES.getProperty("dburl");
        dburl = dburl.replace("@PATH@", dbPath);
        String dbengine = PROPERTIES.getProperty("dbengine");
        return new DBConnection(dburl, null, null, dbengine);
    }
    
    private final Model model;
    private final Model incoming;
    private final DataSource datasource;
    private final TuplesToQueriesToListenersMap listeners;
    private final Set<GraphSpaceBase> baseListeners;
    
    private final BlockingQueue<Runnable> addRemove;
    private final Thread adderRemover;
    private final GarbageCollector gc;
    private final Thread gcer;
    
    EclipsedGraphSpace(ModelMaker maker, final GarbageCollector collector) {
        model = maker.openModel(PROPERTIES.getProperty("model"));
        
        model.register(new StatementListener() {
            @Override public void addedStatement(final Statement statement) {
                addRemove.offer(new Runnable() {
                    public void run() {
                        added(statement);
                    }
                });
            }
            @Override public void removedStatement(final Statement statement) {
                addRemove.offer(new Runnable() {
                    public void run() {
                        removed(statement);
                    }
                });
            }
        });
        
        incoming = ModelFactory.createMemModelMaker().createFreshModel();
        
        datasource = DatasetFactory.create(model);
        datasource.addNamedModel(uri("#match").uri, incoming);
        
        addRemove = new LinkedBlockingQueue<Runnable>();
        
        adderRemover = new Thread(new Runnable() {
            public void run() {
                try {
                    while ( ! Thread.interrupted()) {
                        addRemove.take().run();
                    }
                } catch (InterruptedException ie) { }
            }
        }, getClass().getName() + " add/remove");
        adderRemover.setDaemon(true);
        adderRemover.start();
        
        listeners = new TuplesToQueriesToListenersMap();
        baseListeners = Collections.synchronizedSet(new HashSet<GraphSpaceBase>());
        
        gc = collector;
        gcer = new Thread(new Runnable() {
            public void run() {
                try {
                    while ( ! Thread.interrupted()) {
                        Thread.sleep(60 * 1000L);
                        gc(60 * 60 * 1000L);
                    }
                } catch (InterruptedException ie) {
                } catch (Throwable t) {
                    Log.error("Error in gc", t);
                }
            }
        }, getClass().getName() + " gc");
        gcer.setDaemon(true);
        gcer.setPriority(3);
        gcer.start();
    }
    
    /**
     * Create a new Eclipse <code>GraphSpace</code> with no persistent
     * storage.
     */
    EclipsedGraphSpace() {
        this(ModelFactory.createMemModelMaker(), new MemCopyGarbageCollector());
    }
    
    /**
     * Create a new Eclipse <code>GraphSpace</code> with persistent
     * storage.
     * 
     * @param connection database connection
     */
    EclipsedGraphSpace(IDBConnection connection) throws SQLException {
        this(ModelFactory.createModelRDBMaker(connection), new DbMarkSweepGarbageCollector(connection));
    }
    
    /**
     * Create a new Eclipse <code>GraphSpace</code> with persistent
     * storage.
     * 
     * @param parent parent directory of persistent database
     * @throws ClassNotFoundException if the database driver cannot be found
     * @throws SQLException if the database cannot be opened
     */
    EclipsedGraphSpace(IPath parent) throws ClassNotFoundException, SQLException {
        this(dbConnectionFor(parent.append(PROPERTIES.getProperty("dblogname")).toOSString(),
                             parent.append("EclipsedGraphSpace.db").toOSString()));
        Log.ok("GraphSpace up with " + model.size() + " statements");
    }
    
    public void close() {
        adderRemover.interrupt();
        gcer.interrupt();
        try {
            adderRemover.join();
            gcer.join();
        } catch (InterruptedException ie) { }
        model.close();
    }
    
    /**
     * Run the garbage collector.
     * 
     * @param age root candidates younger than <code>age</code> will be used
     *            as roots
     */
    void gc(long age) throws Exception {
        gc.gc(model, incoming, age);
    }

    private void added(Statement statement) {
        for (Map.Entry<Tuple, QueriesToListenersMap> tql : listeners.entrySet()) {
            if (TupleUtil.matches(tql.getKey(), statement)) {
                for (Map.Entry<Query, List<GraphSpaceListener>> ql : tql.getValue().entrySet()) {
                    addedMatching(statement, ql.getKey(), ql.getValue());
                }
            }
        }
        Tuple tuple = TupleUtil.tuple(statement);
        synchronized (baseListeners) {
            for (Iterator<GraphSpaceBase> it = baseListeners.iterator(); it.hasNext(); ) {
                try {
                    it.next().add(tuple);
                } catch (RemoteException re) {
                    it.remove();
                }
            }
        }
    }
    
    private void removed(Statement statement) {
        Tuple tuple = TupleUtil.tuple(statement);
        synchronized (baseListeners) {
            for (Iterator<GraphSpaceBase> it = baseListeners.iterator(); it.hasNext(); ) {
                try {
                    it.next().remove(tuple);
                } catch (RemoteException re) {
                    it.remove();
                }
            }
        }
    }

    @SuppressWarnings("unchecked")
    private void addedMatching(Statement statement, Query query, List<GraphSpaceListener> listeners) {
        List<Solution> solutions = new ArrayList<Solution>();
        synchronized (incoming) {
            try {
                incoming.add(statement);
                QueryExecution qexec = QueryExecutionFactory.create(query, datasource);
                ResultSet results = qexec.execSelect();
                while (results.hasNext()) {
                    solutions.add(SolutionUtil.solution(results.nextSolution()));
                }
            } finally {
                incoming.removeAll();
            }
        }
        for (Solution solution : solutions) {
            for (GraphSpaceListener listener : listeners) {
                try {
                    listener.notify(solution);
                } catch (ConnectException ce) {
                    // XXX removing, but is he really gone?
                    Log.info("Listener disappeared: " + ce.getMessage().replace('\n', '_'));
                    unregister(listener);
                } catch (Exception e) {
                    Log.warn("Exception from listener", e);
                }
            }
        }
    }
    
    public GraphSpace initialize(String xml) {
        throw new UnsupportedOperationException("Cannot initialize");
    }
    
    public GraphSpace add(Tuple tuple) {
        synchronized (incoming) {
            model.add(TupleUtil.statementIn(tuple, model));
            return this;
        }
    }
    
    public GraphSpace add(Tuple... tuples) {
        synchronized (incoming) {
            for (Tuple tuple : tuples)
                add(tuple);
        }
        return this;
    }
    
    public GraphSpace add(List<Tuple> tuples) {
        synchronized (incoming) {
            for (Tuple tuple : tuples)
                add(tuple);
        }
        return this;
    }
    
    public GraphSpace remove(Tuple tuple) {
        synchronized (incoming) {
            model.remove(TupleUtil.statementIn(tuple, model));
            return this;
        }
    }
    
    public GraphSpace remove(Tuple... tuples) {
        synchronized (incoming) {
            for (Tuple tuple : tuples)
                remove(tuple);
        }
        return this;
    }
    
    public GraphSpace remove(List<Tuple> tuples) {
        synchronized (incoming) {
            for (Tuple tuple : tuples)
                remove(tuple);
        }
        return this;
    }
    
    public GraphSpace removeAll(Tuple template) {
        synchronized (incoming) {
            Selector selector = TupleUtil.selectorIn(template, model);
            model.removeAll(selector.getSubject(), selector.getPredicate(), selector.getObject());
        }
        return this;
    }
    
    public GraphSpace removeAll() {
        synchronized (incoming) {
            model.removeAll();
        }
        return this;
    }
    
    public Iterable<Tuple> get(Tuple template) {
        final StmtIterator it;
        synchronized (incoming) {
            it = model.listStatements(TupleUtil.selectorIn(template, model));
        }
        // XXX this will not make it over the wire!
        return new Iterable<Tuple>() {
            public Iterator<Tuple> iterator() {
                return new Iterator<Tuple>() {
                    public boolean hasNext() {
                        return it.hasNext();
                    }

                    public Tuple next() {
                        return TupleUtil.tuple(it.nextStatement());
                    }

                    public void remove() {
                        it.remove();
                    }
                };
            }
        };
    }
    
    private Query parse(String queryString) {
        Query query = QueryFactory.make();
        query.setPrefix("arqfn", "http://jena.hpl.hp.com/ARQ/function#");
        query.setPrefix("arqprop", "http://jena.hpl.hp.com/ARQ/property#");
        query.setPrefix("graphspace", BASE_URI + SEPARATOR);
        query = QueryFactory.parse(query, queryString, null, Syntax.syntaxARQ);
        return query;
    }
    
    public Iterable<Solution> query(String queryString) {
        try {
            final ResultSet results;
            synchronized (incoming) {
                QueryExecution qexec = QueryExecutionFactory.create(parse(queryString), model);
                results = qexec.execSelect();
            }
            return SolutionUtil.solutionIterable(results);
        } catch (JenaException je) {
            throw new RuntimeException(je.getMessage());
        }
    }
    
    public void update(String updateString) {
        try {
            updateString = "PREFIX graphspace: <" + BASE_URI + SEPARATOR + ">\n" + updateString; // XXX hack
            UpdateRequest update = UpdateFactory.create(updateString);
            GraphStore graph = GraphStoreFactory.create();
            synchronized (incoming) {
                graph.setDefaultGraph(model.getGraph());
                update.exec(graph);
            }
        } catch (JenaException je) {
            throw new RuntimeException(je.getMessage());
        }
    }
    
    public void register(GraphSpaceListener listener, Tuple template, String queryString) {
        listeners.add(template, parse(queryString), listener);
    }
    
    public void unregister(GraphSpaceListener listener) {
        for (QueriesToListenersMap ql : listeners.values()) {
            for (List<GraphSpaceListener> l : ql.values()) {
                l.remove(listener);
            }
        }
    }
    
    public void register(GraphSpaceBase listener) {
        synchronized (incoming) {
            baseListeners.add(listener);
            try {
                listener.initialize(toXML());
            } catch (RemoteException re) {
                baseListeners.remove(listener);
            }
        }
    }
    
    public void unregister(GraphSpaceBase listener) {
        baseListeners.remove(listener);
    }
    
    public void register(GraphSpaceListener answerer, URIString source) {
        register(answerer,
                 tuple(null, uri("#questioning"), source),
                 PROPERTIES.getProperty("answererquery"));
    }
    
    public void ask(final GraphSpaceListener questioner, URIString source, String message, String... options) {
        Tuple question = unique("#question");
        register(
            new GraphSpaceListener() {
                public void notify(Solution solution) throws Exception {
                    questioner.notify(solution);
                    unregister(questioner);
                }
            },
            tuple(question.subject, uri("#answer"), (String)null),
            PROPERTIES.getProperty("questionerquery"));
        StringBuilder optionList = new StringBuilder();
        for (String option : options) {
            optionList.append(",").append(option);
        }
        add(question,
            timestamp(question.subject),
            tuple(question.subject, uri("#questioning"), source),
            tuple(question.subject, uri("#message"), message),
            tuple(question.subject, uri("#options"), optionList.substring(1)));
    }

    public void answer(URIString question, String answer) {
        add(tuple(question, uri("#answer"), answer));
    }
    
    public long size() {
        return model.size();
    }

    public String toXML() {
        ByteArrayOutputStream out = new ByteArrayOutputStream();
        model.write(out, "RDF/XML");
        return out.toString();
    }
}

interface GarbageCollector {
    
    /**
     * Run the garbage collector.
     * 
     * <p>Root candidates are subjects with a <code>rdf:type</code>.
     * Candidates with a <code>graphspace:timestamp</code> property may be
     * used as roots as long as they are young enough. Candidates with no
     * such property will always be used as roots.</p>
     * 
     * @param model model to collect garbage in
     * @param mutex if the collector must stop the world, it may do so by
     *              obtaining the lock on <code>mutex</code>
     * @param age root candidates younger than <code>age</code> will be used
     *            as roots
     */
    public void gc(Model model, Object mutex, long age) throws Exception;
}

class MemCopyGarbageCollector implements GarbageCollector {
    
    public void gc(Model model, Object mutex, long age) {
        Date cutoff = new Date(new Date().getTime() - age);
        Graph all = model.getGraph();
        Graph live = Factory.createGraphMem();
        GraphExtract extract = new GraphExtract(TripleBoundary.stopNowhere);
        Property gsTimestamp = model.createProperty(uri("#timestamp").uri);
        
        synchronized (mutex) {
            ResIterator roots = model.listSubjectsWithProperty(model.createProperty(rdf_type.uri));
            while (roots.hasNext()) {
                Resource root = roots.nextResource();
                boolean young = true;
                try {
                    Statement timestamp = root.getProperty(gsTimestamp);
                    if (timestamp == null)
                        continue;
                    Date ts = Util.iso8601().parse(timestamp.getString());
                    young = ts.after(cutoff);
                } catch (Exception e) {
                    Log.warn("Possible tuple leak: " + root, e); // XXX
                } finally {
                    if (young)
                        extract.extractInto(live, root.asNode(), all);
                }
            }
            Difference diff = new Difference(all, live);
            ExtendedIterator dead = diff.find(null, null, null);
            while (dead.hasNext()) {
                dead.next();
                dead.remove();
            }
        }
    }
}

class DbMarkSweepGarbageCollector implements GarbageCollector {
    
    static abstract class DriverRDBConstants extends DriverRDB {
        static final String LITERAL = RDBCodeLiteral;
    }
    
    final String MODEL_NAME = EclipsedGraphSpace.PROPERTIES.getProperty("model"); // XXX
    
    private final Connection db;
    private final IRDBDriver driver;
    private String table;
    private int id;
    
    public DbMarkSweepGarbageCollector(IDBConnection connection) throws SQLException {
        db = DriverManager.getConnection(connection.getConnection().getMetaData().getURL());
        driver = connection.getDriver();
        table = null;
    }
    
    void prepareTable() throws SQLException {
        PreparedStatement find = db.prepareStatement("SELECT ID FROM JENA_GRAPH WHERE NAME = ?");
        find.setString(1, MODEL_NAME); // XXX
        java.sql.ResultSet results = find.executeQuery();
        if ( ! results.next()) throw new SQLException("Unable to find model in database");
        id = results.getInt(1);
        String table = "JENA_G" + id + "T1_STMT";
        find.close();
        
        try {
            db.createStatement().executeUpdate("ALTER TABLE " + table + " ADD COLUMN MARK SMALLINT NOT NULL DEFAULT 1");
        } catch (SQLException sqle) {
            if ( ! sqle.getSQLState().equals("X0Y32")) throw sqle;
        }
        
        this.table = table;
    }
    
    public synchronized void gc(Model model, Object mutex, long age) throws SQLException {
        Date cutoff = new Date(new Date().getTime() - age);
        Property gsTimestamp = model.createProperty(uri("#timestamp").uri);
        
        if (table == null) prepareTable();
        
        java.sql.Statement stmt = db.createStatement();
        
        // clear
        synchronized (mutex) {
            stmt.executeUpdate("UPDATE " + table + " SET MARK = 0");
        }
        
        // mark
        PreparedStatement mark = db.prepareStatement(
                "SELECT OBJ, MARK FROM " + table + " WHERE SUBJ = ? AND GRAPHID = " + id + " AND MARK = 0 FOR UPDATE",
                TYPE_FORWARD_ONLY,
                CONCUR_UPDATABLE);
        final int OBJ = 1, MARK = 2;
        ResIterator roots = model.listSubjectsWithProperty(model.createProperty(rdf_type.uri));
        Queue<String> toVisit = new LinkedList<String>();
        while (roots.hasNext()) {
            Resource root = roots.nextResource();
            boolean young = true;
            try {
                Statement timestamp = root.getProperty(gsTimestamp);
                if (timestamp != null) {
                    Date ts = Util.iso8601().parse(timestamp.getString());
                    young = ts.after(cutoff);
                }
            } catch (Exception e) {
                Log.warn("Possible tuple leak: " + root, e); // XXX
            }
            if ( ! young) {
                continue;
            }
            toVisit.add(driver.nodeToRDBString(root.asNode(), false));
            while ( ! toVisit.isEmpty()) {
                mark.setString(1, toVisit.remove());
                java.sql.ResultSet results = mark.executeQuery();
                while (results.next()) {
                    String obj = results.getString(OBJ);
                    results.updateInt(MARK, 1);
                    if ( ! obj.startsWith(DriverRDBConstants.LITERAL)) {
                        toVisit.add(obj);
                    }
                    results.updateRow();
                }
            }
        }
        mark.close();
        
        //sweep
        stmt.executeUpdate("DELETE FROM " + table + " WHERE MARK = 0 AND GRAPHID = " + id);
        stmt.close();
    }
}

class TuplesToQueriesToListenersMap extends ConcurrentHashMap<Tuple, QueriesToListenersMap> {
    private static final long serialVersionUID = 1L;
    
    void add(Tuple tuple, Query query, GraphSpaceListener listener) {
        this.putIfAbsent(tuple, new QueriesToListenersMap());
        this.get(tuple).add(query, listener);
    }
}

class QueriesToListenersMap extends ConcurrentHashMap<Query, List<GraphSpaceListener>> {
    private static final long serialVersionUID = 1L;
    
    void add(Query query, GraphSpaceListener listener) {
        this.putIfAbsent(query, new CopyOnWriteArrayList<GraphSpaceListener>());
        this.get(query).add(listener);
    }
}