/*
 * Codetrail
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package codetrail.eclipse;

import java.rmi.RemoteException;
import java.util.HashSet;
import java.util.Set;

import org.eclipse.core.resources.*;
import org.eclipse.core.runtime.CoreException;

public class ResourceWatcher implements IResourceChangeListener {
    
    public final String extension;
    private final String folderName; // XXX configurable
        
    protected ResourceWatcher(String extension, String folder) {
        this.extension = extension;
        this.folderName = folder;
    }
    
    protected void connect() throws RemoteException {
        ResourcesPlugin.getWorkspace().addResourceChangeListener(this, IResourceChangeEvent.POST_CHANGE);
        for (IProject project : ResourcesPlugin.getWorkspace().getRoot().getProjects()) {
            if (project.isOpen()) {
                visit(project, new Visitor() {
                    public void visit(IProject project, Set<IFile> files) throws Exception {
                        initialize(project, files);
                    }
                });
            }
        }
    }
    
    protected void disconnect() throws RemoteException {
        ResourcesPlugin.getWorkspace().removeResourceChangeListener(this);
    }
    
    protected IFolder folderIn(IProject project) throws CoreException {
        IFolder folder = project.getFolder(folderName);
        folder.refreshLocal(IResource.DEPTH_INFINITE, null);
        
        if ( ! folder.exists()) {
            folder.create(IResource.NONE, false, null);
        }
        
        return folder;
    }
    
    protected IFile fileIn(IProject project, String filename) throws CoreException {
        IFolder folder = folderIn(project);
        return folder.getFile(filename + "." + extension);
    }
    
    public void resourceChanged(IResourceChangeEvent change) {
        ChangeVisitor visitor = new ChangeVisitor();
        try {
            change.getDelta().accept(visitor);
        } catch (CoreException ce) {
            Log.error("Error visiting resource change", ce);
        }
        for (IProject project : visitor.changed) {
            visit(project, new Visitor() {
                public void visit(IProject project, Set<IFile> files) throws Exception {
                    summarize(project, files);
                }
            });
        }
    }
    
    private void visit(IProject project, Visitor visitor) {
        FolderVisitor summarizer = new FolderVisitor();
        try {
            if (project.getFolder(folderName).exists()) {
                project.getFolder(folderName).accept(summarizer);
            }
        } catch (CoreException ce) {
            Log.error("Error visiting resources", ce);
        }
        try {
            visitor.visit(project, summarizer.resources);
        } catch (Exception e) {
            Log.error("Error reviewing changes", e);
        }
    }
    
    class ChangeVisitor implements IResourceDeltaVisitor {
        Set<IProject> changed = new HashSet<IProject>();
        
        public boolean visit(IResourceDelta delta) {
            if ((delta.getResource() instanceof IFile) &&
                    extension.equals(delta.getResource().getFileExtension())) {
                changed.add(delta.getResource().getProject());
                try {
                    switch (delta.getKind()) {
                    case IResourceDelta.ADDED:
                        added((IFile)delta.getResource());
                        break;
                    case IResourceDelta.CHANGED:
                        changed((IFile)delta.getResource());
                        break;
                    case IResourceDelta.REMOVED:
                        removed((IFile)delta.getResource());
                        break;
                    }
                } catch (Exception e) {
                    Log.error("Error visiting changes", e);
                }
            }
            return folderName.startsWith(delta.getProjectRelativePath().toPortableString()); // XXX configurable
        }
    }
    
    class FolderVisitor implements IResourceVisitor {
        Set<IFile> resources = new HashSet<IFile>();
        
        public boolean visit(IResource resource) throws CoreException {
            if ((resource instanceof IFile) &&
                    extension.equals(resource.getFileExtension())) {
                resources.add((IFile)resource);
            }
            return true;
        }
    }
    
    interface Visitor {
        public void visit(IProject project, Set<IFile> files) throws Exception;
    }
    
    protected void initialize(IProject project, Set<IFile> files) throws Exception { }
    
    protected void added(IFile file) throws Exception { }
    
    protected void changed(IFile file) throws Exception { }
    
    protected void removed(IFile file) throws Exception { }
    
    protected void summarize(IProject project, Set<IFile> files) throws Exception { }
}
