/*
 * Codetrail
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package codetrail.eclipse.mod;

import static codetrail.Tuple.*;
import static codetrail.URIString.rdf_type;
import static codetrail.URIString.uri;
import static codetrail.Util.*;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.rmi.RemoteException;
import java.util.*;

import org.eclipse.core.resources.*;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.jdt.core.IJavaElement;
import org.w3c.dom.*;

import codetrail.*;
import codetrail.eclipse.*;
import codetrail.eclipse.study.Data;

/**
 * Bookmarks module: creates web site bookmarks related to project resources.
 * 
 * @author maxg
 */
public class Bookmarks extends ResourceWatcher implements Module {
    
    static final Properties PROPERTIES = Util.propertiesFor(Bookmarks.class);
    
    private GraphSpace space;
    final SortedSet<Solution> visits;
    final BrowseListener browseListener;
    final RequestListener requestListener;
    
    public Bookmarks() {
        super(PROPERTIES.getProperty("extension"), "bookmarks"); // XXX configurable
        visits = Collections.synchronizedSortedSet(new TreeSet<Solution>(new Comparator<Solution>() {
            public int compare(Solution s1, Solution s2) {
                return s2.get("time").compareTo(s1.get("time"));
            }
        }));
        browseListener = new BrowseListener();
        requestListener = new RequestListener();
    }
    
    public String getName() {
        return "Bookmarks";
    }

    public void connect(GraphSpace space) throws RemoteException {
        if (space.equals(this.space)) return;
        
        this.space = space;
        
        space.add(
                tuple(uri("NC:BookmarksRoot"),
                      rdf_type,
                      uri("#ncBookmarksRoot")),
                tuple(uri("NC:BookmarksRoot"),
                      uri("http://home.netscape.com/NC-rdf#Folder"),
                      uri("#bookmarks")),
                tuple(uri("#bookmarks"),
                      rdf_type,
                      uri("http://home.netscape.com/NC-rdf#Folder")),
                tuple(uri("#bookmarks"),
                      uri("http://home.netscape.com/NC-rdf#Name"),
                      "Eclipse Bookmarks")
        );
        
        space.register(browseListener,
                       tuple(null, uri("#action"), uri("#browse")),
                       PROPERTIES.getProperty("graphspace-query"));
        for (Solution s : space.query(PROPERTIES.getProperty("initialize-query"))) {
            browseListener.notify(s);
        }
        
        space.register(requestListener,
                       tuple(null, uri("#action"), uri("#bookmark-request")),
                       PROPERTIES.getProperty("request-query"));
        
        super.connect();
    }
    
    public void disconnect() throws RemoteException {
        if (space == null) return;
        
        space.unregister(browseListener);
        space.unregister(requestListener);
        super.disconnect();
        space = null;
    }
    
    /**
     * Create a bookmark associated with the given project artifact.
     * 
     * <p>Calls should be wrapped in order to coalesce resource change
     * events.</p>
     * 
     * @see IWorkspace#run(org.eclipse.core.resources.IWorkspaceRunnable,
     * org.eclipse.core.runtime.IProgressMonitor)
     * 
     * @param url URL of the bookmarked site
     * @param title title of the bookmarked site
     * @param project project of artifact
     * @param uriInProject location of artifact in project
     * @throws Exception if any of a variety of core, parser, DOM, IO, or
     *                   transformer exceptions occurs
     */
    void bookmark(String url, String title, IProject project, String uriInProject) throws Exception {
        Log.ok("Bookmarking " + url + ", " + title + ", " + project.getName() + ", " + uriInProject);
        Data.ideEvent("bookmark create", null, url);
        
        IFile file = fileIn(project, title.trim().replaceAll("[^\\w ,\\._-]", ""));
        Document doc;
        if ( ! file.exists()) {
            file.create(null, IResource.NONE, null);
            doc = DOC_BUILDERER.newDocumentBuilder().newDocument();
        } else {
            doc = DOC_BUILDERER.newDocumentBuilder().parse(file.getContents());
        }
        
        Element plist = ensureChild(doc, "plist");
        plist.setAttribute("version", "1.0");
        
        Element dict = ensureChild(doc, plist, "dict");
        
        Element urlString = ensureKeyValue(doc, dict, "URL", "string");
        if (urlString.getTextContent().equals("")) {
            urlString.setTextContent(url);
        } else if ( ! urlString.getTextContent().equals(url)) {
            Log.warn("Stored URL " + urlString.getTextContent() + " doesn't match " + url);
            return;
        }
        
        Element subDict = ensureKeyValue(doc, dict, Activator.PLUGIN_ID, "dict");
        
        Element titleString = ensureKeyValue(doc, subDict, "title", "string");
        if (titleString.getTextContent().equals("")) {
            titleString.setTextContent(title);
        } // else use existing title
        
        Element array = ensureKeyValue(doc, subDict, "resources", "array");
        
        for (Node resource : iterable(NODE, array.getChildNodes())) {
            if (resource.getTextContent().equals(uriInProject)) {
                return; // XXX assumes we made no other changes
            }
        }
        
        Element child;
        array.appendChild(child = doc.createElement("string"));
        child.setTextContent(uriInProject);
        
        ByteArrayOutputStream out = new ByteArrayOutputStream();
        Util.transform(doc, out, PROPERTIES.getProperty("dtd-pub"), PROPERTIES.getProperty("dtd-sys"));
        file.setContents(new ByteArrayInputStream(out.toByteArray()), IResource.NONE, null);
    }
    
    void bookmark(String url, String title, IResource item) throws Exception {
        // XXX bookmark(url, title, item.getProject(), "resource:" + item.getProjectRelativePath().toPortableString());
        bookmark(url, title, item.getProject(), item.getProjectRelativePath().toPortableString());
    }

    void bookmark(String url, String title, IJavaElement item) throws Exception {
        // XXX bookmark(url, title, item.getJavaProject().getProject(), "java:" + item.getHandleIdentifier());
        bookmark(url, title, item.getJavaProject().getProject(), item.getResource().getProjectRelativePath().toPortableString());
    }
    
    class BrowseListener implements GraphSpaceListener {
        public void notify(Solution solution) {
            // XXX not threadsafe
            for (Iterator<Solution> it = visits.iterator(); it.hasNext(); ) {
                if (it.next().get("url").equals(solution.get("url"))) {
                    it.remove();
                    break;
                }
            }
            visits.add(solution);
            Solution last = visits.last();
            if (visits.size() > 10)
                visits.remove(last);
        }
    }
    
    class RequestListener implements GraphSpaceListener {
        public void notify(Solution solution) {
            try {
                bookmark(solution.get("url"),
                         solution.get("title"),
                         ResourcesPlugin.getWorkspace().getRoot().getProject(solution.get("project")),
                         solution.get("path"));
            } catch (Exception e) {
                Log.error("Error adding bookmark", e);
            }
        }
    }
    
    @Override protected void initialize(IProject project, Set<IFile> weblocs) throws Exception {
        summarize(project, weblocs);
    }
    
    @Override protected void summarize(IProject project, Set<IFile> weblocs) throws Exception {
        List<Tuple> tuples = new ArrayList<Tuple>();
        URIString projectUri = uri("#eclipse-project-" + project.getName());
        
        space.update(PROPERTIES.getProperty("clear-statement").replaceAll("@PROJECT@", project.getName()));
        
        if (weblocs.isEmpty()) {
            space.removeAll(tuple(null, null, projectUri));
            space.removeAll(tuple(projectUri, null, (String)null));
            space.removeAll(tuple(projectUri, null, (URIString)null));
        } else {
            tuples.addAll(Arrays.asList(
                    tuple(uri("#bookmarks"),
                          uri("http://home.netscape.com/NC-rdf#child"),
                          projectUri),
                    tuple(uri("#bookmarks"),
                          uri("#contains"),
                          projectUri),
                    tuple(projectUri,
                          rdf_type,
                          uri("http://home.netscape.com/NC-rdf#Folder")),
                    tuple(projectUri,
                          uri("http://home.netscape.com/NC-rdf#Name"),
                          project.getName())
            ));
        }
        
        String tag = project.getName().hashCode() + "-" + System.currentTimeMillis() + "-" + (int)(Math.random()*100);
        Data.ideEvent("bookmark accounting", tag);
        for (IFile webloc : weblocs) {
            Document doc;
            try {
                doc = DOC_BUILDERER.newDocumentBuilder().parse(webloc.getContents());
            } catch (CoreException ce) {
                continue;
            }
            Element plist = firstElementByTagName(doc, "plist");
            Element dict = firstElementByTagName(plist, "dict");
            Element url = retrieveKeyValue(dict, "URL");
            
            dict = retrieveKeyValue(dict, Activator.PLUGIN_ID);
            Element title = retrieveKeyValue(dict, "title");
            
            Tuple bookmark = unique("#bookmark");
            tuples.addAll(Arrays.asList(
                    bookmark,
                    tuple(bookmark.subject, uri("#url"), url.getTextContent()),
                    tuple(bookmark.subject, uri("#title"), title.getTextContent()),
                    tuple(bookmark.subject, uri("#project"), project.getName())
            ));
            
            Element array = retrieveKeyValue(dict, "resources");
            for (Node resource : iterable(NODE, array.getChildNodes())) {
                if (resource.getNodeName().equals("string")) {
                    tuples.add(tuple(bookmark.subject, uri("#resource"), resource.getTextContent()));
                }
            }
            
            tuples.addAll(Arrays.asList(
                    tuple(projectUri,
                          uri("http://home.netscape.com/NC-rdf#child"),
                          bookmark.subject),
                    tuple(uri("#bookmarks"),
                          uri("#contains"),
                          bookmark.subject),
                    tuple(bookmark.subject,
                          rdf_type,
                          uri("http://home.netscape.com/NC-rdf#Bookmark")),
                    tuple(bookmark.subject,
                          uri("http://home.netscape.com/NC-rdf#Name"),
                          title.getTextContent()),
                    tuple(bookmark.subject,
                          uri("http://home.netscape.com/NC-rdf#URL"),
                          url.getTextContent())
            ));
            
            Data.ideEvent("bookmark exists", tag, url.getTextContent());
        }
        
        space.add(tuples);
        
        Tuple event = unique("#event");
        space.add(
                event,
                tuple(event.subject, uri("#action"), uri("#bookmark")),
                timestamp(event.subject)
        );
    }
}