/*
 * Codetrail
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package codetrail.eclipse.mod;

import static codetrail.Tuple.tuple;
import static codetrail.URIString.uri;

import java.net.MalformedURLException;
import java.net.URL;
import java.rmi.RemoteException;
import java.util.*;

import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.jface.action.*;
import org.eclipse.jface.viewers.*;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.ui.*;
import org.eclipse.ui.model.WorkbenchLabelProvider;
import org.eclipse.ui.part.ViewPart;

import codetrail.*;
import codetrail.eclipse.Activator;
import codetrail.eclipse.Log;
import codetrail.eclipse.study.Data;

/**
 * Bookmarks module workbench view.
 * 
 * @author maxg
 */
public class BookmarksView extends ViewPart {
    private TreeViewer viewer;
    private Action go;
    // XXX private Action delete;
    
    public void createPartControl(Composite parent) {
        viewer = new TreeViewer(parent, SWT.MULTI | SWT.H_SCROLL | SWT.V_SCROLL);
        try {
            viewer.setContentProvider(new BookmarksViewContentProvider(viewer));
        } catch (RemoteException re) {
            Log.error("Error creating bookmarks view", re);
        }
        viewer.setLabelProvider(new BookmarksViewLabelProvider());
        viewer.setComparator(new BookmarksViewComparator());
        viewer.setInput(getViewSite());
        makeActions();
        hookContextMenu();
        hookDoubleClickAction();
        //contributeToActionBars();
    }
    
    private void makeActions() {
        go = new Action() {
            public void run() {
                ISelection selection = viewer.getSelection();
                Object obj = ((IStructuredSelection)selection).getFirstElement();
                if (obj instanceof Bookmark) {
                    Bookmark bookmark = (Bookmark)obj;
                    Log.ok("Visiting bookmark " + bookmark.url);
                    Data.ideEvent("bookmark visit", null, bookmark.url.toString());
                    try {
                        PlatformUI.getWorkbench().getBrowserSupport().createBrowser(Activator.PLUGIN_ID).openURL(bookmark.url);
                    } catch (PartInitException pie) {
                        Log.error("Error opening browser", pie);
                    }
                }
            }
        };
        go.setText("Open");
        go.setImageDescriptor(Activator.imageDescriptorFromPlugin(Activator.PLUGIN_ID, "icons/web_bookmark_go.gif"));
        
        // XXX needs to be implemented
        //delete = new Action() {
        //    public void run() {
        //        System.out.println("delete");
        //    }
        //};
        //delete.setText("Delete");
        //delete.setActionDefinitionId("org.eclipse.ui.edit.delete");
        //delete.setImageDescriptor(PlatformUI.getWorkbench().getSharedImages().getImageDescriptor(ISharedImages.IMG_TOOL_DELETE));
    }
    
    private void hookContextMenu() {
        MenuManager menuMgr = new MenuManager("#PopupMenu");
        menuMgr.setRemoveAllWhenShown(true);
        menuMgr.addMenuListener(new IMenuListener() {
            public void menuAboutToShow(IMenuManager manager) {
                BookmarksView.this.fillContextMenu(manager);
            }
        });
        Menu menu = menuMgr.createContextMenu(viewer.getControl());
        viewer.getControl().setMenu(menu);
        getSite().registerContextMenu(menuMgr, viewer);
    }
    
    private void fillContextMenu(IMenuManager manager) {
        if (((IStructuredSelection)viewer.getSelection()).getFirstElement() instanceof Bookmark) {
            manager.add(go);
            // XXX manager.add(delete);
        }
        manager.add(new Separator());
        // Other plug-ins can contribute there actions here
        manager.add(new Separator(IWorkbenchActionConstants.MB_ADDITIONS));
    }
    
    /*
    private void contributeToActionBars() {
        IActionBars bars = getViewSite().getActionBars();
        fillLocalPullDown(bars.getMenuManager());
        fillLocalToolBar(bars.getToolBarManager());
    }
    
    private void fillLocalPullDown(IMenuManager manager) {
        manager.add(action1);
        manager.add(new Separator());
        manager.add(action2);
    }
    
    private void fillLocalToolBar(IToolBarManager manager) {
        manager.add(action1);
        manager.add(action2);
        manager.add(new Separator());
        drillDownAdapter.addNavigationActions(manager);
    }
    */
    
    private void hookDoubleClickAction() {
        viewer.addDoubleClickListener(new IDoubleClickListener() {
            public void doubleClick(DoubleClickEvent event) {
                go.run();
            }
        });
    }
    
    public void setFocus() {
        viewer.getControl().setFocus();
    }
}

class Bookmark {
    final IResource container;
    final String title;
    final URL url;
    
    Bookmark(IResource container, String title, String url) throws MalformedURLException {
        this.container = container;
        this.title = title;
        this.url = new URL(url);
    }
    
    @Override public String toString() {
        return title;
    }
}

class BookmarksViewLabelProvider extends LabelProvider {
    
    private final ILabelProvider delegate = WorkbenchLabelProvider.getDecoratingWorkbenchLabelProvider();

    @Override public String getText(Object obj) {
        if (obj instanceof IResource) {
            return ((IResource)obj).getName();
        }
        return obj.toString();
    }
    
    @Override public Image getImage(Object obj) {
        if (obj instanceof IResource) {
            return delegate.getImage(obj);
        } else if (obj instanceof Bookmark) {
            // XXX
        }
        return null;
    }
}

class BookmarksViewComparator extends ViewerComparator {

    @Override public int compare(Viewer viewer, Object obj1, Object obj2) {
        if (obj1 instanceof IResource && obj2 instanceof IResource) {
            String path1 = ((IResource)obj1).getProjectRelativePath().toString();
            String path2 = ((IResource)obj2).getProjectRelativePath().toString();
            // order from most to least specific
            return path2.compareTo(path1);
        }
        return super.compare(viewer, obj1, obj2);
    }
}

class BookmarksViewContentProvider implements ITreeContentProvider, GraphSpaceListener {
    
    private String lastEvent = "";
    private final Set<IResource> containers = new HashSet<IResource>();
    private final List<Bookmark> bookmarks = new ArrayList<Bookmark>();
    private final TreeViewer viewer;
    
    BookmarksViewContentProvider(TreeViewer viewer) throws RemoteException {
        this.viewer = viewer;
        GraphSpace space = Activator.activator().space();
        space.register(this,
                       tuple(null, uri("#action"), uri("#view")),
                       Bookmarks.PROPERTIES.getProperty("view-view-query"));
        space.register(this,
                       tuple(null, uri("#action"), uri("#bookmark")),
                       Bookmarks.PROPERTIES.getProperty("view-bookmark-query"));
    }

    public void inputChanged(Viewer v, Object oldInput, Object newInput) { }
    
    public void dispose() { }
    
    public Object[] getElements(Object parent) {
        return containers.toArray(new Object[0]);
    }
    
    public Object getParent(Object child) {
        if (child instanceof Bookmark) {
            return ((Bookmark)child).container;
        }
        return null;
    }
    
    public Object[] getChildren(Object parent) {
        if (parent instanceof IResource) {
            return getBookmarks((IResource)parent);
        }
        return new Object[0];
    }
    
    public boolean hasChildren(Object parent) {
        return parent instanceof IResource;
    }
        
    private Object[] getBookmarks(IResource container) {
        List<Bookmark> filtered = new ArrayList<Bookmark>();
        for (Bookmark bookmark : bookmarks) {
            if (container.equals(bookmark.container)) {
                filtered.add(bookmark);
            }
        }
        return filtered.toArray(new Object[0]);
    }
    
    public void notify(Solution solution) throws Exception {
        String event = solution.get("event");
        if ( ! lastEvent.equals(event)) {
            containers.clear();
            bookmarks.clear();
            lastEvent = event;
        }
        
        String url = solution.get("url");
        if (url != null) {
            String bookmarkpath = solution.get("bookmarkpath");
            if (solution.get("resourcepath").startsWith(bookmarkpath)) {
                IResource container = ResourcesPlugin.getWorkspace().getRoot().getProject(solution.get("project")).findMember(bookmarkpath);
                containers.add(container);
                bookmarks.add(new Bookmark(container, solution.get("title"), url));
            }
        }
        
        viewer.getControl().getDisplay().asyncExec(new Runnable() {
            public void run() {
                viewer.refresh(false);
                viewer.expandAll();
            }
        });
    }
}
