/*
 * Codetrail
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package codetrail.eclipse.mod;

import static codetrail.Tuple.*;
import static codetrail.URIString.rdf_type;
import static codetrail.URIString.uri;
import static codetrail.Util.*;

import java.io.*;
import java.net.URL;
import java.net.URLEncoder;
import java.rmi.RemoteException;
import java.util.*;
import java.util.concurrent.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPathExpressionException;

import org.eclipse.core.resources.*;
import org.eclipse.core.runtime.*;
import org.eclipse.search.core.text.*;
import org.w3c.dom.*;
import org.xml.sax.SAXException;

import codetrail.*;
import codetrail.eclipse.Log;
import codetrail.eclipse.ResourceWatcher;
import codetrail.eclipse.study.Data;

/**
 * Documentation inference module: learns sources of documentation.
 * 
 * @author maxg
 */
public class DocGuesser extends ResourceWatcher implements Module, GraphSpaceListener, Runnable {
    
    static final Properties PROPERTIES = Util.propertiesFor(DocGuesser.class);
    
    static final URIString MOD_DOC_GUESSER = uri("#" + PROPERTIES.getProperty("id"));
    static final URIString IGNORE = uri("#ignore");
    static final String SPLITTER = PROPERTIES.getProperty("splitter-regex");
    static final String TOKEN = PROPERTIES.getProperty("token-placeholder");
    
    static final int MIN_MATCH_COMPONENTS = 4;
    static final int MIN_MATCH_SITES = 2;
    
    static final int SEARCH_START = Integer.parseInt(PROPERTIES.getProperty("search-api-start"));
    static final int SEARCH_COUNT = Integer.parseInt(PROPERTIES.getProperty("search-api-count"));
    static final int SEARCH_MAX = Integer.parseInt(PROPERTIES.getProperty("search-api-max"));
    static final int WORK_UNITS = 10;
    
    static final List<String> STOPWORDS = commaSeparate(PROPERTIES.getProperty("inurl-stopwords"));
    
    final Set<DocGuesserSearcher> workspaceSearchers;
    final Set<DocGuesserJob> workspaceAttachers;
    
    private final Trie<String, Solution> sites;
    private final BlockingQueue<Solution> urls;
    
    private GraphSpace space;
    private Thread guesser;
    
    public DocGuesser() {
        super(PROPERTIES.getProperty("extension"), "webdocs"); // XXX configurable
        workspaceSearchers = Collections.synchronizedSet(new HashSet<DocGuesserSearcher>());
        workspaceAttachers = Collections.synchronizedSet(new HashSet<DocGuesserJob>());
        sites = new Trie<String, Solution>();
        urls = new LinkedBlockingQueue<Solution>();
    }
    
    public String getName() {
        return PROPERTIES.getProperty("name");
    }

    public void connect(GraphSpace space) throws RemoteException {
        if (space.equals(this.space)) return;
        
        guesser = new Thread(this, getClass().getName() + " guess");
        guesser.setDaemon(true);
        guesser.setPriority(3);
        guesser.start();
        
        this.space = space;
        space.register(this,
                       tuple(null, uri("#action"), uri("#browse")),
                       PROPERTIES.getProperty("graphspace-query"));
        
        space.update(PROPERTIES.getProperty("clear-all-statement"));
        
        super.connect();
    }

    public void disconnect() throws RemoteException {
        if (space == null) return;
        
        guesser.interrupt();
        guesser = null;
        
        space.unregister(this);
        super.disconnect();
        space = null;
    }
    
    public void notify(Solution solution) {
        urls.offer(solution);
    }
    
    static List<String> components(String url) {
        url = url.replaceFirst("\\w+:/?/?", "");
        String[] hostPath = url.split("/", 2);
        String[] path = hostPath[1].split(SPLITTER);
        List<String> components = new ArrayList<String>();
        components.add(hostPath[0]);
        components.addAll(Arrays.asList(path));
        return components;
    }
    
    static Map<String, List<String>> sites(List<String> current, Trie<String, Solution> previous) {
        SortedMap<List<String>, Solution> nearby = previous.near(current, MIN_MATCH_COMPONENTS);
        if (nearby.size() < MIN_MATCH_SITES) return Collections.emptyMap();
        
        Map<String, List<String>> sites = new HashMap<String, List<String>>();
        int idx = -1;
        Iterator<List<String>> it = nearby.keySet().iterator();
        for (int i = 0; i < MIN_MATCH_SITES; i++) {
            List<String> key = it.next();
            if (current.size() != key.size()) return Collections.emptyMap(); // closest match not equal length
            for (int j = 0; j < current.size(); j++) {
                if ( ! current.get(j).equals(key.get(j))) {
                    if (idx < 0) {
                        idx = j;
                    } else if (j != idx) {
                        return Collections.emptyMap(); // mismatch in a different component
                    }
                    sites.put(key.get(j), key);
                }
            }
        }
        extra: for (int i = MIN_MATCH_SITES; i < nearby.size(); i++) {
            List<String> key = it.next();
            if (current.size() != key.size()) break;
            for (int j = 0; j < current.size(); j++) {
                if ( ! (current.get(j).equals(key.get(j)) || j == idx)) continue extra;
            }
            sites.put(key.get(idx), key);
        }
        sites.put(current.get(idx), current);
        return sites;
    }
    
    static SortedSet<String> tokens(String template, Map<String, Solution> matched, IProgressMonitor monitor) throws CancellationException, ParserConfigurationException, SAXException, IOException, XPathExpressionException {
        List<String> components = components(template.replace(TOKEN, ""));
        StringBuilder query = new StringBuilder();
        query.append("site:").append(components.get(0));
        for (int i = 1; i < components.size(); i++) { // XXX ignore last component
            String component = components.get(i);
            if (component.length() > 1 && ! STOPWORDS.contains(component)) {
                query.append(" inurl:").append(component);
            }
        }
        Set<String> titleWords = new HashSet<String>();
        if (matched.size() > 1) {
            Iterator<String> it = matched.keySet().iterator();
            for (String word : matched.get(it.next()).get("title").split("\\s+")) {
                if (word.matches("\\W+")) continue;
                titleWords.add(word);
            }
            while (it.hasNext()) {
                titleWords.retainAll(Arrays.asList(matched.get(it.next()).get("title").split("\\s+")));
            }
        }
        for (String word : titleWords) {
            query.append(" ").append(word); // XXX intitle operator doesn't seem to work
        }
        Log.ok("Guess docs on " + template + ": query " + query);
        
        Pattern regex = Pattern.compile("\\Q" + template.replace(TOKEN, "\\E(.*)\\Q") + "\\E");
        Matcher match;
        
        String call = PROPERTIES.getProperty("search-api-url");
        call = call.replace("@ID@", PROPERTIES.getProperty("search-api-id"));
        call = call.replace("@QUERY@", URLEncoder.encode(query.toString(), "UTF-8"));
        call = call.replace("@COUNT@", Integer.toString(SEARCH_COUNT));
        
        SortedSet<String> urls = new TreeSet<String>();
        DocumentBuilder builder = DOC_BUILDERER.newDocumentBuilder();
        for (int start = SEARCH_START; start + SEARCH_COUNT - 1 <= SEARCH_MAX; start += SEARCH_COUNT) {
            monitor.worked(WORK_UNITS);
            if (monitor.isCanceled()) throw new CancellationException();
            URL search = new URL(call.replace("@START@", Integer.toString(start)));
            Document doc;
            try {
                doc = builder.parse(search.openStream());
            } catch (IOException ioe) {
                Log.warn("Guess docs: retrying query", ioe);
                doc = builder.parse(search.openStream());
            }
            NodeList nodes = evaluate(doc, "//Result");
            iterating: for (Element result : iterable(ELEMENT, nodes)) {
                String title = result.getElementsByTagName("Title").item(0).getTextContent().trim();
                String url = result.getElementsByTagName("Url").item(0).getTextContent().trim();
                for (String word : titleWords) {
                    if ( ! title.contains(word)) continue iterating;
                }
                if ((match = regex.matcher(url)).matches()) {
                    urls.add(match.group(1));
                }
            }
            if (nodes.getLength() < SEARCH_COUNT) break;
        }
        return urls;
    }
    
    void ignore(String template, String replace) throws RemoteException {
        String pattern = Pattern.quote(template).replace(replace, "\\E.*\\Q");
        Log.ok("Guess docs: ignoring " + pattern);
        space.add(
                tuple(MOD_DOC_GUESSER,
                      rdf_type,
                      uri("#module")),
                tuple(MOD_DOC_GUESSER,
                      IGNORE,
                      pattern)
        );
    }
    
    void documentation(String template, SortedSet<String> tokens, IProject project) throws Exception {
        tokens = new TreeSet<String>(tokens);
        
        IFile file = fileIn(project, template.replaceFirst("\\w+://", "").replaceAll("[^\\w ,\\._-]", ""));
        Document doc;
        if ( ! file.exists()) {
            file.create(null, IResource.NONE, null);
            doc = DOC_BUILDERER.newDocumentBuilder().newDocument();
        } else {
            doc = DOC_BUILDERER.newDocumentBuilder().parse(file.getContents());
        }
        
        Element webdoc = ensureChild(doc, "webdoc");
        webdoc.setAttribute("version", "0.1");
        
        Element dict = ensureChild(doc, webdoc, "dict");
        
        Element templateString = ensureKeyValue(doc, dict, "template", "string");
        if (templateString.getTextContent().equals("")) {
            templateString.setTextContent(template);
        } else if ( ! templateString.getTextContent().equals(template)) {
            Log.warn("Stored template " + templateString.getTextContent() + " doesn't match " + template);
            return;
        }
        
        Element replaceString = ensureKeyValue(doc, dict, "replace", "string");
        if (replaceString.getTextContent().equals("")) {
            replaceString.setTextContent(TOKEN);
        }
        
        Element array = ensureKeyValue(doc, dict, "tokens", "array");
        
        for (Node tokenNode : iterable(NODE, array.getChildNodes())) {
            tokens.remove(tokenNode.getTextContent());
        }
        
        Element child;
        for (String token : tokens) {
            array.appendChild(child = doc.createElement("string"));
            child.setTextContent(token);
        }
        
        java.io.ByteArrayOutputStream out = new ByteArrayOutputStream();
        transform(doc, out);
        file.setContents(new ByteArrayInputStream(out.toByteArray()), IResource.NONE, null);
    }
    
    void documentation(String template, SortedSet<String> tokens, Set<IProject> projects, IProgressMonitor monitor, int work) throws Exception {
        for (IProject project : projects) {
            documentation(template, tokens, project);
            monitor.worked(work / projects.size());
            if (monitor.isCanceled()) throw new CancellationException();
        }
    }
    
    @Override protected void initialize(IProject project, Set<IFile> webdocs) throws Exception {
        for (IFile webdoc : webdocs) {
            added(webdoc);
        }
        summarize(project, webdocs);
    }
    
    @Override protected void added(IFile webdoc) throws Exception {
        List<Tuple> tuples = new ArrayList<Tuple>();
        
        Document doc;
        try {
            doc = DOC_BUILDERER.newDocumentBuilder().parse(webdoc.getContents());
        } catch (CoreException ce) {
            return;
        }
        Element dict = firstElementByTagName(firstElementByTagName(doc, "webdoc"), "dict");
        
        String template = retrieveKeyValue(dict, "template").getTextContent();
        String replace = retrieveKeyValue(dict, "replace").getTextContent();
        
        ignore(template, replace);
        
        // XXX documentation source needs to be used only for the project it's in
        Tuple docs = unique("#documentation-source");
        tuples.addAll(Arrays.asList(
                docs,
                tuple(docs.subject, uri("#built-from"), webdoc.getFullPath().toPortableString()),
                tuple(docs.subject, uri("#template"), template),
                tuple(docs.subject, uri("#replace"), replace)
        ));
        
        Element array = retrieveKeyValue(dict, "tokens");
        for (Node token : iterable(NODE, array.getChildNodes())) {
            if (token.getNodeName().equals("string")) {
                tuples.add(tuple(docs.subject, uri("#token"), token.getTextContent().toLowerCase())); // XXX to lower case
            }
        }
        
        space.add(tuples);
    }
    
    @Override protected void changed(IFile webdoc) throws Exception {
        removed(webdoc);
        added(webdoc);
    }
    
    @Override protected void removed(IFile webdoc) throws Exception {
        space.update(PROPERTIES.getProperty("clear-statement").replace("@FILENAME@", webdoc.getFullPath().toPortableString()));
    }
    
    @Override protected void summarize(IProject project, Set<IFile> webdocs) throws Exception {
        // XXX this method is for study logging purposes only
        String tag = project.getName().hashCode() + "-" + System.currentTimeMillis() + "-" + (int)(Math.random()*100);
        Data.ideEvent("guessdocs accounting", tag);
        for (IFile webdoc : webdocs) {
            try {
                Document doc = DOC_BUILDERER.newDocumentBuilder().parse(webdoc.getContents());
                Element dict = firstElementByTagName(firstElementByTagName(doc, "webdoc"), "dict");
                String template = retrieveKeyValue(dict, "template").getTextContent();
                Data.ideEvent("guessdocs exists", tag, template);
            } catch (CoreException ce) {
                continue;
            }
        }
    }
    
    public void run() {
        while ( ! Thread.currentThread().isInterrupted()) {
            try {
                Solution solution = urls.take();
                List<String> components = components(solution.get("url"));
                sites.put(components, solution);

                Map<String, List<String>> matching = sites(components, sites);
                if (matching.isEmpty()) continue;
                
                Map.Entry<String, List<String>> site = matching.entrySet().iterator().next();
                String template = sites.get(site.getValue()).get("url").replace(site.getKey(), TOKEN);
                
                Log.ok("Guess docs on " + template + ": saw " + matching.keySet());
                Data.ideEvent("guessdocs search");
                DocGuesserSearcher searcher = new DocGuesserSearcher(sites.get(components), template, matching);
                searcher.search(".*", ResourcesPlugin.getWorkspace().getRoot());
            } catch (InterruptedException ie) {
                return;
            } catch (Exception e) {
                Log.error("Error guessing docs", e);
            }
        }
    }
    
    class DocGuesserSearcher extends TextSearchRequestor {
        
        private final Solution solution;
        private final String template;
        private final Map<String, List<String>> sites;
        private final Pattern regex;
        private final Map<String, List<String>> matched;
        final Set<IProject> projects;
        
        private boolean searching = true;

        public DocGuesserSearcher(Solution solution, String template, Map<String, List<String>> sites) {
            this.solution = solution;
            this.template = template;
            this.sites = sites;
            this.matched = new HashMap<String, List<String>>();
            this.projects = Collections.synchronizedSet(new HashSet<IProject>());
            
            StringBuilder regex = new StringBuilder();
            for (String keyword : sites.keySet()) {
                regex.append("|").append(Pattern.quote(keyword));
            }
            regex.replace(0, 1, "\\b(").append(")\\b");
            this.regex = Pattern.compile(regex.toString(), Pattern.CASE_INSENSITIVE);
        }
        
        void search(String fileNames, IResource... roots) {
            TextSearchScope scope = TextSearchScope.newSearchScope(roots, Pattern.compile(fileNames), false);
            workspaceSearchers.add(this);
            TextSearchEngine.create().search(scope, this, regex, null);
        }

        @Override public boolean acceptFile(IFile file) {
            return true;
        }

        @Override public boolean acceptPatternMatch(TextSearchMatchAccess match) {
            String humanKeyword = match.getFileContent(match.getMatchOffset(), match.getMatchLength());
            for (String machineKeyword : sites.keySet()) {
                if (machineKeyword.equalsIgnoreCase(humanKeyword)) {
                    matched.put(humanKeyword, sites.get(machineKeyword));
                    projects.add(match.getFile().getProject());
                }
            }
            if (searching && matched.size() >= MIN_MATCH_SITES) {
                searching = false;
                
                Log.ok("Guess docs on " + template + ": asking user");
                
                DocGuesserJob job = new DocGuesserJob(template, matched, projects);
                
                Iterator<String> it = matched.keySet().iterator();
                StringBuilder keywordList = new StringBuilder().append("\"").append(it.next()).append("\"");
                while (it.hasNext()) {
                    String keyword = it.next();
                    keywordList.append(it.hasNext() ? ", \"" : " and \"").append(keyword).append("\"");
                }
                String question = PROPERTIES.getProperty("user-question");
                question = question.replace("@KEYWORDS@", keywordList);
                String[] answers = PROPERTIES.getProperty("user-answers").split(",");
                try {
                    space.ask(job, uri(solution.get("source")), question, answers);
                } catch (RemoteException e) {
                    Log.error("Guess docs on " + template + ": failed to ask user");
                }
            }
            return true;
        }
        
        @Override public void endReporting() {
            if (searching) {
                Log.ok("Guess docs on " + template + ": negative, browsed tokens not in code");
                Data.ideEvent("guessdocs negative");
            }
            workspaceSearchers.remove(this);
            synchronized(workspaceSearchers) {
                workspaceSearchers.notify();
            }
        }
    }
    
    class DocGuesserJob extends WorkspaceJob implements GraphSpaceListener {
        
        final String template;
        final Map<String, Solution> matched;
        final Set<IProject> projects;
        
        public DocGuesserJob(String template, Map<String, List<String>> matched, Set<IProject> projects) {
            super(PROPERTIES.getProperty("job-name").replace("@SITE@", components(template).get(0)));
            this.template = template;
            this.matched = new HashMap<String, Solution>();
            for (Map.Entry<String, List<String>> site : matched.entrySet()) {
                this.matched.put(site.getKey(), sites.get(site.getValue()));
            }
            this.projects = projects;
        }
        
        public void notify(Solution solution) throws RemoteException {
            // user answered question about attaching docs
            
            ignore(template, TOKEN); // XXX ignore now since attachment takes a while
            
            if (solution.get("answer").equals(PROPERTIES.getProperty("user-answer-yes"))) {
                workspaceAttachers.add(this);
                this.schedule();
            } else {
                Log.ok("Guess docs on " + template + ": ignore");
                Data.ideEvent("guessdocs ignore");
            }
        }

        public IStatus runInWorkspace(IProgressMonitor monitor) throws CoreException {
            try {
                int halfWork = SEARCH_MAX/SEARCH_COUNT * WORK_UNITS;
                monitor.beginTask(PROPERTIES.getProperty("task-name").replace("@SITE@", components(template).get(0)), 2*halfWork);
                
                SortedSet<String> tokens = tokens(template, matched, monitor);
                
                Log.ok("Guess docs on " + template + ": positive, " + tokens.size() + " tokens");
                Data.ideEvent("guessdocs positive", Integer.toString(tokens.size()), template);
                
                // XXX the searcher might still be searching, so we copy the projects it found so far
                Set<IProject> projects = new HashSet<IProject>();
                synchronized (this.projects) {
                    projects.addAll(this.projects);
                }
                documentation(template, tokens, projects, monitor, halfWork);
                
                monitor.done();
            } catch (CancellationException ce) {
                Log.info("Guess docs on " + template + ": cancelled by user");
                Data.ideEvent("guessdocs", "cancel");
                return Status.CANCEL_STATUS;
            } catch (Exception e) {
                Log.error("Error realizing documentor", e);
            } finally {
                workspaceAttachers.remove(this);
                synchronized (workspaceAttachers) {
                    workspaceAttachers.notify();
                }
            }
            return Status.OK_STATUS;
        }
    }
}
