/*
 * Codetrail
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package codetrail.eclipse.mod;

import static codetrail.Tuple.tuple;
import static codetrail.URIString.uri;

import java.io.IOException;
import java.net.URLEncoder;
import java.util.*;

import org.eclipse.core.runtime.IPath;
import org.eclipse.jdt.core.*;
import org.eclipse.jdt.ui.JavaUI;
import org.eclipse.jface.text.ITextSelection;
import org.eclipse.ui.texteditor.ITextEditor;

import codetrail.*;

/**
 * Element reporter module Java reporter.
 * 
 * @author maxg
 */
public class JavaReporter implements Reporter {
    
    private static final String ECLIPSE_ID = "org.eclipse.platform.doc.isv";

    public List<Tuple> selection(GraphSpace space, URIString event, ITextEditor editor, ITextSelection selection) throws JavaModelException, IOException {
        IJavaElement input = JavaUI.getEditorInputJavaElement(editor.getEditorInput());
        if (input instanceof ICodeAssist) {
            return selection(event, (ICodeAssist)input, selection);
        }
        return null;
    }
    
    private List<Tuple> selection(URIString event, ICodeAssist input, ITextSelection selection) throws JavaModelException, IOException {
        IJavaElement[] elements = input.codeSelect(selection.getOffset(), selection.getLength());
        if (elements.length > 0) {
            return selected(event, elements[0]);
        }
        return null;
    }
    
    private List<Tuple> selected(URIString event, IJavaElement element) throws IOException, JavaModelException {
        String id = element.getHandleIdentifier();
        URIString eltURI = uri("#code-element-java-" + URLEncoder.encode(id, "UTF-8"));
        List<Tuple> tuples = new ArrayList<Tuple>(Arrays.asList(
                tuple(event, uri("#artifact"), eltURI),
                tuple(eltURI, uri("#language"), "java"),
                tuple(eltURI, uri("#java-handle-id"), id)
        ));
        
        String docURL = null;
        String docIndexURL = null;
        switch (element.getElementType()) {
        case IJavaElement.TYPE:
            IType type = (IType)element;
            tuples.add(tuple(eltURI, uri("#kind"), "type"));
            tuples.add(tuple(eltURI, uri("#name"), type.getFullyQualifiedName()));
            docURL = documentationURL(type);
            if (docURL != null) docURL += "#skip-navbar_top";
            docIndexURL = getJavadocBaseLocation(type);
            break;
        case IJavaElement.FIELD:
            IField field = (IField)element;
            tuples.add(tuple(eltURI, uri("#kind"), "field"));
            tuples.add(tuple(eltURI, uri("#class"), field.getDeclaringType().getFullyQualifiedName()));
            tuples.add(tuple(eltURI, uri("#name"), field.getElementName()));
            docURL = documentationURL(field);
            docIndexURL = getJavadocBaseLocation(field.getDeclaringType());
            break;
        case IJavaElement.METHOD:
            IMethod method = (IMethod)element;
            tuples.add(tuple(eltURI, uri("#kind"), "method"));
            tuples.add(tuple(eltURI, uri("#class"), method.getDeclaringType().getFullyQualifiedName()));
            tuples.add(tuple(eltURI, uri("#name"), method.getElementName() + method.getSignature())); // XXX
            docURL = documentationURL(method);
            docIndexURL = getJavadocBaseLocation(method.getDeclaringType());
            break;
        case IJavaElement.LOCAL_VARIABLE:
            tuples.add(tuple(eltURI, uri("#kind"), "local"));
            tuples.add(tuple(eltURI, uri("#type"), ((ILocalVariable)element).getTypeSignature()));
            // XXX documentation link to type
            break;
        }
        
        String docType = "javadoc";
        if (docURL != null && docURL.startsWith("jar:") && docURL.contains(ECLIPSE_ID)) {
            docURL = docURL.replaceFirst("jar:.*!", "http://help.eclipse.org/stable/topic/" + ECLIPSE_ID);
            docIndexURL = docIndexURL.replaceFirst("jar:.*!", "http://help.eclipse.org/stable/index.jsp?topic=/" + ECLIPSE_ID);
            docType = "eclipse";
        }
        if (docURL != null && docURL.startsWith("http")) {
            tuples.add(tuple(eltURI, uri("#documentation"), docURL));
            tuples.add(tuple(eltURI, uri("#documentation-index"), docIndexURL));
            tuples.add(tuple(eltURI, uri("#documentation-type"), docType));
        }
        
        return tuples;
    }
    
    /**
     * @see org.eclipse.jdt.internal.core.BinaryType.getJavadocContents(IProgressMonitor)
     *      internal Eclipse implementation partially duplicated here
     */
    static String documentationURL(IType type) throws JavaModelException {
        String base = getJavadocBaseLocation(type);
        if (base == null) {
            return null;
        }
        StringBuffer pathBuffer = new StringBuffer(base);

        if (!(pathBuffer.charAt(pathBuffer.length() - 1) == '/')) {
            pathBuffer.append('/');
        }
        IPackageFragment pack= type.getPackageFragment();
        String typeQualifiedName = null;
        if (type.isMember()) {
            IType currentType = type;
            StringBuffer typeName = new StringBuffer();
            while (currentType != null) {
                typeName.insert(0, currentType.getElementName());
                currentType = currentType.getDeclaringType();
                if (currentType != null) {
                    typeName.insert(0, '.');
                }
            }
            typeQualifiedName = typeName.toString();
        } else {
            typeQualifiedName = type.getElementName();
        }
        
        pathBuffer.append(pack.getElementName().replace('.', '/')).append('/').append(typeQualifiedName).append(".html");
        return pathBuffer.toString();
    }
    
    static String documentationURL(IField field) throws JavaModelException {
        String typeDoc = documentationURL(field.getDeclaringType());
        if (typeDoc == null) return null;
        
        return typeDoc + "#" + field.getElementName();
    }
    
    static String documentationURL(IMethod method) throws JavaModelException {
        String typeDoc = documentationURL(method.getDeclaringType());
        if (typeDoc == null) return null;
        
        /*
         * XXX known issues
         *  method.getElementName() correct for all c'tors?
         *  no reason to suspect Signature.toString is always the right thing
         *  appending ourselves rather than using one of the other
         *    Signature.toString methods because generic args appear as the
         *    type variable name, not java.lang.Object
         */
        StringBuilder anchorBuild = new StringBuilder();
        for (String type : method.getParameterTypes()) {
            anchorBuild.append(", ").append(Signature.toString(type));
        }
        String anchor = anchorBuild.replace(0, 2, method.getElementName() + "(").append(")").toString();
        anchor = anchor.replaceAll("<.*>", "");
        if (Flags.isVarargs(method.getFlags())) {
            anchor = anchor.replace("[])", "...)");
        }
        return typeDoc + "#" + anchor;
    }
    
    /**
     * @see org.eclipse.jdt.internal.core.JavaElement.getJavadocBaseLocation()
     *      internal Eclipse implementation duplicated here
     */
    static String getJavadocBaseLocation(IType type) throws JavaModelException {
        IPackageFragmentRoot root= (IPackageFragmentRoot) type.getAncestor(IJavaElement.PACKAGE_FRAGMENT_ROOT);
        if (root == null) {
            return null;
        }

        if (root.getKind() == IPackageFragmentRoot.K_BINARY) {
            IClasspathEntry entry= root.getRawClasspathEntry();
            if (entry == null) {
                return null;
            }
            if (entry.getEntryKind() == IClasspathEntry.CPE_CONTAINER) {
                entry= getRealClasspathEntry(root.getJavaProject(), entry.getPath(), root.getPath());
                if (entry == null) {
                    return null;
                }
            }
            return getLibraryJavadocLocation(entry);
        }
        return null;
    }
    
    /**
     * @see org.eclipse.jdt.internal.core.JavaElement.getRealClasspathEntry(IJavaProject, IPath, IPath)
     *      internal Eclipse implementation duplicated here
     */
    static IClasspathEntry getRealClasspathEntry(IJavaProject project, IPath containerPath, IPath libPath) throws JavaModelException {
        IClasspathContainer container= JavaCore.getClasspathContainer(containerPath, project);
        if (container != null) {
            IClasspathEntry[] entries= container.getClasspathEntries();
            for (int i= 0; i < entries.length; i++) {
                IClasspathEntry curr = entries[i];
                if (curr == null) {
                    break;
                }
                IClasspathEntry resolved= JavaCore.getResolvedClasspathEntry(curr);
                if (resolved != null && libPath.equals(resolved.getPath())) {
                    return curr; // return the real entry
                }
            }
        }
        return null; // not found
    }
    
    /**
     * @see org.eclipse.jdt.internal.core.JavaElement.getLibraryJavadocLocation(IClasspathEntry)
     *      internal Eclipse implementation duplicated here
     */
    static String getLibraryJavadocLocation(IClasspathEntry entry) {
        switch(entry.getEntryKind()) {
            case IClasspathEntry.CPE_LIBRARY :
            case IClasspathEntry.CPE_VARIABLE :
                break;
            default :
                return null;
        }
        
        IClasspathAttribute[] extraAttributes= entry.getExtraAttributes();
        for (int i= 0; i < extraAttributes.length; i++) {
            IClasspathAttribute attrib= extraAttributes[i];
            if (IClasspathAttribute.JAVADOC_LOCATION_ATTRIBUTE_NAME.equals(attrib.getName())) {
                String value = attrib.getValue();
                return value;
            }
        }
        return null;
    }
}
