/*
 * Codetrail
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package codetrail.eclipse.mod;

import static codetrail.Tuple.tuple;
import static codetrail.URIString.rdf_type;
import static codetrail.URIString.uri;
import static org.eclipse.jdt.core.IClasspathAttribute.JAVADOC_LOCATION_ATTRIBUTE_NAME;

import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.rmi.RemoteException;
import java.util.*;

import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.resources.WorkspaceJob;
import org.eclipse.core.runtime.*;
import org.eclipse.jdt.core.*;

import codetrail.*;
import codetrail.eclipse.Log;
import codetrail.eclipse.study.Data;

/**
 * Javadoc attacher module: attaches browser-visited Javadoc sites. Attached
 * to relevant libraries as Javadoc locations.
 * 
 * @author maxg
 */
public class JavadocAttacher implements Module, GraphSpaceListener {
    
    static final Properties PROPERTIES = Util.propertiesFor(JavadocAttacher.class);
    
    static final URIString MOD_JAVADOC_ATTACHER = uri("#" + PROPERTIES.getProperty("id"));
    static final URIString IGNORE = uri("#ignore");
        
    private GraphSpace space;
    
    public String getName() {
        return PROPERTIES.getProperty("name");
    }

    public void connect(GraphSpace space) throws RemoteException {
        if (space.equals(this.space)) return;
        
        this.space = space;
        space.register(this,
                       tuple(null, uri("#action"), uri("#load")),
                       PROPERTIES.getProperty("graphspace-query"));
    }
    
    public void disconnect() throws RemoteException {
        if (space == null) return;
        
        space.unregister(this);
        space = null;
    }
    
    public void notify(Solution solution) throws IOException, JavaModelException {
        // user browsed to possible Javadoc location
        
        URL visited = new URL(solution.get("url"));
        URL pkgList = new URL(visited, "package-list");
        BufferedReader reader = new BufferedReader(new InputStreamReader(pkgList.openStream()));
        List<String> pkgs = new ArrayList<String>();
        String pkg;
        while ((pkg = reader.readLine()) != null) {
            pkgs.add(pkg);
        }
        reader.close();
        
        IJavaModel model = JavaCore.create(ResourcesPlugin.getWorkspace().getRoot());
        
        Map<IJavaProject, IClasspathEntry> matches = new HashMap<IJavaProject, IClasspathEntry>();
        IClasspathEntry entry;
        for (IJavaProject project : model.getJavaProjects()) {
            if (project.isOpen()) {
                entry = match(project, pkgs);
                if (entry != null) {
                    matches.put(project, entry);
                }
            }
        }
        
        if (matches.isEmpty()) return;
        
        MatchAttacher attacher = new MatchAttacher(visited, matches);
        String question = PROPERTIES.getProperty("user-question");
        question = question.replace("@LIBRARY@", Util.commaSeparated(attacher.libraries()));
        question = question.replace("@PROJECT@", Util.commaSeparated(attacher.projects()));
        String[] answers = PROPERTIES.getProperty("user-answers").split(",");
        space.ask(attacher, uri(solution.get("source")), question, answers);
        Log.ok("Asking about Javadoc " + visited);
        Data.ideEvent("javadocs ask");
    }
    
    private IClasspathEntry match(IJavaProject project, List<String> pkgs) throws JavaModelException {
        IClasspathEntry candidate = null;
        for (String pkg : pkgs) {
            IJavaElement element = project.findElement(new Path(pkg.replace(".", "/")));
            if (element == null) continue;
            if (candidate != null) {
                if (element.getPath().equals(candidate.getPath())) continue;
                else return null; // XXX maybe fail less fast?
            }
            for (IClasspathEntry entry : project.getRawClasspath()) {
                if (element.getPath().equals(entry.getPath())) {
                    candidate = entry;
                    break;
                }
            }
        }
        if (candidate == null) return null;
        for (IClasspathAttribute attr : candidate.getExtraAttributes()) {
            if (attr.getName().equals(JAVADOC_LOCATION_ATTRIBUTE_NAME))
                return null;
        }
        return candidate;
    }
    
    class MatchAttacher implements GraphSpaceListener {
        private final URL visited;
        private final Map<IJavaProject, IClasspathEntry> matches;
        
        MatchAttacher(URL visited, Map<IJavaProject, IClasspathEntry> matches) {
            this.visited = visited;
            this.matches = matches;
        }
        
        Collection<String> projects() {
            Set<String> projects = new HashSet<String>();
            for (IJavaProject project : matches.keySet()) {
                projects.add(project.getElementName());
            }
            return projects;
        }
        
        Collection<String> libraries() {
            Set<String> libraries = new HashSet<String>();
            for (IClasspathEntry entry : matches.values()) {
                libraries.add(entry.getPath().lastSegment());
            }
            return libraries;
        }

        public void notify(Solution solution) throws JavaModelException, MalformedURLException, RemoteException {
            // user answered question about attaching Javadoc
            
            if (solution.get("answer").equals(PROPERTIES.getProperty("user-answer-yes"))) {
                Log.ok("Javadoc attach " + visited);
                Data.ideEvent("javadocs attach", null, visited.toString());
                for (Map.Entry<IJavaProject, IClasspathEntry> entry : matches.entrySet()) {
                    attach(entry.getKey(), entry.getValue(), new URL(visited, "./"));
                }
            } else {
                Log.ok("Javadoc ignore " + visited);
                Data.ideEvent("javadocs ignore");
                space.add(
                        tuple(MOD_JAVADOC_ATTACHER,
                              rdf_type,
                              uri("#module")),
                        tuple(MOD_JAVADOC_ATTACHER,
                              IGNORE,
                              visited.toExternalForm())
                );
            }
        }
        
        private void attach(final IJavaProject project, IClasspathEntry entry, URL location) throws JavaModelException {
            if (entry.getEntryKind() != IClasspathEntry.CPE_LIBRARY) return;
            List<IClasspathAttribute> extraAttrs = new ArrayList<IClasspathAttribute>();
            extraAttrs.addAll(Arrays.asList(entry.getExtraAttributes()));
            extraAttrs.add(JavaCore.newClasspathAttribute(JAVADOC_LOCATION_ATTRIBUTE_NAME,
                                                          location.toExternalForm()));
            final IClasspathEntry[] entries = project.getRawClasspath();
            for (int ii = 0; ii < entries.length; ii++) {
                if (entries[ii].equals(entry)) {
                    entries[ii] = JavaCore.newLibraryEntry(entry.getPath(),
                                                           entry.getSourceAttachmentPath(),
                                                           entry.getSourceAttachmentRootPath(),
                                                           entry.getAccessRules(),
                                                           extraAttrs.toArray(new IClasspathAttribute[0]),
                                                           entry.isExported());
                    break;
                }
            }
            
            new WorkspaceJob(JavadocAttacher.PROPERTIES.getProperty("job-name")) {
                @Override public IStatus runInWorkspace(IProgressMonitor monitor) throws JavaModelException {
                    project.setRawClasspath(entries, monitor);
                    return Status.OK_STATUS;
                }
            }.schedule();
        }
    }
}
