/*
 * Codetrail
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package codetrail.eclipse.mod;

import static codetrail.Tuple.*;
import static codetrail.URIString.uri;

import java.io.IOException;
import java.io.InputStream;
import java.rmi.RemoteException;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

import org.eclipse.compare.rangedifferencer.*;
import org.eclipse.core.resources.*;
import org.eclipse.core.runtime.CoreException;

import codetrail.*;
import codetrail.eclipse.Log;

/**
 * Edit reporter module: reports changes made in the code editor.
 * 
 * @author maxg
 */
public class ReportEdits implements Module, IResourceChangeListener, IResourceDeltaVisitor, Runnable {
    
    private final BlockingQueue<IFile> changed;
    
    private GraphSpace space;
    private Thread differ;
    
    public ReportEdits() {
        changed = new LinkedBlockingQueue<IFile>();
    }
    
    public String getName() {
        return "Code edit tracker";
    }

    public void connect(GraphSpace space) {
        if (space.equals(this.space)) return;
        
        this.space = space;
        
        differ = new Thread(this, getClass().getName() + " differ");
        differ.setDaemon(true);
        differ.setPriority(3);
        differ.start();
        
        ResourcesPlugin.getWorkspace().addResourceChangeListener(this, IResourceChangeEvent.POST_CHANGE);
    }
    
    public void disconnect() {
        ResourcesPlugin.getWorkspace().removeResourceChangeListener(this);
        differ.interrupt();
        differ = null;
        space = null;
    }

    public void resourceChanged(IResourceChangeEvent event) {
        try {
            event.getDelta().accept(this);
        } catch (CoreException ce) {
            ce.printStackTrace(); // XXX
        }
    }
    
    public boolean visit(IResourceDelta delta) throws CoreException {
        if (delta.getKind() != IResourceDelta.CHANGED || delta.getResource().isDerived()) return false;
        if ((delta.getFlags() & IResourceDelta.CONTENT) != 0 && delta.getResource() instanceof IFile) {
            changed.offer((IFile)delta.getResource());
        }
        return true;
    }
    
    public void run() {
        while ( ! Thread.currentThread().isInterrupted()) {
            try {
                IFile file = changed.take();
                IFileState[] history = file.getHistory(null);
                if (history.length < 1) continue;
                
                TextComparator now = new TextComparator(file);
                TextComparator before = new TextComparator(history[0]);
                
                RangeDifference[] diffs = RangeDifferencer.findDifferences(now, before);
                for (RangeDifference diff : diffs) {
                    StringBuilder code = new StringBuilder();
                    for (int ii = diff.leftStart(); ii < diff.leftEnd(); ii++) {
                        code.append(now.lines[ii]).append("\n");
                    }
                    edit(file, code.toString());
                }
            } catch (InterruptedException ie) {
                return;
            } catch (Exception e) {
                Log.error("Error reporting edit", e);
            }
        }
    }
    
    private void edit(IFile file, String code) throws RemoteException {
        if (code.equals("")) return;
        Tuple edit = unique("#edit");
        space.add(
                edit,
                timestamp(edit.subject),
                tuple(edit.subject, uri("#source"), uri("#eclipse-XXX")), // XXX
                tuple(edit.subject, uri("#project"), file.getProject().getName()),
                tuple(edit.subject, uri("#path"), file.getProjectRelativePath().toPortableString()),
                tuple(edit.subject, uri("#new-code"), code)
        );
    }
}

class TextComparator implements IRangeComparator {
    
    final String[] lines;

    public TextComparator(IStorage storage) throws CoreException, IOException {
        InputStream is = storage.getContents();
        byte[] bytes = new byte[is.available()];
        is.read(bytes);
        is.close();
        String content = new String(bytes);
        lines = content.split("[\r\n]+");
    }

    public int getRangeCount() {
        return lines.length;
    }

    public boolean rangesEqual(int thisIndex, IRangeComparator other, int otherIndex) {
        return lines[thisIndex].equals(((TextComparator)other).lines[otherIndex]);
    }

    public boolean skipRangeComparison(int length, int maxLength, IRangeComparator other) {
        return false;
    }
}