/*
 * Codetrail
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package codetrail.eclipse.mod;

import static codetrail.Tuple.*;
import static codetrail.URIString.uri;

import java.rmi.RemoteException;
import java.util.*;

import org.eclipse.jface.text.ITextSelection;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.*;
import org.eclipse.ui.part.MultiPageEditorPart;
import org.eclipse.ui.texteditor.ITextEditor;

import codetrail.*;
import codetrail.eclipse.Log;

/**
 * Element reporter module: reports known elements under the code editor cursor.
 * 
 * @author maxg
 */
public class ReportElements implements Module, IWindowListener, IPageListener, ISelectionListener {
    
    static final Properties PROPERTIES = Util.propertiesFor(ReportElements.class);
    
    private static final Map<String, Reporter> REPORTERS;
    static {
        REPORTERS = new HashMap<String, Reporter>();
        REPORTERS.put("org.eclipse.jdt.ui.CompilationUnitEditor", new JavaReporter());
        REPORTERS.put("org.eclipse.wst.javascript.core.javascriptsource.source", new JavaScriptReporter());
    }
    private static final Reporter GENERIC_REPORTER = new GenericReporter();
    
    private GraphSpace space;
    private Display display;
    
    public String getName() {
        return "Code element tracker";
    }

    public void connect(GraphSpace space) throws RemoteException {
        if (space.equals(this.space)) return;
        
        this.space = space;
        this.display = PlatformUI.getWorkbench().getDisplay();
        PlatformUI.getWorkbench().addWindowListener(this);
        for (IWorkbenchWindow window : PlatformUI.getWorkbench().getWorkbenchWindows()) {
            windowOpened(window);
        }
    }
    
    public void disconnect() throws RemoteException {
        PlatformUI.getWorkbench().removeWindowListener(this);
        space = null;
    }
    
    public void windowActivated(IWorkbenchWindow window) { }
    public void windowClosed(IWorkbenchWindow window) { }
    public void windowDeactivated(IWorkbenchWindow window) { }
    public void windowOpened(IWorkbenchWindow window) {
        window.addPageListener(this);
        for (IWorkbenchPage page : window.getPages()) {
            pageOpened(page);
        }
    }
    
    public void pageActivated(IWorkbenchPage page) { }
    public void pageClosed(IWorkbenchPage page) { }
    public void pageOpened(final IWorkbenchPage page) {
        page.addPostSelectionListener(this);
        display.asyncExec(new Runnable() {
            public void run() {
                selectionChanged(page.getActivePart(), page.getSelection());
            }
        });
    }

    public void selectionChanged(IWorkbenchPart part, ISelection selection) {
        if ((part instanceof IEditorPart) && (selection instanceof ITextSelection)) {
            try {
                selectionChanged((IEditorPart)part, (ITextSelection)selection);
            } catch (Exception e) {
                Log.warn("Error reporting selection", e);
            }
        }
    }
    
    private void selectionChanged(IEditorPart editor, ITextSelection selection) throws Exception {
        if (space == null) {
            editor.getEditorSite().getPage().removePostSelectionListener(this);
            editor.getEditorSite().getWorkbenchWindow().removePageListener(this);
            return;
        }
        
        if (editor instanceof MultiPageEditorPart) {
            IEditorPart[] parts = ((MultiPageEditorPart)editor).findEditors(editor.getEditorInput());
            for (IEditorPart part : parts) {
                if (part instanceof ITextEditor) {
                    editor = part; // XXX why the first one?
                    break;
                }
            }
        }
        if ( ! (editor instanceof ITextEditor)) return;
        
        String id = editor.getSite().getId();
        Reporter reporter = REPORTERS.get(id);
        if (reporter == null) reporter = GENERIC_REPORTER;
        
        Tuple event = unique("#event");
        List<Tuple> data = reporter.selection(space, event.subject, (ITextEditor)editor, selection);
        if (data == null) return;
        
        List<Tuple> tuples = new ArrayList<Tuple>(Arrays.asList(
                event,
                tuple(event.subject, uri("#action"), uri("#select")),
                tuple(event.subject, uri("#source"), "#eclipse-XXX"), // XXX
                timestamp(event.subject)
        ));
        tuples.addAll(data);
        space.add(tuples);
    }
}

interface Reporter {
    List<Tuple> selection(GraphSpace space, URIString event, ITextEditor editor, ITextSelection selection) throws Exception;
}