/*
 * Codetrail
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package codetrail.eclipse.mod;

import static codetrail.Tuple.*;
import static codetrail.URIString.rdf_type;
import static codetrail.URIString.uri;

import java.rmi.RemoteException;

import org.eclipse.ui.*;
import org.eclipse.ui.part.FileEditorInput;

import codetrail.*;
import codetrail.eclipse.Log;

/**
 * View reporter module: reports files viewed in the code editor.
 * 
 * @author maxg
 */
public class ReportViews implements Module, IWindowListener, IPageListener, IPartListener {
    
    static final URIString MOD_REPORT_VIEWS = uri("#mod-report-views");
    
    private GraphSpace space;
    private String lastPath = "";
    
    public String getName() {
        return "Code view tracker";
    }

    public void connect(GraphSpace space) {
        if (space.equals(this.space)) return;
        
        this.space = space;
        PlatformUI.getWorkbench().addWindowListener(this);
        for (IWorkbenchWindow window : PlatformUI.getWorkbench().getWorkbenchWindows()) {
            windowOpened(window);
        }
    }
    
    public void disconnect() {
        PlatformUI.getWorkbench().removeWindowListener(this);
        space = null;
    }
    
    public void windowActivated(IWorkbenchWindow window) { }
    public void windowClosed(IWorkbenchWindow window) { }
    public void windowDeactivated(IWorkbenchWindow window) { }
    public void windowOpened(IWorkbenchWindow window) {
        window.addPageListener(this);
        for (IWorkbenchPage page : window.getPages()) {
            pageOpened(page);
        }
    }
    
    public void pageActivated(IWorkbenchPage page) { }
    public void pageClosed(IWorkbenchPage page) { }
    public void pageOpened(final IWorkbenchPage page) {
        page.addPartListener(this);
        IEditorPart editor = page.getActiveEditor();
        if (editor != null) {
            partActivated(editor);
        }
    }

    public void partActivated(IWorkbenchPart part) {
        if (part instanceof IEditorPart) {
            editorActivated((IEditorPart)part);
        }
    }
    public void partBroughtToTop(IWorkbenchPart part) {
        if (part instanceof IEditorPart) {
            editorActivated((IEditorPart)part);
        }
    }
    public void partClosed(IWorkbenchPart part) { }
    public void partDeactivated(IWorkbenchPart part) { }
    public void partOpened(IWorkbenchPart part) { }
    
    private void editorActivated(IEditorPart editor) {
        IEditorInput input = editor.getEditorInput();
        if (input instanceof FileEditorInput) {
            try {
                editorActivated((FileEditorInput)input);
            } catch (RemoteException re) {
                Log.warn("Error reporting editor", re);
            }
        }
    }
    
    private void editorActivated(FileEditorInput input) throws RemoteException {
        String path = input.getFile().getProjectRelativePath().toPortableString();
        if (path.equals(lastPath)) {
            return;
        }
        lastPath = path;
        Tuple event = unique("#event");
        space.add(tuple(MOD_REPORT_VIEWS, rdf_type, uri("#module")));
        space.removeAll(tuple(MOD_REPORT_VIEWS, uri("#currently-viewing"), (URIString)null));
        space.add(
                tuple(MOD_REPORT_VIEWS, uri("#currently-viewing"), event.subject),
                event,
                tuple(event.subject, uri("#action"), uri("#view")),
                tuple(event.subject, uri("#source"), uri("#eclipse-XXX")), // XXX
                timestamp(event.subject),
                tuple(event.subject, uri("#project"), input.getFile().getProject().getName()),
                tuple(event.subject, uri("#path"), path)
        );
    }
}
