/*
 * Codetrail
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package codetrail.eclipse.mod;

import static codetrail.Tuple.tuple;
import static codetrail.URIString.uri;

import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.rmi.RemoteException;
import java.util.*;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

import org.eclipse.core.resources.*;
import org.eclipse.core.runtime.CoreException;

import codetrail.*;
import codetrail.eclipse.Log;

/**
 * Resources module: tracks the current workspace resource tree.
 * 
 * @author maxg
 */
public class Resources implements Module, GraphSpaceListener, IResourceChangeListener, IResourceVisitor, IResourceDeltaVisitor, Runnable {
    
    private final BlockingQueue<MinimalResourceDelta> deltas;
    private final LinkedList<String> parents;
    private final Map<String, List<String>> added;
    private final Map<String, List<String>> removed;
    
    private GraphSpace space;
    private Thread recorder;
    
    public Resources() {
        deltas = new LinkedBlockingQueue<MinimalResourceDelta>();
        parents = new LinkedList<String>();
        added = new HashMap<String, List<String>>();
        removed = new HashMap<String, List<String>>();
    }
    
    public String getName() {
        return "Resource tracker";
    }
    
    public void connect(GraphSpace space) throws RemoteException {
        if (space.equals(this.space)) return;
        
        this.space = space;
        
        recorder = new Thread(this, getClass().getName() + " rec");
        recorder.setDaemon(true);
        recorder.setPriority(3);
        recorder.start();
        
        ResourcesPlugin.getWorkspace().addResourceChangeListener(this, IResourceChangeEvent.POST_CHANGE);
        
        space.register(this,
                       tuple(null, uri("#action"), uri("#resources-list-request")),
                       "SELECT ?x WHERE { }");
    }
    
    public void disconnect() throws RemoteException {
        ResourcesPlugin.getWorkspace().removeResourceChangeListener(this);
        recorder.interrupt();
        recorder = null;
        space = null;
    }
    
    public void notify(Solution solution) {
        try {
            ResourcesPlugin.getWorkspace().getRoot().accept(this);
        } catch (CoreException ce) {
            Log.error("Error visiting workspace", ce);
        }
    }
    
    public void resourceChanged(IResourceChangeEvent event) {
        try {
            event.getDelta().accept(this);
        } catch (CoreException ce) {
            Log.error("Error visiting resource delta", ce);
        }
    }
    
    public boolean visit(IResource resource) throws CoreException {
        return handle(resource, IResourceDelta.ADDED);
    }
    
    public boolean visit(IResourceDelta delta) throws CoreException {
        return handle(delta.getResource(), delta.getKind());
    }
    
    private boolean handle(IResource resource, int kind) {
        if (resource.getParent() == null) return true;
        if (resource.getName().startsWith(".")) return false;
        if (resource.isDerived()) return false;
        
        if (kind == IResourceDelta.CHANGED) return true;
        
        deltas.offer(new MinimalResourceDelta(resource, kind));
        
        return true;
    }
    
    private void handle(String parentName, List<String> added, List<String> removed) throws UnsupportedEncodingException, RemoteException, InterruptedException {
        URIString parent = uri("#r6e-" + URLEncoder.encode(parentName, "UTF-8"));
        
        if (added != null) {
            space.add(tuple(parent, uri("#r6e-add"), Util.separated(added, "\n")));
        }
        
        if (removed != null) {
            space.add(tuple(parent, uri("#r6e-del"), Util.separated(removed, "\n")));
        }
        
        Thread.sleep(10); // XXX try to prevent Eclipse or Firefox from pegging the CPU
    }
    
    public void run() {
        while ( ! Thread.currentThread().isInterrupted()) {
            try {
                MinimalResourceDelta delta = deltas.take();
                
                Map<String, List<String>> children;
                switch (delta.kind) {
                case (IResourceDelta.ADDED):
                    children = added;
                    break;
                case (IResourceDelta.REMOVED):
                    children = removed;
                    break;
                default:
                    continue;
                }
                
                String parent = delta.resource.getParent().getFullPath().toPortableString();
                String child = delta.resource.getName();
                while (( ! parents.isEmpty()) && ( ! parent.startsWith(parents.peek()))) {
                    String parentName = parents.poll();
                    handle(parentName, added.remove(parentName), removed.remove(parentName));
                }
                if ( ! parent.equals(parents.peek())) {
                    parents.addFirst(parent);
                    children.put(parent, new ArrayList<String>());
                }
                children.get(parent).add(child);
                
                if (deltas.isEmpty()) {
                    while ( ! parents.isEmpty()) {
                        String parentName = parents.poll();
                        handle(parentName, added.remove(parentName), removed.remove(parentName));
                    }
                }
            } catch (InterruptedException ie) {
                return;
            } catch (Exception e) {
                Log.error("Error recording resource", e);
            }
        }
    }
}

class MinimalResourceDelta {
    final IResource resource;
    final int kind;
    
    public MinimalResourceDelta(IResource resource, int kind) {
        this.resource = resource;
        this.kind = kind;
    }
}