/*
 * Codetrail
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package codetrail.eclipse.mod;

import static codetrail.Tuple.*;
import static codetrail.URIString.*;

import java.rmi.RemoteException;
import java.util.*;

import codetrail.*;
import codetrail.eclipse.Log;
import codetrail.eclipse.study.Data;

/**
 * Snippets module: identifies copied code.
 * 
 * @author maxg
 */
public class Snippets implements Module {
    
    static final Properties PROPERTIES = Util.propertiesFor(Snippets.class);
    static final int MAX_SNIPPETS = 30;
    static final String SPLITTER = "(\\s+)|((?<=[\\W&&\\S])(?=\\w))";
    
    private final HashSet<Snippet> snippets;
    
    private GraphSpace space;
    private final BrowseListener browseListener;
    private final EditListener editListener;
    
    public Snippets() {
        snippets = new LinkedHashSet<Snippet>();
        browseListener = new BrowseListener();
        editListener = new EditListener();
    }
    
    public String getName() {
        return "Snippet tracker";
    }

    public void connect(GraphSpace space) throws RemoteException {
        if (space.equals(this.space)) return;
        
        this.space = space;
        space.register(browseListener,
                       tuple(null, uri("#snippet"), (String)null),
                       PROPERTIES.getProperty("snippet-query"));
        space.register(browseListener,
                       tuple(null, uri("#action"), uri("#browse")),
                       PROPERTIES.getProperty("browse-query"));
        space.register(editListener,
                       tuple(null, rdf_type, uri("#edit")),
                       PROPERTIES.getProperty("edit-query"));
    }

    public void disconnect() throws RemoteException {
        if (space == null) return;
        
        space.unregister(browseListener);
        space.unregister(editListener);
        space = null;
    }
    
    private void compare(Snippet snippet, Solution edit) throws RemoteException {
        Match match = align(snippet.tokens, edit.get("code").split(SPLITTER));
        if (match == null) return;
        
        Log.ok("Snippet match, score " + match.score + ", "
                + snippet.solution.get("url") + ", " + snippet.solution.get("title") + ", "
                + edit.get("project") + ", " + edit.get("path"));
        Data.ideEvent("snippet", Integer.toString(match.score));
        Tuple event = unique("#event");
        space.add(
                event,
                tuple(event.subject, uri("#action"), uri("#bookmark-request")),
                timestamp(event.subject),
                tuple(event.subject, uri("#url"), snippet.solution.get("url")),
                tuple(event.subject, uri("#title"), "c.p. " + snippet.solution.get("title")),
                tuple(event.subject, uri("#project"), edit.get("project")),
                tuple(event.subject, uri("#path"), edit.get("path"))
        );
    }
    
    enum Dir { DOWN, RGHT, DIAG };
    /**
     * Computes the Smith-Waterman distance between the given strings.
     * 
     * @param s one string
     * @param t another string
     */
    static Match align(String[] s, String[] t) {
        int lenS = s.length;
        int lenT = t.length;
        int[][] sim = new int[lenS+1][lenT+1];
        Dir[][] dir = new Dir[lenS+1][lenT+1];
        int gap = 1;
        int best = 0, bestI = 0, bestJ = 0;
        int down, rght, diag;
        
        for (int i = 1; i <= lenS; i++) {
            for (int j = 1; j <= lenT; j++) {
                down = sim[i-1][j] - gap;
                rght = sim[i][j-1] - gap;
                diag = sim[i-1][j-1] + (s[i-1].equals(t[j-1]) ? 1 : -1);
                if (down > rght) {
                    if (down > diag) {
                        sim[i][j] = down;
                        dir[i][j] = Dir.DOWN;
                    } else {
                        sim[i][j] = diag;
                        dir[i][j] = Dir.DIAG;
                    }
                } else {
                    if (rght > diag) {
                        sim[i][j] = rght;
                        dir[i][j] = Dir.RGHT;
                    } else {
                        sim[i][j] = diag;
                        dir[i][j] = Dir.DIAG;
                    }
                }
                if (sim[i][j] < 0) {
                    sim[i][j] = 0;
                }
                if (sim[i][j] > best) {
                    best = sim[i][j];
                    bestI = i;
                    bestJ = j;
                }
            }
        }
        
        if (best < 5 || best < Math.min(lenS, lenT)/4) {
            return null; // XXX really?
        }
        
        List<String> matchS = new ArrayList<String>();
        List<String> matchT = new ArrayList<String>();
        
        while (bestI > 0 && bestJ > 0 && sim[bestI][bestJ] > 0) {
            switch (dir[bestI][bestJ]) {
            case DOWN:
                matchS.add(0, s[--bestI]);
                break;
            case RGHT:
                matchT.add(0, t[--bestJ]);
                break;
            case DIAG:
                matchS.add(0, s[--bestI]);
                matchT.add(0, t[--bestJ]);
                break;
            }
        }
        while (bestI > 0 && sim[bestI][bestJ] > 0) {
            matchS.add(0, s[--bestI]);
        }
        while (bestJ > 0 && sim[bestI][bestJ] > 0) {
            matchT.add(0, t[--bestJ]);
        }
        
        return new Match(best, bestI, bestJ, matchS, matchT);
    }
    
    class BrowseListener implements GraphSpaceListener {
        public void notify(Solution solution) {
            snippets.add(new Snippet(solution));
            if (snippets.size() > MAX_SNIPPETS) {
                synchronized (snippets) {
                    Iterator<?> it = snippets.iterator();
                    it.next();
                    it.remove();
                }
            }
        }
    }
    
    class EditListener implements GraphSpaceListener {
        public void notify(Solution solution) throws RemoteException {
            synchronized (snippets) {
                for (Snippet snippet : snippets) {
                    compare(snippet, solution);
                }
            }
        }
    }
    
    static class Snippet {
        final Solution solution;
        final String[] tokens;
        
        Snippet(Solution solution) {
            this.solution = solution;
            this.tokens = solution.get("snippet").split(SPLITTER);
        }
        
        @Override public boolean equals(Object o) {
            if (o instanceof Snippet)
                return solution.equals(((Snippet)o).solution);
            return false;
        }
        
        @Override public int hashCode() {
            return solution.hashCode() + 17;
        }
    }
    
    static class Match {
        final int score;
        final int indexS, indexT;
        final String[] matchS, matchT;
        
        Match(int score, int indexS, int indexT, List<String> matchS, List<String> matchT) {
            this(score, indexS, indexT, matchS.toArray(new String[0]), matchT.toArray(new String[0]));
        }
        
        Match(int score, int indexS, int indexT, String[] matchS, String[] matchT) {
            this.score = score;
            this.indexS = indexS;
            this.indexT = indexT;
            this.matchS = matchS;
            this.matchT = matchT;
        }
        
        @Override public boolean equals(Object o) {
            if (o instanceof Match) {
                Match m = (Match)o;
                return score == m.score
                       && indexS == m.indexS
                       && indexT == m.indexT
                       && Arrays.equals(matchS, m.matchS)
                       && Arrays.equals(matchT, m.matchT);
            }
            return false;
        }
        
        @Override public int hashCode() {
            return score + indexS + indexT + Arrays.hashCode(matchS) + Arrays.hashCode(matchT);
        }
        
        @Override public String toString() {
            return "Match<"+score+";"+indexS+":"+Arrays.toString(matchS)+";"+indexT+":"+Arrays.toString(matchT)+">";
        }
    }
}