/*
 * Codetrail
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package codetrail.eclipse.pref;

import java.security.Principal;
import java.util.*;
import java.util.List;

import org.eclipse.jface.preference.FieldEditor;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.*;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.*;

import codetrail.Util;

/**
 * A field editor for an enumeration type preference.
 * The choices are presented as a list of check boxes.
 * 
 * @see org.eclipse.jface.preference#RadioGroupFieldEditor
 *      provided radio button field editor
 */
public class CheckGroupFieldEditor extends FieldEditor {
    
    private List<? extends Principal> options;
    private int numColumns;
    private int indent = HORIZONTAL_GAP;
    private List<String> value;
    private Composite checkContainer;
    private Button[] checkButtons;
    private boolean useGroup;

    protected CheckGroupFieldEditor() { }

    public CheckGroupFieldEditor(String name, String labelText, int numColumns, List<? extends Principal> options, Composite parent) {
        this(name, labelText, numColumns, options, parent, false);
    }

    public CheckGroupFieldEditor(String name, String labelText, int numColumns, List<? extends Principal> options, Composite parent, boolean useGroup) {
        init(name, labelText);
        this.options = options;
        this.numColumns = numColumns;
        this.useGroup = useGroup;
        createControl(parent);
    }

    protected void adjustForNumColumns(int numColumns) {
        Control control = getLabelControl();
        if (control != null) {
            ((GridData) control.getLayoutData()).horizontalSpan = numColumns;
        }
        ((GridData) checkContainer.getLayoutData()).horizontalSpan = numColumns;
    }

    protected void doFillIntoGrid(Composite parent, int numColumns) {
        if (useGroup) {
            Control control = getRadioBoxControl(parent);
            GridData gd = new GridData(GridData.FILL_HORIZONTAL);
            control.setLayoutData(gd);
        } else {
            Control control = getLabelControl(parent);
            GridData gd = new GridData();
            gd.horizontalSpan = numColumns;
            control.setLayoutData(gd);
            control = getRadioBoxControl(parent);
            gd = new GridData();
            gd.horizontalSpan = numColumns;
            gd.horizontalIndent = indent;
            control.setLayoutData(gd);
        }

    }

    protected void doLoad() {
        updateValue(getPreferenceStore().getString(getPreferenceName()));
    }

    protected void doLoadDefault() {
        updateValue(getPreferenceStore().getDefaultString(getPreferenceName()));
    }

    protected void doStore() {
        if (value == null) {
            getPreferenceStore().setToDefault(getPreferenceName());
            return;
        }

        getPreferenceStore().setValue(getPreferenceName(), Util.commaSeparated(value));
    }

    public int getNumberOfControls() {
        return 1;
    }

    public Composite getRadioBoxControl(Composite parent) {
        if (checkContainer == null) {

            Font font = parent.getFont();

            if (useGroup) {
                Group group = new Group(parent, SWT.NONE);
                group.setFont(font);
                String text = getLabelText();
                if (text != null) {
                    group.setText(text);
                }
                checkContainer = group;
                GridLayout layout = new GridLayout();
                layout.horizontalSpacing = HORIZONTAL_GAP;
                layout.numColumns = numColumns;
                checkContainer.setLayout(layout);
            } else {
                checkContainer = new Composite(parent, SWT.NONE);
                GridLayout layout = new GridLayout();
                layout.marginWidth = 0;
                layout.marginHeight = 0;
                layout.horizontalSpacing = HORIZONTAL_GAP;
                layout.numColumns = numColumns;
                checkContainer.setLayout(layout);
                checkContainer.setFont(font);
            }

            checkButtons = new Button[options.size()];
            for (int i = 0; i < checkButtons.length; i++) {
                final Button check = new Button(checkContainer, SWT.CHECK | SWT.LEFT);
                checkButtons[i] = check;
                check.setText(options.get(i).getName());
                check.setData(options.get(i).getClass().getName());
                check.setFont(font);
                check.addSelectionListener(new SelectionAdapter() {
                    @Override public void widgetSelected(SelectionEvent event) {
                        String oldValue = Util.commaSeparated(value);
                        updateValue(check.getSelection(), (String)check.getData());
                        setPresentsDefaultValue(false);
                        fireValueChanged(VALUE, oldValue, Util.commaSeparated(value));
                    }
                });
            }
            checkContainer.addDisposeListener(new DisposeListener() {
                public void widgetDisposed(DisposeEvent event) {
                    checkContainer = null;
                    checkButtons = null;
                }
            });
        } else {
            checkParent(checkContainer, parent);
        }
        return checkContainer;
    }

    public void setIndent(int indent) {
        if (indent < 0) {
            this.indent = 0;
        } else {
            this.indent = indent;
        }
    }
    
    private void updateValue(boolean add, String singleValue) {
        if (add) {
            if ( ! value.contains(singleValue)) {
                value.add(singleValue);
                Collections.sort(value);
            }
        } else {
            value.remove(singleValue);
        }
    }

    private void updateValue(String selectedValue) {
        value = Util.commaSeparate(selectedValue);
        if (checkButtons == null) {
            return;
        }

        for (int i = 0; i < checkButtons.length; i++) {
            Button check = checkButtons[i];
            boolean selection = false;
            if (this.value.contains(check.getData())) {
                selection = true;
            }
            check.setSelection(selection);
        }
    }

    @Override public void setEnabled(boolean enabled, Composite parent) {
        if (!useGroup) {
            super.setEnabled(enabled, parent);
        }
        for (int i = 0; i < checkButtons.length; i++) {
            checkButtons[i].setEnabled(enabled);
        }
    }
}
