/*
 * Codetrail
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package codetrail.eclipse.pref;

import static codetrail.Tuple.tuple;

import java.rmi.RemoteException;

import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.preference.FieldEditor;
import org.eclipse.jface.resource.JFaceResources;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.*;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.*;

import codetrail.*;

/**
 * A field editor for an unordered list of values backed by the graph space.
 * The editor displays a list containing the values and buttons for
 * manipulating the list.
 * 
 * @see org.eclipse.jface.preference#ListEditor
 *      provided abstract list field editor
 */
public class GraphSpaceListEditor extends FieldEditor {
    
    private final GraphSpace space;
    private final URIString subject, predicate;
    
    private List list;
    private Composite buttonBox;
    private Button removeButton;
    private SelectionListener selectionListener;
    
    public GraphSpaceListEditor(GraphSpace space, URIString subject, URIString predicate,
                                String labelText, Composite parent) {
        this.space = space;
        this.subject = subject;
        this.predicate = predicate;
        this.setLabelText(labelText);
        createControl(parent);
    }

    protected void adjustForNumColumns(int numColumns) {
        Control control = getLabelControl();
        ((GridData)control.getLayoutData()).horizontalSpan = numColumns;
        ((GridData)list.getLayoutData()).horizontalSpan = numColumns - 1;
    }

    protected void doFillIntoGrid(Composite parent, int numColumns) {
        Control control = getLabelControl(parent);
        GridData gd = new GridData();
        gd.horizontalSpan = numColumns;
        control.setLayoutData(gd);
        
        list = getListControl(parent);
        gd = new GridData(GridData.FILL_HORIZONTAL);
        gd.verticalAlignment = GridData.FILL;
        gd.horizontalSpan = numColumns - 1;
        gd.grabExcessHorizontalSpace = false; // XXX
        list.setLayoutData(gd);
        
        buttonBox = getButtonBoxControl(parent);
        gd = new GridData();
        gd.verticalAlignment = GridData.BEGINNING;
        buttonBox.setLayoutData(gd);
    }

    protected void doLoad() {
        try {
            for (Tuple t : space.get(tuple(subject, predicate, (String)null))) {
                list.add(t.object().toString());
            }
        } catch (RemoteException re) { }
    }

    /**
     * Should never be called.
     */
    protected void doLoadDefault() {
        throw new UnsupportedOperationException("Cannot load default for graph space list editor");
    }
    
    /**
     * Should never be called.
     * 
     * @see GraphSpaceListEditor#store()
     */
    protected void doStore() {
        throw new UnsupportedOperationException("Cannot store from graph space list editor");
    }

    public int getNumberOfControls() {
        return 2;
    }
    
    private void createButtons(Composite box) {
        removeButton = createPushButton(box, "ListEditor.remove");
    }
    
    private Button createPushButton(Composite parent, String key) {
        Button button = new Button(parent, SWT.PUSH);
        button.setText(JFaceResources.getString(key));
        button.setFont(parent.getFont());
        GridData data = new GridData(GridData.FILL_HORIZONTAL);
        int widthHint = convertHorizontalDLUsToPixels(button,
                IDialogConstants.BUTTON_WIDTH);
        data.widthHint = Math.max(widthHint, button.computeSize(SWT.DEFAULT,
                SWT.DEFAULT, true).x);
        button.setLayoutData(data);
        button.addSelectionListener(getSelectionListener());
        return button;
    }
    
    void createSelectionListener() {
        selectionListener = new SelectionAdapter() {
            public void widgetSelected(SelectionEvent event) {
                Widget widget = event.widget;
                if (widget == removeButton) {
                    removePressed();
                } else if (widget == list) {
                    selectionChanged();
                }
            }
        };
    }
    
    Composite getButtonBoxControl(Composite parent) {
        if (buttonBox == null) {
            buttonBox = new Composite(parent, SWT.NULL);
            GridLayout layout = new GridLayout();
            layout.marginWidth = 0;
            buttonBox.setLayout(layout);
            createButtons(buttonBox);
            buttonBox.addDisposeListener(new DisposeListener() {
                public void widgetDisposed(DisposeEvent event) {
                    removeButton = null;
                }
            });
        } else {
            checkParent(buttonBox, parent);
        }
        
        selectionChanged();
        return buttonBox;
    }
    
    List getListControl(Composite parent) {
        if (list == null) {
            list = new List(parent, SWT.BORDER | SWT.SINGLE | SWT.V_SCROLL | SWT.H_SCROLL);
            list.setFont(parent.getFont());
            list.addSelectionListener(getSelectionListener());
            list.addDisposeListener(new DisposeListener() {
                public void widgetDisposed(DisposeEvent event) {
                    list = null;
                }
            });
        } else {
            checkParent(list, parent);
        }
        return list;
    }
    
    /**
     * Returns this field editor's selection listener.
     * The listener is created if nessessary.
     *
     * @return the selection listener
     */
    private SelectionListener getSelectionListener() {
        if (selectionListener == null) {
            createSelectionListener();
        }
        return selectionListener;
    }
    
    private void removePressed() {
        setPresentsDefaultValue(false);
        int index = list.getSelectionIndex();
        if (index >= 0) {
            list.remove(index);
            selectionChanged();
        }
    }
    
    private void selectionChanged() {
        int index = list.getSelectionIndex();
        removeButton.setEnabled(index >= 0);
    }
    
    @Override public void setFocus() {
        if (list != null) {
            list.setFocus();
        }
    }
    
    @Override public void setEnabled(boolean enabled, Composite parent) {
        super.setEnabled(enabled, parent);
        getListControl(parent).setEnabled(enabled);
        removeButton.setEnabled(enabled);
    }
    
    @Override public void store() {
        try {
            for (Tuple t : space.get(tuple(subject, predicate, (String)null))) {
                String name = t.object().toString();
                if (list.indexOf(name) < 0) {
                    space.remove(tuple(subject, predicate, name));
                }
            }
        } catch (RemoteException re) { }
    }
}
