/*
 * Codetrail
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package codetrail;

import static org.junit.Assert.assertEquals;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.*;

import org.junit.*;

public class TrieTest {

    @Before public void setUp() { }

    @After public void tearDown() { }
    
    private <T> void put(Trie<String, T> trie, String splitter, T keyValue) {
        trie.put(keyValue.toString().split(splitter), keyValue);
    }
    
    private <T> SortedMap<List<String>, T> result(String splitter, T... keyValues) {
        SortedMap<List<String>, T> result = new TreeMap<List<String>, T>(Trie.LENGTH_COMPARATOR);
        for (T keyValue : keyValues) {
            result.put(Arrays.asList(keyValue.toString().split(splitter)), keyValue);
        }
        return result;
    }
    
    @Test public void testNearStrings() {
        Trie<String, String> trie = new Trie<String, String>();
        
        String[] abc = new String[] { "a", "b", "c" };
        
        put(trie, ",", "a,f,g");
        put(trie, ",", "a,h,i");
        put(trie, ",", "a,b,c");
        put(trie, ",", "a,b");
        put(trie, ",", "a,k,l");
        put(trie, ",", "w,x");
        assertEquals(result(",", "a,b", "a,f,g", "a,h,i", "a,k,l"),
                trie.near(abc, 0));
        assertEquals(result(",", "a,b", "a,f,g", "a,h,i", "a,k,l"),
                trie.near(abc, 2));
        assertEquals(result(","),
                trie.near(abc, 3));
        
        put(trie, ",", "a,b,d");
        assertEquals(result(",", "a,b,d"),
                trie.near(abc, 0));
        assertEquals(result(",", "a,b,d"),
                trie.near(abc, 3));
        
        put(trie, ",", "a,b,e");
        put(trie, ",", "a,b,j,m");
        assertEquals(result(",", "a,b,d", "a,b,e", "a,b,j,m"),
                trie.near(abc, 0));
        assertEquals(result(",", "a,b,d", "a,b,e", "a,b,j,m"),
                trie.near(abc, 3));
        
        put(trie, ",", "a,b,c,n");
        assertEquals(result(",", "a,b,d", "a,b,e", "a,b,c,n", "a,b,j,m"),
                trie.near(abc, 0));
        assertEquals(result(",", "a,b,d", "a,b,e", "a,b,c,n", "a,b,j,m"),
                trie.near(abc, 3));
    }
    
    @Test public void testNearXulPlanetURLs() throws MalformedURLException {
        Trie<String, URL> trie = new Trie<String, URL>();
        
        String rsrc, node, eqno, domp;
        String pattern = "[/\\.#]";
        for (String url : new String[] {
                rsrc = "http://www.xulplanet.com/references/xpcomref/ifaces/nsIRDFResource.html",
                node = "http://www.xulplanet.com/references/xpcomref/ifaces/nsIRDFNode.html",
                eqno = "http://www.xulplanet.com/references/xpcomref/ifaces/nsIRDFNode.html#method_EqualsNode",
                domp = "http://www.xulplanet.com/references/xpcomref/ifaces/nsIDOMParser.html",
                "http://www.xulplanet.com/references/xpcomref/comps/c_rdfresourcefactory1.html",
                "http://www.xulplanet.com/references/xpcomref/",
                "http://www.xulplanet.com/tutorials/",
                "http://www.xulplanet.com/references/elemref/ref_stringbundle.html",
                "http://www.google.com/ig"
                }) {
            trie.put(url.split(pattern), new URL(url));
        }
        assertEquals(result(pattern, new URL(node), new URL(rsrc), new URL(eqno)),
                trie.near(domp.split(pattern), 4));
        assertEquals(result(pattern, new URL(domp), new URL(node), new URL(eqno)),
                trie.near(rsrc.split(pattern), 4));
    }
    
    @Test public void testNearMdcXpcomURLs() throws MalformedURLException {
        Trie<String, URL> trie = new Trie<String, URL>();
        
        String file, lofi, dirs;
        String pattern = "[/\\.#]";
        for (String url : new String[] {
                //"http://developer.mozilla.org/en/docs/Main_Page",
                //"http://developer.mozilla.org/en/docs/XPCOM",
                //"http://developer.mozilla.org/en/docs/XPCOM_API_Reference",
                file = "http://developer.mozilla.org/en/docs/nsIFile",
                lofi = "http://developer.mozilla.org/en/docs/nsILocalFile",
                //"http://developer.mozilla.org/en/docs/nsILocalFile:load",
                dirs = "http://developer.mozilla.org/en/docs/nsIDirectoryService"
                }) {
            trie.put(url.split(pattern), new URL(url));
        }
        assertEquals(result(pattern, new URL(file), new URL(lofi)),
                trie.near(dirs.split(pattern), 4));
    }
}
