/*
 * Codetrail
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package codetrail.eclipse;

import static codetrail.Tuple.*;
import static codetrail.URIString.*;
import static org.junit.Assert.*;

import java.io.File;
import java.sql.SQLException;
import java.util.*;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

import org.junit.*;

import codetrail.*;

import com.hp.hpl.jena.db.IDBConnection;
import com.hp.hpl.jena.rdf.model.*;

public class EclipsedGraphSpaceTest {
    
    static final Properties PROPERTIES = Util.propertiesFor(EclipsedGraphSpaceTest.class);
    static final File TEST_DIR = new File("test-data", EclipsedGraphSpaceTest.class.getSimpleName());
    static final File LOG_FILE = new File(TEST_DIR, "log");
    static final File DB_DIR = new File(TEST_DIR, "db");
    
    private static IDBConnection db;
    private static EclipsedGraphSpace dbSpace;
    
    private EclipsedGraphSpace space;
    
    @BeforeClass public static void classSetUp() throws Exception {
        db = EclipsedGraphSpace.dbConnectionFor(LOG_FILE.getPath(), DB_DIR.getPath());
        dbSpace = new EclipsedGraphSpace(db);
    }
    
    @AfterClass public static void classTearDown() throws Exception {
        dbSpace.close();
        dbSpace = null;
        
        db.close();
        db = null;
        
        LinkedList<File> files = new LinkedList<File>();
        files.add(LOG_FILE);
        files.add(DB_DIR);
        for (ListIterator<File> it = files.listIterator(); it.hasNext(); ) {
            File file = it.next();
            if (file.isHidden()) {
                it.remove();
            } else if ( ! file.getCanonicalPath().startsWith(TEST_DIR.getCanonicalPath())) {
                it.remove();
            } else if (file.isDirectory()) {
                for (File child : file.listFiles()) {
                    it.add(child);
                    it.previous();
                }
            }
        }
        while ( ! files.isEmpty()) {
            files.removeLast().delete();
        }
    }
    
    @Before public void setUp() throws Exception {
        space = new EclipsedGraphSpace();
    }

    @After public void tearDown() throws Exception {
        if (space != null)
            space.close();
        space = null;
        
        dbSpace.removeAll();
    }

    @Test public void testAddEdge() {
        String pattern =
            "(?s).*<rdf:Description +rdf:about=\"\\S+#foo\">\\s*"
            + "<\\S+:bar +rdf:resource=\"\\S+#baz\"/>\\s*"
            + "</rdf:Description>.*";
        assertFalse(space.toXML().matches(pattern));
        space.add(tuple(uri("#foo"), uri("#bar"), uri("#baz")));
        assertTrue(space.toXML().matches(pattern));
    }

    @Test public void testAddLeaf() {
        String pattern =
            "(?s).*<rdf:Description +rdf:about=\"\\S+#fred\">\\s*"
            + "<\\S+:quux>meta</\\S+>\\s*"
            + "</rdf:Description>.*";
        assertFalse(space.toXML().matches(pattern));
        space.add(tuple(uri("#fred"), uri("#quux"), "meta"));
        assertTrue(space.toXML().matches(pattern));
    }
    
    @Test public void testAddTuples() {
        String pattern =
            "(?s).*<rdf:Description +rdf:about=\"\\S+#meta\">\\s*"
            + "<\\S+:abc>123</\\S+>\\s*"
            + "</rdf:Description>\\s*"
            + "<rdf:Description +rdf:about=\"\\S+#fred\">\\s*"
            + "<\\S+:quux rdf:resource=\"\\S+#meta\"/>\\s*"
            + "</rdf:Description>.*";
        assertFalse(space.toXML().matches(pattern));
        space.add(new Tuple[] {
                tuple(uri("#fred"), uri("#quux"), uri("#meta")),
                tuple(uri("#meta"), uri("#abc"), "123")
        });
        assertTrue(space.toXML().matches(pattern));
    }
    
    @Test public void testQuery() {
        String query = "SELECT ?s ?o WHERE { ?s graphspace:p ?o }";
        for (@SuppressWarnings("unused")
             Solution solution : space.query(query)) {
            fail();
        }
        space.add(
                tuple(uri("#s"), uri("#p"), "oh"),
                tuple(uri("#s"), uri("#q"), "no")
        );
        Solution expected = new Solution();
        expected.put("s", uri("#s").uri);
        expected.put("o", "oh");
        Iterable<Solution> solutions = space.query(query);
        assertTrue(solutions.iterator().hasNext());
        for (Solution solution : solutions) {
            assertEquals(expected, solution);
        }
    }
    
    @Test public void testUpdateInsert() {
        String pattern =
            "(?s).*<rdf:Description +rdf:about=\"\\S+#s\">\\s*"
            + "<\\S+:p>object</\\S+>\\s*"
            + "</rdf:Description>.*";
        assertFalse(space.toXML().matches(pattern));
        space.update("INSERT { graphspace:s graphspace:p \"object\" }");
        assertTrue(space.toXML().matches(pattern));
    }
    
    @Test public void testUpdateDelete() {
        String foo =
            "(?s).*<rdf:Description +rdf:about=\"\\S+#foo\">\\s*"
            + "<\\S+:y>z</\\S+>\\s*"
            + "<\\S+:x>y</\\S+>\\s*"
            + "</rdf:Description>.*";
        String barbaz =
            "(?s).*<rdf:Description +rdf:about=\"\\S+#bar\">\\s*"
            + "<\\S+:x>z</\\S+>\\s*"
            + "</rdf:Description>\\s*"
            + "<rdf:Description +rdf:about=\"\\S+#baz\">\\s*"
            + "<\\S+:y>x</\\S+>\\s*"
            + "</rdf:Description>.*";
        
        space.add(
                tuple(uri("#foo"), uri("#x"), "y"),
                tuple(uri("#foo"), uri("#y"), "z"),
                tuple(uri("#bar"), uri("#x"), "z"),
                tuple(uri("#baz"), uri("#y"), "x")
        );
        
        assertTrue(space.toXML().matches(foo));
        space.update("DELETE { ?s ?p ?o } WHERE { ?s ?p ?o . ?s graphspace:x \"y\" }");
        assertTrue(space.toXML().matches(barbaz));
    }
    
    @Test(timeout=1000) public void testRegister() throws InterruptedException {
        Solution solution;
        
        GraphSpaceListenerImpl listener = new GraphSpaceListenerImpl();
        space.register(listener,
                       tuple(null, null, (String)null),
                       "SELECT ?x WHERE { GRAPH graphspace:match { ?x ?y ?z } { ?x ?y ?z } }");
        assertEquals(0, listener.calls);
        space.add(tuple(uri("#a"), uri("#b"), uri("#c")));
        solution = listener.assignments.take();
        assertEquals(1, listener.calls);
        assertEquals(uri("#a").uri, solution.get("x"));
        
        space.unregister(listener);
        
        space.add(tuple(uri("#d"), uri("#e"), uri("#f")));
        // shouldn't get a callback, but can't wait around for it to not happen
        
        space.register(listener,
                       tuple(null, uri("#p1"), uri("#o")),
                       "SELECT ?p2 WHERE { GRAPH graphspace:match { ?s graphspace:p1 graphspace:o } { ?s graphspace:p2 ?p2 } }");
        space.add(new Tuple[] {
                tuple(uri("#abc123"), uri("#p1"), uri("#o")),
                tuple(uri("#abc123"), uri("#p2"), "42")
        });
        solution = listener.assignments.take();
        assertEquals(2, listener.calls);
        assertEquals("42", solution.get("p2"));
    }
    
    @Test(timeout=1000) public void testQuestionAndAnswer() throws InterruptedException {
        Solution solution;
        
        GraphSpaceListenerImpl questioner = new GraphSpaceListenerImpl();
        GraphSpaceListenerImpl answerer = new GraphSpaceListenerImpl();
        
        space.register(answerer, uri("#somewhere"));
        space.ask(questioner, uri("#somewhere"), "The question", "Option A", "Option B");
        solution = answerer.assignments.take();
        assertEquals(0, questioner.calls);
        assertEquals(1, answerer.calls);
        assertEquals("The question", solution.get("message"));
        
        space.answer(uri(solution.get("question")), "Option A");
        solution = questioner.assignments.take();
        assertEquals(1, questioner.calls);
        assertEquals(1, answerer.calls);
        assertEquals("Option A", solution.get("answer"));
    }
    
    void doTestGarbageCollectionSmall(EclipsedGraphSpace space) throws Exception {
        String longstring = "-a-rather-long-seed".replaceAll(".", "-very-long-url");
        Tuple t1, t2;
        space.add(
                tuple(uri("#thing1"), rdf_type, uri("#type1")),
                tuple(uri("#thing1"), uri("#timestamp"), Util.iso8601().format(new Date(new Date().getTime()-5000))),
                tuple(uri("#thing1"), uri("#foo"), uri("#bar")),
                tuple(uri("#bar"), uri("#fred"), "qwerty"),
                tuple(uri("#thing2"), rdf_type, uri("#type2")),
                tuple(uri("#thing2"), uri("#timestamp"), Util.iso8601().format(new Date(new Date().getTime()-5000))),
                tuple(uri("#thing2"), uri("#foo"), uri("#baz")),
                tuple(uri("#baz"), uri("#fred"), "asdf"),
                tuple(uri("#thing3"), rdf_type, uri("#type3")),
                tuple(uri("#thing3"), uri("#foo"), uri("#bar")),
                tuple(uri("#thing4"), rdf_type, uri("#type4")),
                t1 = timestamp(uri("#thing4")),
                tuple(uri("#thing4"), uri("#foo"), uri("#quux")),
                tuple(uri("#quux"), uri("#fred"), "zxcv"),
                tuple(uri("#thing5"), rdf_type, uri("#type5")),
                t2 = timestamp(uri("#thing5")),
                tuple(uri("#thing5"), uri("#foo"), uri("#fred" + longstring)),
                tuple(uri("#fred" + longstring), uri("#abc"), "123")
        );
        space.gc(4500);
        Collection<Tuple> expected = new ArrayList<Tuple>(Arrays.asList(
                t1, t2,
                tuple(uri("#thing3"), rdf_type, uri("#type3")),
                tuple(uri("#thing3"), uri("#foo"), uri("#bar")),
                tuple(uri("#bar"), uri("#fred"), "qwerty"),
                tuple(uri("#thing4"), rdf_type, uri("#type4")),
                tuple(uri("#thing4"), uri("#foo"), uri("#quux")),
                tuple(uri("#quux"), uri("#fred"), "zxcv"),
                tuple(uri("#thing5"), rdf_type, uri("#type5")),
                tuple(uri("#thing5"), uri("#foo"), uri("#fred" + longstring)),
                tuple(uri("#fred" + longstring), uri("#abc"), "123")
        ));
        for (Tuple tuple : space.get(Tuple.tuple(null, null, (String)null))) {
            assertTrue("Uncollected garbage: " + tuple, expected.remove(tuple));
        }
        assertTrue("Collected alive: " + expected, expected.isEmpty());
    }
    
    void doTestGarbageCollection(EclipsedGraphSpace space) throws Exception {
        List<String> roots = new ArrayList<String>();
        for (String[] chain : new String[][] { { "A", "b", "c", "d" },
                                               { "B", "f", "g", "h" },
                                               { "C", "j", "k", "h" },
                                               { "D", "m", "n", "o" },
                                               { "E", "q", "m", "r" },
                                               { "F", "t", "G", "v" },
                                               { "G", "w", "x", "y" },
                                               { "H", "z", "x", "G" } }) {
            String parent = chain[0];
            roots.add(parent);
            space.add(tuple(uri("#" + parent), rdf_type, uri("#test-root")));
            for (Iterator<String> it = Arrays.asList(chain).listIterator(1); it.hasNext(); ) {
                space.add(tuple(uri("#" + parent), uri("#child"), uri("#" + (parent = it.next()))));
            }
        }
        for (Iterator<String> it = roots.iterator(); it.hasNext(); ) {
            String root = it.next();
            it.remove();
            space.add(timestamp(uri("#" + root)));
            space.gc(0);
            Iterator<Solution> jt = space.query(PROPERTIES.getProperty("gc-query-" + roots.toString())).iterator();
            if (roots.isEmpty()) {
                if (jt.hasNext()) fail("Uncollected garbage:\n" + space.toXML());
            } else {
                if ( ! jt.hasNext()) fail("Collected alive with roots " + roots.toString());
            }
        }
    }
    
    void doTestGarbageCollectionConcur(final EclipsedGraphSpace space) throws Exception {
        space.gc(0);
        for (int ii = 0; ii < 100; ii++) {
            space.add(tuple(uri("http://live/" + ii), rdf_type, uri("#live")));
            space.add(tuple(uri("http://live/" + ii), uri("#a"), "live"));
            space.add(tuple(uri("http://gone/" + ii), rdf_type, uri("#gone")));
            space.add(tuple(uri("http://gone/" + ii), uri("#a"), "gone"));
            space.add(tuple(uri("http://dead/" + ii), uri("#a"), "dead"));
        }
        assertEquals(500, space.size());
        Thread remover = new Thread(new Runnable() {
            public void run() {
                boolean removing;
                do {
                    removing = false;
                    for (Iterator<Tuple> it = space.get(tuple(null, rdf_type, uri("#gone"))).iterator(); it.hasNext(); ) {
                        removing = true;
                        try { Thread.sleep(2); } catch (InterruptedException ie) { }
                        it.next();
                        it.remove();
                    }                    
                } while (removing);
            }
        });
        remover.start();
        space.gc(0);
        remover.join();
        
        Iterator<Solution> it;
        // 100 <,a,live>
        it = space.query("SELECT COUNT(*) WHERE { ?x a graphspace:live }").iterator();
        assertTrue(it.hasNext() && it.next().get(".1").startsWith("100"));
        // 100 <,gs:a,"live">
        it = space.query("SELECT COUNT(*) WHERE { ?x graphspace:a 'live' }").iterator();
        assertTrue(it.hasNext() && it.next().get(".1").startsWith("100"));
        
        // 0 <,a,gone>, they were removed by the remover
        it = space.query("SELECT COUNT(*) WHERE { ?x a graphspace:gone }").iterator();
        assertTrue( ! it.hasNext() || it.next().get(".1").startsWith("0"));
        // some <,gs:a,"gone">, otherwise there was no interleaving
        it = space.query("SELECT COUNT(*) WHERE { ?x graphspace:a 'gone' }").iterator();
        assertTrue(it.hasNext() && ! it.next().get(".1").startsWith("0"));
        
        // 0 <,gs:a,"dead">, should be removed by garbage collector
        it = space.query("SELECT COUNT(*) WHERE { ?x graphspace:a 'dead' }").iterator();
        assertTrue( ! it.hasNext() || it.next().get(".1").startsWith("0"));
        
        space.gc(0);
        
        // 0 <,gs:a,"gone">, should be removed now by garbage collector
        it = space.query("SELECT COUNT(*) WHERE { ?x graphspace:a 'gone' }").iterator();
        assertTrue( ! it.hasNext() || it.next().get(".1").startsWith("0"));
    }
    
    @Test public void testGarbageCollectionSmallMem() throws Exception {
        doTestGarbageCollectionSmall(space);
    }
    
    @Test public void testGarbageCollectionMem() throws Exception {
        doTestGarbageCollection(space);
    }
    
    @Test public void testGarbageCollectionConcurMem() throws Exception {
        doTestGarbageCollectionConcur(space);
    }
    
    @Ignore @Test public void testGarbageCollectionInitDb() throws Exception {
        dbSpace.add(unique("#thing"));
        Iterable<Tuple> tuples = dbSpace.get(tuple(null, rdf_type, uri("#thing")));
        tuples.iterator();
        try {
            dbSpace.gc(0);
        } catch (SQLException sqle) {
            if ("X0X95".equals(sqle.getSQLState())) {
                // open ResultSet error valid
            } else if ("40XL1".equals(sqle.getSQLState())) {
                // deadlock timeout error valid
            } else {
                throw sqle;
            }
            tuples = null;
            dbSpace.gc(0);
            return;
        }
        fail("Initialization should fail with open cursors");
    }
    
    @Test public void testGarbageCollectionSmallDb() throws Exception {
        doTestGarbageCollectionSmall(dbSpace);
    }
    
    @Test public void testGarbageCollectionDb() throws Exception {
        doTestGarbageCollection(dbSpace);
    }
    
    @Test public void testGarbageCollectionConcurDb() throws Exception {
        doTestGarbageCollectionConcur(dbSpace);
    }
    
    long createLargeModel(ModelMaker maker, int chains) {
        Model model = maker.openModel(EclipsedGraphSpace.PROPERTIES.getProperty("model"));
        
        Property hasChild = model.createProperty(BASE_URI + SEPARATOR + "child");
        List<Resource> branch = new ArrayList<Resource>();
        for (int chain = 1; chain <= chains; chain++) {
            Resource parent = model.createResource(BASE_URI + SEPARATOR + UUID.randomUUID());
            model.add(parent,
                      model.createProperty("http://www.w3.org/1999/02/22-rdf-syntax-ns#type"),
                      model.createResource(BASE_URI + SEPARATOR + "test-root"));
            for (int i = 0; i < Math.pow(2, chain); i++) {
                Resource child = model.createResource(BASE_URI + SEPARATOR + UUID.randomUUID());
                model.add(parent, hasChild, child);
                int props = (int)Math.pow(Math.random() + 0.7, 5);
                for (int j = 0; j < props; j++) {
                    model.add(child,
                              model.createProperty(BASE_URI + SEPARATOR + UUID.randomUUID()),
                              UUID.randomUUID().toString());
                }
                if (Math.random() < 0.3) {
                    branch.add(child);
                }
                parent = child;
            }
            model.add(model.createResource(BASE_URI + SEPARATOR + "garbage-" + UUID.randomUUID()),
                      hasChild,
                      parent);
        }
        for (Resource parent : branch) {
            for (int chainlen : new int[] { 5, 5, 100 }) {
                for (int i = 0; i < chainlen; i++) {
                    Resource child = model.createResource(BASE_URI + SEPARATOR + UUID.randomUUID());
                    model.add(parent, hasChild, child);
                    parent = child;
                }
            }
        }
        long size = model.size();
        model.close();
        return size;
    }
    
    @Ignore @Test public void testGarbageCollectionHugeDb() throws Exception {
        IDBConnection db = EclipsedGraphSpace.dbConnectionFor(LOG_FILE.getPath(), DB_DIR.getPath());
        int chains = 7;
        long size = createLargeModel(ModelFactory.createModelRDBMaker(db), chains);
        
        space = new EclipsedGraphSpace(db);
        assertEquals(size, space.size());
        space.gc(1000);
        assertEquals(size - chains, space.size());
    }
}

class GraphSpaceListenerImpl implements GraphSpaceListener {
    
    int calls = 0;
    BlockingQueue<Solution> assignments = new LinkedBlockingQueue<Solution>();

    public synchronized void notify(Solution solution) {
        calls++;
        assignments.offer(solution);
    }
}