/*
 * Codetrail
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package codetrail.eclipse.mod;

import static codetrail.Util.DOC_BUILDERER;
import static codetrail.Util.evaluate;
import static org.junit.Assert.assertEquals;

import java.io.InputStream;

import org.eclipse.core.resources.*;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.jdt.core.IJavaProject;
import org.junit.*;
import org.w3c.dom.Document;

import codetrail.Solution;
import codetrail.test.GraphSpaceStub;
import codetrail.test.TestUtil;

public class BookmarksTest {
    
    private final IProgressMonitor monitor = new NullProgressMonitor();
    private IJavaProject javaProject;
    private Bookmarks bookmarks;
    private GraphSpaceStub space;

    @Before public void setUp() throws Exception {
        IWorkspaceRoot root = ResourcesPlugin.getWorkspace().getRoot();
        IProject project = root.getProject(this.getClass().getSimpleName() + "Project");
        
        project.create(monitor);
        project.open(monitor);
        
        javaProject = TestUtil.createJavaProject(project, monitor);
        
        bookmarks = new Bookmarks();
        space = new GraphSpaceStub();
        bookmarks.connect(space);
    }

    @After public void tearDown() throws Exception {
        bookmarks.disconnect();
        javaProject.getProject().delete(true, monitor);
        Thread.yield();
    }
    
    private static void assertBookmarked(String url, String title, IProject project, String path) throws Exception {
        InputStream in = project.getFolder("bookmarks").getFile(title + ".webloc").getContents();
        Document doc = DOC_BUILDERER.newDocumentBuilder().parse(in);
        assertEquals("URL",
                evaluate(doc, "/plist/dict/key[1]").item(0).getTextContent());
        assertEquals(url,
                evaluate(doc, "/plist/dict/string").item(0).getTextContent());
        assertEquals("edu.mit.csail.uid.codetrail.core",
                evaluate(doc, "/plist/dict/key[2]").item(0).getTextContent());
        assertEquals("title",
                evaluate(doc, "/plist/dict/dict/key[1]").item(0).getTextContent());
        assertEquals(title,
                evaluate(doc, "/plist/dict/dict/string").item(0).getTextContent());
        assertEquals("resources",
                evaluate(doc, "/plist/dict/dict/key[2]").item(0).getTextContent());
        assertEquals(path,
                evaluate(doc, "/plist/dict/dict/array/string").item(0).getTextContent());
    }

    @Test public void testBookmark() throws Exception {
        bookmarks.bookmark(
                "http://web.mit.edu/",
                "MIT",
                javaProject.getProject(),
                javaProject.getProject().getFolder("folder").getFile("file.txt").getProjectRelativePath().toPortableString());
        assertBookmarked("http://web.mit.edu/", "MIT", javaProject.getProject(), "folder/file.txt");
    }
    
    @Test public void testRequest() throws Exception {
        Solution req = new Solution(
                "url", "http://web.mit.edu/",
                "title", "MIT",
                "project", javaProject.getProject().getName(),
                "path", "folder/file.txt");
        bookmarks.requestListener.notify(req);
        assertBookmarked("http://web.mit.edu/", "MIT", javaProject.getProject(), "folder/file.txt");
    }
}
