/*
 * Codetrail
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package codetrail.eclipse.mod;

import static codetrail.Tuple.tuple;
import static codetrail.URIString.rdf_type;
import static codetrail.URIString.uri;
import static java.util.concurrent.TimeUnit.SECONDS;
import static org.junit.Assert.*;

import java.io.*;
import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.resources.*;
import org.eclipse.core.runtime.*;
import org.eclipse.jdt.core.IJavaProject;
import org.junit.*;

import codetrail.*;
import codetrail.test.GraphSpaceStub;
import codetrail.test.TestUtil;

public class DocGuesserTest {
    
    /**
     * Directory of test files.
     */
    final String TEST_DIR = "test-data" + File.separator + this.getClass().getSimpleName();
    
    private final IProgressMonitor monitor = new NullProgressMonitor();
    private IJavaProject javaProject;
    private DocGuesser docGuesser;
    private GraphSpaceStub space;
    private List<Solution> visits;
    
    @Before public void setUp() throws Exception {
        IWorkspaceRoot root = ResourcesPlugin.getWorkspace().getRoot();
        IProject project = root.getProject(this.getClass().getSimpleName() + "Project");
        
        project.create(monitor);
        project.open(monitor);
        
        javaProject = TestUtil.createJavaProject(project, monitor);
        
        docGuesser = new DocGuesser();
        space = new GraphSpaceStub();
        docGuesser.connect(space);
        
        visits = new ArrayList<Solution>();
        visits.add(new Solution("url", "http://www.xulplanet.com/references/xpcomref/ifaces/nsILocalFile.html",
                                "title", "Interface Reference - nsILocalFile",
                                "source", uri("#firefoxy").uri));
        visits.add(new Solution("url", "http://www.xulplanet.com/references/xpcomref/ifaces/nsIDirectoryService.html",
                                "title", "Interface Reference - nsIDirectoryService",
                                "source", uri("#firefoxy").uri));
        visits.add(new Solution("url", "http://www.xulplanet.com/references/xpcomref/ifaces/nsIFile.html",
                                "title", "Interface Reference - nsIFile",
                                "source", uri("#firefoxy").uri));
    }

    @After public void tearDown() throws Exception {
        docGuesser.disconnect();
        javaProject.getProject().delete(true, monitor);
        Thread.yield();
    }
    
    /**
     * Test visits to doc site when no matching keywords are in the project.
     */
    @Test public void testNoMatch() throws Exception {
        for (Solution visit : visits) {
            docGuesser.notify(visit);
            Thread.yield();
        }
        assertEquals(0, space.askCount);
    }
    
    /**
     * Test visits to doc site when matching keywords are in the project.
     */
    @Test public void testMatch() throws Exception {
        addFilesToProject(TEST_DIR, "xpcom.js");
        for (Solution visit : visits) {
            docGuesser.notify(visit);
            Thread.yield();
        }
        Thread.interrupted(); // XXX why were we interrupted?
        synchronized (docGuesser.workspaceSearchers) {
            docGuesser.workspaceSearchers.wait(5 * 1000);
        }
        assertEquals(1, space.askCount);
        space.added.clear();
        synchronized (docGuesser.workspaceAttachers) {
            docGuesser.workspaceAttachers.wait(60 * 1000);
        }
        Tuple proto = tuple(null, uri("#token"), "nsIRDFNode".toLowerCase());
        Tuple tuple = space.added.take();
        while ((tuple = space.added.poll(1, SECONDS)) != null) {
            if (TupleUtil.matches(proto, tuple))
                return;
        }
        fail("No tuple matching " + proto);
    }
    
    /**
     * Test visits to doc site after answering "ignore."
     */
    @Test public void testIgnoreMatch() throws Exception {
        addFilesToProject(TEST_DIR, "xpcom.js");
        space.answerIndex = 1;
        for (Solution visit : visits) {
            docGuesser.notify(visit);
            Thread.yield();
        }
        Thread.interrupted(); // XXX why were we interrupted?
        synchronized (docGuesser.workspaceSearchers) {
            docGuesser.workspaceSearchers.wait(5 * 1000);
        }
        assertEquals(1, space.askCount);
        Thread.yield();
        assertEquals(0, docGuesser.workspaceAttachers.size());
        assertTrue(space.added.contains(tuple(uri("#mod-doc-guesser"),
                                              rdf_type,
                                              uri("#module"))));
        assertTrue(space.added.contains(tuple(uri("#mod-doc-guesser"),
                                              uri("#ignore"),
                                              "http://www.xulplanet.com/references/xpcomref/ifaces/.*.html")));
    }
    
    private void addFilesToProject(String directory, String... files) throws CoreException, FileNotFoundException {
        for (String filename : files) {
            File source = new File(directory + File.separator + filename);
            assertTrue(source.canRead());
            IFile destination = javaProject.getProject().getFile(new Path(filename));
            assertFalse(destination.exists());
            destination.create(new FileInputStream(source), true, monitor);
        }
        javaProject.open(monitor);
    }
}
