/*
 * Codetrail
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package codetrail.eclipse.mod;

import static org.junit.Assert.assertEquals;

import org.eclipse.core.resources.*;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.jdt.core.*;
import org.junit.*;

import codetrail.test.TestUtil;

public class JavaReporterTest {
    
    private static final String JAVADOC = "http://java.sun.com/j2se/1.5.0/docs/api/";
    
    private final IProgressMonitor monitor = new NullProgressMonitor();
    private IJavaProject javaProject;

    @Before public void setUp() throws Exception {
        IWorkspaceRoot root = ResourcesPlugin.getWorkspace().getRoot();
        IProject project = root.getProject(this.getClass().getSimpleName() + "Project");
        
        project.create(monitor);
        project.open(monitor);
        
        javaProject = TestUtil.createJavaProject(project, monitor);
    }

    @After public void tearDown() throws Exception {
        javaProject.getProject().delete(true, monitor);
        Thread.yield(); // project needs time to die
    }

    @Test public void testTypeDocumentationURL() throws JavaModelException {
        // vanilla
        assertEquals(JAVADOC + "java/util/HashMap.html",
                JavaReporter.documentationURL(javaProject.findType("java.util.HashMap")));
        // inner
        assertEquals(JAVADOC + "java/util/Map.Entry.html",
                JavaReporter.documentationURL(javaProject.findType("java.util.Map.Entry")));
    }

    @Test public void testFieldDocumentationURL() throws JavaModelException {
        // static
        assertEquals(JAVADOC + "java/lang/Boolean.html#TYPE",
                JavaReporter.documentationURL(javaProject.findType("java.lang.Boolean").getField("TYPE")));
        // member
        assertEquals(JAVADOC + "java/awt/Point.html#x",
                JavaReporter.documentationURL(javaProject.findType("java.awt.Point").getField("x")));
    }

    @Test public void testMethodDocumentationURL() throws JavaModelException {
        // no arguments
        assertMethod("java/lang/ClassLoader.html#getParent()",
                "java.lang.ClassLoader", "getParent");
        // final argument is non-var-arg array
        assertMethod("java/lang/ClassLoader.html#setSigners(java.lang.Class, java.lang.Object[])",
                "java.lang.ClassLoader", "setSigners", "Ljava.lang.Class<*>;", "[Ljava.lang.Object;");
        
        // only arg is var-arg
        assertMethod("java/lang/Class.html#getConstructor(java.lang.Class...)",
                "java.lang.Class", "getConstructor", "[Ljava.lang.Class;");
        // only arg is class
        assertMethod("java/lang/Class.html#getField(java.lang.String)",
                "java.lang.Class", "getField", "Ljava.lang.String;");
        // last arg is var-arg
        assertMethod("java/lang/Class.html#getMethod(java.lang.String, java.lang.Class...)",
                "java.lang.Class", "getMethod", "Ljava.lang.String;", "[Ljava.lang.Class;");
        // only arg is parameterized
        assertMethod("java/lang/Class.html#isAnnotationPresent(java.lang.Class)",
                "java.lang.Class", "isAnnotationPresent", "Ljava.lang.Class<+Ljava.lang.annotation.Annotation;>;");
        
        // type of var-arg is parameterized
        assertMethod("java/util/Collections.html#addAll(java.util.Collection, T...)",
                "java.util.Collections", "addAll", "Ljava.util.Collection<-TT;>;", "[TT;");
        // type of only arg is parameterized
        assertMethod("java/util/Collections.html#singleton(T)",
                "java.util.Collections", "singleton", "TT;");
        
        // no argument constructor
        assertMethod("java/lang/String.html#String()", "java.lang.String", "String");
        // constructor with argument; byte
        assertMethod("java/lang/String.html#String(byte[])", "java.lang.String", "String", "[B");
        // boolean
        assertMethod("java/lang/String.html#valueOf(boolean)", "java.lang.String", "valueOf", "Z");
        // char
        assertMethod("java/lang/String.html#valueOf(char)", "java.lang.String", "valueOf", "C");
        // primitive array
        assertMethod("java/lang/String.html#valueOf(char[])", "java.lang.String", "valueOf", "[C");
        // multiple primitives
        assertMethod("java/lang/String.html#valueOf(char[], int, int)", "java.lang.String", "valueOf", "[C", "I", "I");
        // double
        assertMethod("java/lang/String.html#valueOf(double)", "java.lang.String", "valueOf", "D");
        // float
        assertMethod("java/lang/String.html#valueOf(float)", "java.lang.String", "valueOf", "F");
        // int
        assertMethod("java/lang/String.html#valueOf(int)", "java.lang.String", "valueOf", "I");
        // long
        assertMethod("java/lang/String.html#valueOf(long)", "java.lang.String", "valueOf", "J");
        
        // constructor with parameterized argument
        assertMethod("java/util/HashMap.html#HashMap(java.util.Map)",
                "java.util.HashMap", "HashMap", "Ljava.util.Map<+TK;+TV;>;");
        // types of all args are parameterized
        assertMethod("java/util/HashMap.html#put(K, V)",
                "java.util.HashMap", "put", "TK;", "TV;");
    }
    
    private void assertMethod(String doc, String type, String method, String... args) throws JavaModelException {
        assertEquals(JAVADOC + doc, JavaReporter.documentationURL(javaProject.findType(type).getMethod(method, args)));
    }
}
