/*
 * Codetrail
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package codetrail.eclipse.mod;

import static codetrail.URIString.*;
import static codetrail.Tuple.*;
import static org.eclipse.jdt.core.IClasspathAttribute.JAVADOC_LOCATION_ATTRIBUTE_NAME;
import static org.junit.Assert.*;

import java.io.*;
import java.util.*;

import org.eclipse.core.resources.*;
import org.eclipse.core.runtime.*;
import org.eclipse.jdt.core.*;
import org.junit.*;

import codetrail.Solution;
import codetrail.test.GraphSpaceStub;
import codetrail.test.TestUtil;

public class JavadocAttacherTest {
    
    /**
     * URL indicative of a Javadoc site.
     */
    static final String TEST_URL = "http://java.sun.com/products/javamail/javadocs/allclasses-frame.html";
    /**
     * Directory of test jars.
     */
    static final String TEST_DIR = "test-data" + File.separator + JavadocAttacherTest.class.getSimpleName();
    /**
     * Names of test jars in directory.
     */
    static final String[] TEST_FILES = new String[] { "activation.jar", "mail.jar" };
    
    private final IProgressMonitor monitor = new NullProgressMonitor();
    private IJavaProject javaProject;
    private JavadocAttacher javadocAttacher;
    private GraphSpaceStub space;
    private Solution visit;

    @Before public void setUp() throws Exception {
        IWorkspaceRoot root = ResourcesPlugin.getWorkspace().getRoot();
        IProject project = root.getProject(this.getClass().getSimpleName() + "Project");
        
        project.create(monitor);
        project.open(monitor);
        
        javaProject = TestUtil.createJavaProject(project, monitor);
        
        javadocAttacher = new JavadocAttacher();
        space = new GraphSpaceStub();
        javadocAttacher.connect(space);
        
        visit = new Solution();
        visit.put("url", TEST_URL);
        visit.put("source", uri("#firefoxy").uri);
    }
    
    @After public void tearDown() throws Exception {
        javadocAttacher.disconnect();
        javaProject.getProject().delete(true, monitor);
        while(javaProject.getProject().exists()) {
            Thread.sleep(100);
        }
        Thread.yield(); // project needs time to die
    }
    
    /**
     * Test visit to Javadoc site when no matching jars are in the project.
     */
    @Test public void testNoMatch() throws Exception {
        javadocAttacher.notify(visit);
        assertEquals(0, space.askCount);
    }
    
    /**
     * Test visit to Javadoc site when one matching jar is in the project.
     */
    @Test public void testMatch() throws Exception {
        addLibrariesToProject(TEST_DIR, TEST_FILES);
        javadocAttacher.notify(visit);
        assertEquals(1, space.askCount);
        assertClasspathAttrib(TEST_FILES[1],
                              JAVADOC_LOCATION_ATTRIBUTE_NAME,
                              "http://java.sun.com/products/javamail/javadocs/");
    }
    
    /**
     * Test visit to Javadoc site after answering "ignore."
     */
    @Test public void testIgnoreMatch() throws Exception {
        addLibrariesToProject(TEST_DIR, TEST_FILES);
        space.answerIndex = 1;
        javadocAttacher.notify(visit);
        assertEquals(1, space.askCount);
        assertClasspathAttrib(TEST_FILES[1],
                              JAVADOC_LOCATION_ATTRIBUTE_NAME,
                              null);
        assertTrue(space.added.contains(tuple(uri("#mod-javadoc-attacher"),
                                              rdf_type,
                                              uri("#module"))));
        assertTrue(space.added.contains(tuple(uri("#mod-javadoc-attacher"),
                                              uri("#ignore"),
                                              "http://java.sun.com/products/javamail/javadocs/allclasses-frame.html")));
    }
    
    private void addLibrariesToProject(String directory, String[] files) throws CoreException, FileNotFoundException, JavaModelException {
        for (String filename : files) {
            File source = new File(directory + File.separator + filename);
            assertTrue(source.canRead());
            IFile destination = javaProject.getProject().getFile(new Path(filename));
            assertFalse(destination.exists());
            destination.create(new FileInputStream(source), true, monitor);
        }

        Set<IClasspathEntry> entries = new HashSet<IClasspathEntry>();
        entries.addAll(Arrays.asList(javaProject.getRawClasspath()));
        for (String filename : files) {
            IPath path = javaProject.getProject().getFile(new Path(filename)).getFullPath();
            entries.add(JavaCore.newLibraryEntry(path, null, null));
        }
        javaProject.setRawClasspath(entries.toArray(new IClasspathEntry[0]), monitor);
        javaProject.open(monitor);
    }
    
    private void assertClasspathAttrib(String file, String name, String value) throws CoreException {
        for (IClasspathEntry entry : javaProject.getRawClasspath()) {
            if ( ! entry.getPath().lastSegment().equals(file)) continue;
            for (IClasspathAttribute attrib : entry.getExtraAttributes()) {
                if ( ! attrib.getName().equals(name)) continue;
                assertEquals(value, attrib.getValue());
                return;
            }
            if (value == null) return;
            fail("Classpath entry for " + file + " missing attribute " + name);
        }
        fail("No classpath entry for " + file);
    }
}