/*
 * Codetrail
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package codetrail.test;

import static org.junit.Assert.fail;

import java.io.Serializable;
import java.util.*;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

import codetrail.*;

/**
 * <code>GraphSpace</code> stub for testing.
 * 
 * @author maxg
 */
public class GraphSpaceStub implements GraphSpace {
    
    public BlockingQueue<Tuple> added = new LinkedBlockingQueue<Tuple>();
    public BlockingQueue<Tuple> removed = new LinkedBlockingQueue<Tuple>();
    public int askCount = 0;
    public int answerIndex = 0;
    
    public GraphSpace initialize(String xml) { return this; }
    public void close() { }
    
    public GraphSpace add(Tuple tuple) {
        added.add(tuple);
        return this;
    }
    
    public GraphSpace add(Tuple... tuples) {
        for (Tuple tuple : tuples) {
            add(tuple);
        }
        return this;
    }
    
    public GraphSpace add(List<Tuple> tuples) {
        for (Tuple tuple : tuples) {
            add(tuple);
        }
        return this;
    }
    
    public GraphSpace remove(Tuple tuple) {
        removed.add(tuple);
        return this;
    }

    public GraphSpace remove(Tuple... tuples) {
        for (Tuple tuple : tuples) {
            remove(tuple);
        }
        return this;
    }

    public GraphSpace remove(List<Tuple> tuples) {
        for (Tuple tuple : tuples) {
            remove(tuple);
        }
        return this;
    }
    
    public GraphSpace removeAll(Tuple template) {
        for (Tuple tuple : added) {
            if (TupleUtil.matches(template, tuple)) {
                remove(tuple);
            }
        }
        return this;
    }
    
    public GraphSpace removeAll() {
        removed.addAll(added);
        return this;
    }
    
    public Iterable<Tuple> get(Tuple template) {
        return new IterableStub<Tuple>();
    }
    
    public Iterable<Solution> query(String queryString) {
        return new IterableStub<Solution>();
    }
    
    public void update(String updateString) { }
    public void register(GraphSpaceListener listener, Tuple template, String queryString) { }
    public void unregister(GraphSpaceListener listener) { }
    public void register(GraphSpaceBase listener) { }
    public void unregister(GraphSpaceBase listener) { }
    public void register(GraphSpaceListener answerer, URIString source) { }
    
    public void ask(GraphSpaceListener listener, URIString source, String message, String... options) {
        try {
            askCount++;
            Solution s = new Solution();
            s.put("answer", options[answerIndex]);
            listener.notify(s);
        } catch (Exception e) {
            fail(e.getMessage());
        }
    }
    
    public void answer(URIString question, String answer) { }
    public long size() { return 0; }
    public String toXML() { return ""; }
}

class IterableStub<T> implements Iterable<T>, Serializable {
    
    private static final long serialVersionUID = 1L;

    public Iterator<T> iterator() {
        return new Iterator<T>() {
            public boolean hasNext() { return false; }
            public T next() { throw new NoSuchElementException(); }
            public void remove() { }
        };
    }
}