/*
 * Codetrail
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

const Cc = Components.classes;
const Ci = Components.interfaces;

const EXTENSION_NAME = "@NAME@";
const EXTENSION_HUMAN_NAME = "@HUMAN_NAME@";

const THIS_DESCRIPTION = "field study data";
const THIS_CID = Components.ID("{681a76e7-bdf3-4979-acc6-66fcd1f6af4e}");
const THIS_CONTRACTID = "@csail.mit.edu/codetrail/study/Data;1";

Log = function() { return Cc["@csail.mit.edu/jsjr/log;1"].getService().wrappedJSObject; }

const AppInfoId = "@mozilla.org/xre/app-info;1";
const StringBundleId = "@mozilla.org/intl/stringbundle;1";
const PreferencesServicesId = "@mozilla.org/preferences-service;1";
const DirectoryServiceId = "@mozilla.org/file/directory_service;1";
const StorageServiceId = "@mozilla.org/storage/service;1";
const TimerManagerId = "@mozilla.org/updates/timer-manager;1";
const HttpsId = "@mozilla.org/network/protocol;1?name=https";
const CertDbId = "@mozilla.org/security/x509certdb;1";
const CertOverrideId = "@mozilla.org/security/certoverride;1";
const ProxyServiceId = "@mozilla.org/network/protocol-proxy-service;1";
const BinInStreamId = "@mozilla.org/binaryinputstream;1";
const FileInStreamId = "@mozilla.org/network/file-input-stream;1";
const BuffInStreamId = "@mozilla.org/network/buffered-input-stream;1";
const WindowMediatorId = "@mozilla.org/appshell/window-mediator;1"

/*
 * Database setup and SQL statements.
 */

const DataDb = { }

DataDb.table = {
    source: "source",
    timeline: "timeline"
}

DataDb.create = {
    source: "CREATE TABLE IF NOT EXISTS " + DataDb.table.source + " (key TEXT NOT NULL PRIMARY KEY, value TEXT NOT NULL)",
    timeline: "CREATE TABLE IF NOT EXISTS " + DataDb.table.timeline + " ("
            + "id INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,"
            + "source TEXT NOT NULL,"
            + "event TEXT NOT NULL,"
            + "tag TEXT,"
            + "data TEXT,"
            + "ts NOT NULL DEFAULT CURRENT_TIMESTAMP"
            + ")",
    timelineTsIdx: "CREATE INDEX IF NOT EXISTS " + DataDb.table.timeline + "TsIdx ON " + DataDb.table.timeline
            + " (ts ASC, id ASC)"
}

DataDb.stmt = {
    insertSource: "INSERT OR REPLACE INTO " + DataDb.table.source + " (key, value) VALUES (?1, ?2)",
    insertTimeline: "INSERT INTO " + DataDb.table.timeline + " (source, event, tag, data) VALUES (?1, ?2, ?3, ?4)",
    insertTimelineTs: "INSERT INTO " + DataDb.table.timeline + " (source, event, tag, data, ts) VALUES (?1, ?2, ?3, ?4, ?5)",
    updateTimelineTag: "UPDATE " + DataDb.table.timeline + " SET tag = ?2 WHERE id = ?1",
    updateTimelineData: "UPDATE " + DataDb.table.timeline + " SET data = ?2 WHERE id = ?1",
    countTimeline: "SELECT COUNT(*) FROM " + DataDb.table.timeline,
    selectTimeline: "SELECT ts, source, event, tag, data FROM " + DataDb.table.timeline + " WHERE id = "
            + "(SELECT COUNT(*) FROM " + DataDb.table.timeline + ") - ?1"
}

const PrivDb = { }

PrivDb.table = {
    domains: "shared_domains",
    urls: "shared_urls",
    hidden: "not_shared"
}

PrivDb.create = {
    domains: "CREATE TABLE IF NOT EXISTS " + PrivDb.table.domains + " (domain TEXT NOT NULL PRIMARY KEY)",
    urls: "CREATE TABLE IF NOT EXISTS " + PrivDb.table.urls + " (url TEXT NOT NULL PRIMARY KEY)",
    hidden: "CREATE TABLE IF NOT EXISTS " + PrivDb.table.hidden + " ("
            + "id INTEGER NOT NULL PRIMARY KEY,"
            + "domain TEXT NOT NULL,"
            + "url TEXT NOT NULL"
            + ")"
}

PrivDb.stmt = {
    insertDomain: "INSERT OR REPLACE INTO " + PrivDb.table.domains + " (domain) VALUES (?1)",
    insertUrl: "INSERT OR REPLACE INTO " + PrivDb.table.urls + " (url) VALUES (?1)",
    selectShared: "SELECT 1 AS match FROM " + PrivDb.table.domains + " WHERE domain = ?1"
            + " UNION "
            + "SELECT 1 AS match FROM " + PrivDb.table.urls + " WHERE url = ?2",
    insertHidden: "INSERT INTO " + PrivDb.table.hidden + " (id, domain, url) VALUES (?1, ?2, ?3)",
    selectHidden: "SELECT h.id, h.url FROM " + PrivDb.table.hidden + " AS h,"
            + "(SELECT domain AS domain, null AS url FROM " + PrivDb.table.domains
            + " UNION "
            + "SELECT null AS domain, url AS url FROM " + PrivDb.table.urls + ") AS s"
            + " WHERE h.domain = s.domain OR h.url = s.url",
    deleteHidden: "DELETE FROM " + PrivDb.table.hidden + " WHERE id = ?1"
}

/*
 * Field study: data store.
 */

function StudyData() {
    this.wrappedJSObject = this;
}

StudyData.prototype = {
    init: function() {
        this.strings = Cc[StringBundleId]
                           .getService(Ci.nsIStringBundleService)
                               .createBundle("chrome://"+EXTENSION_NAME+"/content/study/data.properties");
        this.initEnabled();
        this.initDatabase();
        this.initHttp();
        
        var app = Cc[AppInfoId].getService(Ci.nsIXULAppInfo);
        this.source("firefox", app.name + " " + app.version + " " + app.appBuildID);
    },
    
    initEnabled: function() {
        try {
            this.enabled = Cc[PreferencesServicesId]
                               .getService(Ci.nsIPrefService)
                                   .getBranch(EXTENSION_NAME + ".")
                                       .getBoolPref("installed");
        } catch (e) {
            this.enabled = false;
        }
        
        if (this.enabled) {
            Cc[TimerManagerId]
                .getService(Ci.nsIUpdateTimerManager)
                    .registerTimer(EXTENSION_NAME + "-data-submit-timer", this, 23 * 60 * 60);
        }
    },
    
    initDatabase: function() {
        var dirs = Cc[DirectoryServiceId].getService(Ci.nsIProperties);
        var store = Cc[StorageServiceId].getService(Ci.mozIStorageService);
        
        this.datafile = dirs.get("ProfD", Ci.nsIFile);
        this.datafile.append(EXTENSION_NAME + "-data.sqlite");
        DataDb.conn = store.openDatabase(this.datafile);
        
        var privfile = dirs.get("ProfD", Ci.nsIFile);
        privfile.append(EXTENSION_NAME + "-priv.sqlite");
        PrivDb.conn = store.openDatabase(privfile);
        
        [ DataDb, PrivDb ].forEach(function (db) {
            for each (var sql in db.create) {
                db.conn.executeSimpleSQL(sql);
            }
            for (var name in db.stmt) {
                db.stmt[name] = db.conn.createStatement(db.stmt[name]);
            }
        });
    },
    
    initHttp: function() {
        this.https = Cc[HttpsId].getService(Ci.nsIHttpProtocolHandler);
        
        var cert = this.strings.GetStringFromName("server_cert");
        this.cert = Cc[CertDbId]
                        .getService(Ci.nsIX509CertDB)
                            .constructX509FromBase64(cert);
        
        if (Cc[CertOverrideId]) {
            // MOZ Firefox 3
            var host = this.strings.GetStringFromName("submit_host");
            var port = 443;
            var certs = Cc[CertOverrideId].getService(Ci.nsICertOverrideService);
            var bits = {};
            var temporary = {};
            if ( ! certs.hasMatchingOverride(host, port, this.cert, bits, temporary)) {
                certs.rememberValidityOverride(host,
                                               port,
                                               this.cert,
                                               Ci.nsICertOverrideService.ERROR_UNTRUSTED,
                                               true);
            } else {
                // XXX what if it has the wrong bits?
            }
        } else {
            // MOZ Firefox 2 handled in nsIBadCertListener callback
        }
    },
    
    /*
     * nsITimerCallback
     */
    
    notify: function(timer) {
        Log().ok("Submit data");
        this.submit(false);
    },
    
    getSharedCount: function() {
        var select = DataDb.stmt.countTimeline;
        try {
            select.executeStep();
            return select.getInt64(0);
        } finally {
            select.reset();
        }
    },
    
    getSharedData: function(row) {
        var select = DataDb.stmt.selectTimeline;
        try {
            select.bindInt64Parameter(0, row);
            select.executeStep();
            return [ select.getUTF8String(0),
                     select.getUTF8String(1),
                     select.getUTF8String(2),
                     select.getUTF8String(3),
                     select.getUTF8String(4) ];
        } finally {
            select.reset();
        }
    },
    
    source: function(key, value) {
        //dump("StudyData.source(" + key + ", " + value + ")\n");
        var insert = DataDb.stmt.insertSource;
        try {
            insert.bindUTF8StringParameter(0, key);
            insert.bindUTF8StringParameter(1, value);
            insert.execute();
        } finally {
            insert.reset();
        }
    },
    
    event: function(insert, source, event, tag, url) {
        try {
            insert.bindUTF8StringParameter(0, source);
            insert.bindUTF8StringParameter(1, event);
            if (tag == null) {
                insert.bindNullParameter(2);
            } else {
                insert.bindUTF8StringParameter(2, tag);
            }
            if (url == null) {
                insert.bindNullParameter(3);
            } else {
                insert.bindUTF8StringParameter(3, url);
            }
            insert.execute();
        } finally {
            insert.reset();
        }
    },
    
    browserEvent: function(event, tag, url) {
        if ( ! this.enabled) return;
        
        var insert = DataDb.stmt.insertTimeline;
        this.event(insert, "firefox", event, tag, url);
    },
    
    waitingForBrowserFocus: true,
    browserFocus: function() {
        if ( ! this.waitingForBrowserFocus) return;
        
        this.waitingForBrowserFocus = false;
        this.browserEvent("window focus");
    },
    
    ideEvent: function(event, tag, url, ts) {
        if ( ! this.enabled) return;
        
        this.waitingForBrowserFocus = true;
        var insert = DataDb.stmt.insertTimelineTs;
        insert.bindUTF8StringParameter(4, ts);
        this.event(insert, "eclipse", event, tag, url);
    },
    
    domain: function(uri) {
        return uri.host;
    },
    
    url: function(uri) {
        return uri.scheme + "://" + uri.hostPort + uri.path;
    },
    
    browse: function(event, uri) {
        if ( ! this.enabled) return 0;
        
        var domain = this.domain(uri);
        var url = this.url(uri);
                
        var select = PrivDb.stmt.selectShared;
        try {
            select.bindUTF8StringParameter(0, domain);
            select.bindUTF8StringParameter(1, url);
            var share = select.executeStep();
        } finally {
            select.reset();
        }
        
        var insert = DataDb.stmt.insertTimeline;
        try {
            insert.bindUTF8StringParameter(0, "firefox");
            insert.bindUTF8StringParameter(1, event);
            insert.bindNullParameter(2);
            if (share) {
                insert.bindUTF8StringParameter(3, url);
            } else {
                insert.bindNullParameter(3);
            }
            insert.execute();
            var id = DataDb.conn.lastInsertRowID;
        } finally {
            insert.reset();
        }
        
        /* DISABLED->
        if ( ! share) {
            insert = PrivDb.stmt.insertHidden;
            try {
                insert.bindInt64Parameter(0, id);
                insert.bindUTF8StringParameter(1, domain);
                insert.bindUTF8StringParameter(2, url);
                insert.execute();
            } finally {
                insert.reset();
            }
        }
        <-DISABLED */
        
        return (share ? -id : id);
    },
    
    tag: function(id, tag) {
        if ( ! this.enabled) return;
        
        var update = DataDb.stmt.updateTimelineTag;
        try {
            update.bindInt64Parameter(0, id);
            update.bindUTF8StringParameter(1, tag);
            update.execute();
        } finally {
            update.reset();
        }
    },
    
    shareDomain: function(domain) {
        var insert = PrivDb.stmt.insertDomain;
        try {
            insert.bindUTF8StringParameter(0, domain);
            insert.execute();
        } finally {
            insert.reset();
        }
    },
    
    shareUrl: function(url) {
        var insert = PrivDb.stmt.insertUrl;
        try {
            insert.bindUTF8StringParameter(0, url);
            insert.execute();
        } finally {
            insert.reset();
        }
    },
    
    updateSharedData: function() {
        var select = PrivDb.stmt.selectHidden;
        var update = DataDb.stmt.updateTimelineData;
        var ids = [ ];
        try {
            while(select.executeStep()) {
                try {
                    update.bindInt64Parameter(0, select.getInt64(0));
                    update.bindUTF8StringParameter(1, select.getUTF8String(1));
                    update.execute();
                    ids.push(select.getInt64(0));
                } finally {
                    update.reset();
                }
            }
        } finally {
            select.reset();
        }
        var celete = PrivDb.stmt.deleteHidden;
        ids.forEach(function(id) {
            try {
                celete.bindInt64Parameter(0, id);
                celete.execute();
            } finally {
                celete.reset();
            }
        });
    },
    
    submit: function(notify) {
        //dump("StudyData.submit()\n");
        if ( ! this.enabled) return;
        
        this.updateSharedData();
        
        var uri = this.https.newURI(this.strings.GetStringFromName("submit_url"), null, null);
        var data = this;
        Cc[ProxyServiceId].getService(Ci.nsIProtocolProxyService).asyncResolve(uri, 0, {
            notify: notify,
            onProxyAvailable: function(request, uri, proxy, status) {
                data.onProxyAvailableWithContext(request, uri, proxy, status, this);
            }
        });
    },
    
    /*
     * nsIProxyCallback
     */
    
    onProxyAvailableWithContext: function(request, uri, proxy, status, context) {
        //dump("StudyData.onProxyAvailable(...)\n");
        var instream = Cc[FileInStreamId].createInstance(Ci.nsIFileInputStream);
        instream.init(this.datafile, -1, -1, Ci.nsIFileInputStream.CLOSE_ON_EOF);
        var buffstream = Cc[BuffInStreamId].createInstance(Ci.nsIBufferedInputStream);
        buffstream.init(instream, 32768);
        
        var channel = this.https.newProxiedChannel(uri, proxy);
        channel.notificationCallbacks = this;
        channel.QueryInterface(Ci.nsIHttpChannel);
        channel.QueryInterface(Ci.nsIUploadChannel);
        channel.setUploadStream(buffstream, "application/octet-stream", -1);
        channel.requestMethod = "POST";
        channel.asyncOpen(this, { wrappedJSObject: { channel: channel, result: "", notify: context.notify }});
    },
    
    /*
     * nsIStreamListener
     */
    
    onStartRequest: function(request, context) {
        //dump("StudyData.onStartRequest(...)\n");
        context.wrappedJSObject.stream = Cc[BinInStreamId].createInstance(Ci.nsIBinaryInputStream);
    },
    
    onDataAvailable: function(request, context, input, offset, count) {
        //dump("StudyData.onDataAvailable(...)\n");
        context.wrappedJSObject.stream.setInputStream(input);
        context.wrappedJSObject.result += context.wrappedJSObject.stream.readBytes(count);
    },
    
    onStopRequest: function(request, context, status) {
        //dump("StudyData.onStopRequest(...)\n");
        // XXX should examine the MD5 in context.wrappedJSObject.result
        
        var channel = context.wrappedJSObject.channel;
        var response = 0;
        try {
            response = channel.responseStatus;
        } catch (e) { }
        var success = status == Components.results.NS_OK && response == 200;
        if (success) {
            Log().ok("Data submission succeeded");
        } else {
            Log().error("Data submission failed: " + status + ", HTTP " + response);
        }
        if (context.wrappedJSObject.notify) {
            var notebox = Cc[WindowMediatorId]
                             .getService(Ci.nsIWindowMediator)
                                 .getMostRecentWindow("navigator:browser")
                                     .document.getElementById("content")
                                         .getNotificationBox();
            var ok = {
                label: "OK",
                callback: function(notification, description) {
                    notification.close();
                }
            };
            if (success) {
                notebox.appendNotification(context.wrappedJSObject.notify.ok,
                                           "CtNotification",
                                           null,
                                           notebox.PRIORITY_INFO_MEDIUM,
                                           [ ok ]);
            } else {
                notebox.appendNotification(context.wrappedJSObject.notify.fail + " (status " + status + ", HTTP " + response + ")",
                                           "CtNotification",
                                           null,
                                           notebox.PRIORITY_WARNING_MEDIUM,
                                           [ ok ]);
            }
        }
    },
    
    /*
     * nsIBadCertListener2
     * MOZ Firefox 3
     */
    
    notifyCertProblem: function(socket, status, target) {
        Log().error("Data submission failed: certificate problem");
    },
    
    /*
     * nsIBadCertListener
     * MOZ Firefox 2
     */
    
    confirmUnknownIssuer: function(socket, cert, add) {
        // the certificate should match our copy
        add.value = Ci.nsIBadCertListener.ADD_TRUSTED_FOR_SESSION;
        if (cert.equals(this.cert)) {
            return true;
        }
        Log().error("Data submission failed: certificate wrong");
        return false;
    },
    
    confirmMismatchDomain: function(socket, target, cert) {
        Log().error("Data submission failed: certificate domain mismatch");
        return false; // the domain should match
    },
    
    confirmCertExpired: function(socket, cert) {
        Log().error("Data submission failed: certificate expired");
        return false; // the cert should be valid
    },
    
    notifyCrlNextupdate: function(socket, target, cert) {
        Log().error("Data submission failed: CRL");
    },
    
    /*
     * nsIChannelEventSink
     */
    
    onChannelRedirect: function(oldChannel, newChannel, flags) { },
    
    /*
     * nsIHttpEventSink
     */
    
    onRedirect: function(oldChannel, newChannel) { },
    
    /*
     * nsIProgressEventSink
     */
    
    onProgress: function(request, context, progress, max) { },
    
    onStatus: function(request, context, status, arg) { },
    
    /*
     * XPCOM
     */
    
    getHelperForLanguage: function(language) {
        return null;
    },
    
    getInterfaces: function getInterfaces(count) {
        var interfaces = [Ci.nsIClassInfo,
                          Ci.nsIInterfaceRequestor,
                          Ci.nsIRequestObserver,
                          Ci.nsIStreamListener,
                          (Ci.nsIBadCertListener2 ? Ci.nsIBadCertListener2 : Ci.nsIBadCertListener), // MOZ Firefox 3 vs 2
                          Ci.nsIChannelEventSink,
                          Ci.nsIHttpEventSink,
                          Ci.nsIProgressEventSink];
        count.value = interfaces.length;
        return interfaces;
    },
    
    getInterface: function(iid) {
        return this.QueryInterface(iid);
    },
    
    QueryInterface: function(iid) {
        if (iid.equals(Ci.nsISupports))
            return this;
        var interfaces = this.getInterfaces({});
        for (var ii = 0; ii < interfaces.length; ii++) {
            if (iid.equals(interfaces[ii]))
                return this;
        }
        throw Components.results.NS_ERROR_NO_INTERFACE;
    }
}

/*
 * Singleton factory for QIable component, with initialization.
 */
function SingletonFactory() {
    this.instance = null;
}

SingletonFactory.prototype.createInstance = function(outer, iid) {
    if (outer != null) throw Components.results.NS_ERROR_NO_AGGREGATION;
    
    if (this.instance == null) {
        this.instance = new StudyData();
        this.instance.init();
    }
    
    return this.instance.QueryInterface(iid);
}

const THE_FACTORY = new SingletonFactory();

/*
 * Module automagically loaded by Firefox.
 */

var Module = new Object();

Module.registerSelf = function(compMgr, fileSpec, location, type) {
    compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);

    compMgr.registerFactoryLocation(THIS_CID,
                                    EXTENSION_HUMAN_NAME + " " + THIS_DESCRIPTION,
                                    THIS_CONTRACTID,
                                    fileSpec, location, type);
}

Module.unregisterSelf = function(compMgr, fileSpec, location) {
    compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);

    compMgr.unregisterFactoryLocation(THIS_CID, fileSpec);
}

Module.getClassObject = function(compMgr, cid, iid) {
    if ( ! iid.equals(Components.interfaces.nsIFactory))
        throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    if (cid.equals(THIS_CID))
        return THE_FACTORY;

    throw Components.results.NS_ERROR_NO_INTERFACE;
}

Module.canUnload = function(compMgr) {
    return true;
}

function NSGetModule(compMgr, fileSpec) {
    return Module;
}