/*
 * Codetrail
 * 
 * Copyright (c) 2007-2008 Massachusetts Institute of Technology
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package jsjr;

import java.io.*;
import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.net.*;
import java.rmi.ConnectException;
import java.util.*;
import java.util.concurrent.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Implementation of remote method invocation connector.
 * 
 * @author maxg
 */
public class ConnectorImpl implements Connector {
    
    private final ServerSocket socket;
    private final Set<ConnectionListener> listeners;
    final Map<String, Object> registered;
    private final Thread thread;
    
    /**
     * Creates a new <code>Connector</code> to listen on the given TCP port.
     * 
     * @param port port to listen on
     * @throws IOException if a socket cannot be opened on the given port
     */
    public ConnectorImpl(int port) throws IOException {
        socket = new ServerSocket(port, 0, InetAddress.getByName(null));
        
        listeners = new HashSet<ConnectionListener>();
        registered = new ConcurrentHashMap<String, Object>();
        
        thread = new Thread(new Runnable() {
            public void run() {
                while ( ! Thread.interrupted()) {
                    try {
                        announce(new ConnectionImpl(ConnectorImpl.this, socket.accept()).started());
                    } catch (SocketException se) {
                    } catch (IOException ioe) {
                        ioe.printStackTrace(); // XXX
                    }
                }
            }            
        });
        thread.setDaemon(true);
        thread.setName("Connector<" + port + ">");
    }
    
    private void announce(Connection connection) {
        for (ConnectionListener listener : listeners) {
            listener.connected(connection);
        }
    }
    
    public void accept() {
        thread.start();
        
        try {
            new Socket(InetAddress.getByName(null), socket.getLocalPort()+1).close();
        } catch (Exception e) { }
    }
    
    public void close() throws IOException {
        thread.interrupt();
        socket.close();
    }
    
    public void addConnectionListener(ConnectionListener listener) {
        listeners.add(listener);
    }
    
    public void export(String id, Object object) {
        registered.put(id, object);
    }
    
    /**
     * Marshal a <code>String</code>.
     * 
     * @param string string to marshal
     * @return marshaled string
     */
    public static String marshal(final String string) {
        return (string == null ? "null" : ("str:" + string.replace("\n", "\nlstr:"))) + "\n";
    }
    
    /**
     * Make marshalable a <code>String</code>.
     * 
     * @param string string to be marshaled
     * @return marshalable for string
     */
    public static Marshalable marshalable(final String string) {
        return new Marshalable() {
            private static final long serialVersionUID = 1L;
            public String marshal() {
                return ConnectorImpl.marshal(string);
            }
        };
    }
}

class ConnectionImpl implements Connection {
    private static final Pattern CALL_PATTERN = Pattern.compile("call:target:(.*)&method:(.*)&formals:(\\d+)&arguments:(\\d+)");
    private static final int TARGET = 1, METHOD = 2, FORMALS = 3, CALL_ARGC = 4;
    private static final Pattern CLASS_PATTERN = Pattern.compile("class:(.*)");
    private static final Pattern REMOTE_PATTERN = Pattern.compile("remote:interface:(.*)&id:(.*)");
    private static final int INTERFACE = 1, ID = 2;
    private static final Pattern SERIAL_PATTERN = Pattern.compile("serial:class:(.*)&arguments:(\\d+)");
    private static final int CLASS = 1, SERIAL_ARGC = 2;
    private static final Pattern STRING_PATTERN = Pattern.compile("str:(.*)");
    private static final Pattern LONGSTRING_PATTERN = Pattern.compile("lstr:(.*)");
    
    private final ConnectorImpl connector;
    private final Socket socket;
    private boolean connected;
    
    private final BlockingQueue<String> incoming;
    private final Stack<Object> received;

    private final BlockingQueue<String> outgoing;
    
    private final String name;
    
    ConnectionImpl(ConnectorImpl connector, Socket socket) throws IOException {
        this.connector = connector;
        this.socket = socket;
        connected = false;
        incoming = new LinkedBlockingQueue<String>();
        received = new Stack<Object>();
        outgoing = new LinkedBlockingQueue<String>();
        name = "Connector<" + socket.getLocalPort() + "/" + socket.getPort() + ">";
    }
    
    private void input(String message) throws Exception {
        Matcher m;
        
        m = CALL_PATTERN.matcher(message);
        if (m.matches()) {
            Object target = connector.registered.get(m.group(TARGET));
            if (target == null) {
                throw new IOException("Invalid call to unknown target: " + m.group(TARGET));
            }
            String method = m.group(METHOD);
            int formals = Integer.parseInt(m.group(FORMALS));
            Class<?>[] classes = new Class[formals];
            for (int ii = formals - 1; ii >= 0; ii--) {
                classes[ii] = (Class<?>)received.pop();
            }
            int argc = Integer.parseInt(m.group(CALL_ARGC));
            Object[] arguments = new Object[argc];
            for (int ii = argc - 1; ii >= 0; ii--) {
                arguments[ii] = received.pop();
            }
            int varargc = 0;
            if (argc > formals) {
                // we have more args than formals
                varargc = argc - formals + 1;
            } else if (formals > 0 && classes[formals-1].isArray()) {
                // we have equal args and formals, but the last formal is an array
                varargc = 1;
            }
            if (varargc > 0) {
                String arrtype = classes[formals-1].getName();
                Object varargs = Array.newInstance(Class.forName(arrtype.substring(2, arrtype.length()-1)), varargc);
                System.arraycopy(arguments, formals-1, varargs, 0, varargc);
                Object[] compacted = new Object[formals];
                System.arraycopy(arguments, 0, compacted, 0, formals-1);
                compacted[formals-1] = varargs;
                arguments = compacted;
            }
            target.getClass().getMethod(method, classes).invoke(target, arguments);
            return;
        }
        
        m = CLASS_PATTERN.matcher(message);
        if (m.matches()) {
            received.push(Class.forName(m.group(1)));
            return;
        }
        
        m = REMOTE_PATTERN.matcher(message);
        if (m.matches()) {
            Class<?> clazz = Class.forName(m.group(INTERFACE) + "Proxy");
            Class<?>[] classes = new Class[] { Connection.class, String.class };
            Object[] arguments = new Object[] { this, m.group(ID) };
            Constructor<?> ctor = clazz.getConstructor(classes);
            ctor.setAccessible(true);
            Object proxy = ctor.newInstance(arguments); // XXX record and reuse proxy
            received.push(proxy);
            return;
        }
        
        m = SERIAL_PATTERN.matcher(message);
        if (m.matches()) {
            int argc = Integer.parseInt(m.group(SERIAL_ARGC));
            Object[] arguments = new Object[argc];
            for (int ii = argc - 1; ii >= 0; ii--) {
                arguments[ii] = received.pop();
            }
            //Class<?>[] classes = new Class[argc];
            //for (int ii = 0; ii < argc; ii++) {
            //    classes[ii] = arguments[ii].getClass(); // XXX doesn't work for nulls
            //}
            Constructor<?> constructor = null;
            for (Constructor<?> ctor : Class.forName(m.group(CLASS)).getConstructors()) {
                if (ctor.getParameterTypes().length == arguments.length) {
                    constructor = ctor;
                    break;
                }
            }
            if (constructor == null) {
                throw new IOException("Invalid " + m.group(CLASS) + " ctor args len: " + argc);
            }
            constructor.setAccessible(true);
            received.push(constructor.newInstance(arguments));
            return;
        }
        
        m = STRING_PATTERN.matcher(message);
        if (m.matches()) {
            received.push(m.group(1));
            return;
        }
        
        m = LONGSTRING_PATTERN.matcher(message);
        if (m.matches()) {
            received.push(received.pop() + "\n" + m.group(1));
            return;
        }
        
        if (message.equals("null")) {
            received.push(null);
            return;
        }
        
        if (message.equals("hello")) {
        	return;
        }
        
        throw new IOException("Invalid message: " + message);
    }
    
    public void call(String remoteId, String method, Marshalable... arguments) throws ConnectException {
        if ( ! connected) {
            throw new ConnectException("Lost connection");
        }
        synchronized (outgoing) {
            for (Marshalable argument : arguments) {
                outgoing.offer(argument.marshal());
            }
            outgoing.offer("call:target:" + remoteId + "&method:" + method + "&arguments:" + arguments.length + "\n");
        }
    }
    
    Connection started() throws IOException {
        start();
        return this;
    }
    
    void start() throws IOException {
        final Thread in = new Thread(new Runnable() {
            public void run() {
                try {
                    while ( ! Thread.interrupted()) {
                        try {
                            input(incoming.take());
                        } catch (InterruptedException ie) {
                            throw ie;
                        } catch (Exception e) {
                            e.printStackTrace(); // XXX
                        }
                    }
                } catch (InterruptedException ie) { }
            }
        }, name + " in");
        in.setDaemon(true);
        in.start();
        
        final Thread writer = new Thread(new Runnable() {
            final Writer output = new OutputStreamWriter(socket.getOutputStream());
            public void run() {
                try {
                    while (socket.isConnected()) {
                        output.write(outgoing.take());
                        output.flush();
                    }
                } catch (InterruptedException ie) {
                } catch (IOException ioe) {
                    ioe.printStackTrace(); // XXX
                }
            }
        }, name + " writer");
        writer.setDaemon(true);
        writer.start();
        
        Thread reader = new Thread(new Runnable() {
            final BufferedReader input = new BufferedReader(new InputStreamReader(socket.getInputStream()));
            public void run() {
                String line;
                try {
                    while ((line = input.readLine()) != null) {
                        incoming.put(line);
                    }
                } catch (SocketException se) {
                } catch (Exception e) {
                    e.printStackTrace(); // XXX
                }
                in.interrupt();
                writer.interrupt();
                connected = false;
            }
        }, name + " reader");
        reader.setDaemon(true);
        reader.start();
        
        connected = true;
        
        outgoing.offer("hello\n");
    }
}